#!/usr/local/bin/perl
##############################################################################
##! @file
##! @brief Looks at asn files and stores information about enumerations so 
##! that it can be merged with with bapitypes.h.
##! 
##! A platform-independent definition of enumerations et al are defined
##! in asn files. Comments have been added to these files but the normal
##! asn tools do not pass them through to the h files.
##!
##! This tool parses the asn files and stores information.
##! Then it parses the generated bapitypes.h file.
##! It merges information stored with information from bapitypes.h
##! and outputs a file with the merged information.
##!
##! @ingroup tp_tools tp_bapi
##!
##! @author Glenn C. Maxey
##
##    $Id: asn_bapitypes.pl,v 1.2 2002/11/20 15:23:01 gmaxe Exp $
##
## Copyright 2002 Created by Voyant Technologies, Inc., Westminster, Colorado, USA.
##
## Revision Information:
##
## $Log: asn_bapitypes.pl,v $
## Revision 1.2  2002/11/20 15:23:01  gmaxe
## Added exit codes so that wrapper scripts can catch errors properly.
##
## Revision 1.1  2002/02/02 01:48:52  gmaxe
## Supports index tokens from all dox files;
## supports used-by index tokens.
## asn_bapitypes.pl is a new file to handle remerging of comments
## into what the tools see.
##
##
##############################################################################

##############################################################################
##! @fn int BEGIN
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_bapi
##############################################################################
sub BEGIN {
   print "\n============  Starting asn_bapitypes.pl ==================================\n";
   $_file_list = "_file_list";
   $in_file = "";
   $f_type = "asn";
   $asn_path = "./";
   
   $comment_dox = "\-\-\/\/\!";  # --//!
   $comment_dox_normal = "\/\/\!";  # //!
   $comment_ignore = "\-\-\/\/"; # --//
   undef (%unproc_file); # key = $in_file; content = $_temp_in no blank lines   
   $code_enum = "ENUMERATED";
   $code_seq = "SEQUENCE";
   $code_def_mark = "\:\:\=";  # ::=
   undef (%ct_item);  # code top item; key = item name; content = dox comment
   undef (%enum_item); #enumeration item; key = enum+item; content = dox comment

   $td_enum = "typedef enum";    #type definition flag
   $td_struct = "typedef struct";    #type definition flag
   $comp_of_flag = "COMPONENTS OF ";  # says that we have to do something different.
   $comp_of_flag2 = "\;\,\;";  # says that we have to do something different.

   push (@INC, `pwd`);
   push (@INC, '../perl');
   if (0){
      print (@INC, "\n");
   }
}

##############################################################################
##! @fn int main
##! @brief The main program.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_bapi
##############################################################################
# sub main {
{
   ##############################################################################
   ## Start of Main Program
   ##############################################################################

   if (@ARGV < 1) {
      &using_asn_bapi();
      exit(1);
   }

   # Get path to asn files if there is one.
   if (@ARGV > 0) {
      $asn_path = @ARGV[0];
      if ($asn_path =~ /\//) {
         print "The path specified is $asn_path\n";
   
      } else {
         print "The input argument \"$asn_path\" requires a forward slash (\/).\n";
         &using_asn_bapi();
         exit(1);
      }
   } # if 1 or more arguments

   if (@ARGV > 1) {
      $bapitypes_file = @ARGV[1];
      print "The master file is $bapitypes_file\n";
      unless (open ( IN_MASTER, $bapitypes_file)) {
         print "Cannot open file \"$bapitypes_file\"\n";
         exit(0);
      }
      close (IN_MASTER);
   } else {
      print "Need to have the master bapitypes file.\n";
      &using_asn_bapi();
      exit(1);
   } # if 2 or more arguments

   if (&create_file_list() == 0) {
      &process_file_list();
   } else {
      print "Cannot create file list\n";
      exit(0);
   }
   
   &fix_enum_list();
   
   &read_manipulate_master();
   
   $out_file = "$bapitypes_file.gen";
   # print "before passing: $out_file\n";
   
   &write_output_file($out_file);


   if (@file_errors) {
      print "\n============  Summary of errors ==================================\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
   }


   ##############################################################################
   ## End of most of work
   ##############################################################################
}

##############################################################################
##! @fn int create_file_list
##! @brief Creates a temporary file that lists all of the files which should
##! be processed.
##!
##! @return The file is created in the directory given by the first input
##! parameter to the program.
##!
##! @bug 
##!
##! @ingroup tp_bapi
##############################################################################
sub create_file_list {
   $_file_list = "$asn_path$_file_list";

   if (system ("ls $asn_path*.$f_type* > $_file_list")) {
      print "ls $asn_path*.$f_type* > $_file_list\n";
      print "There is no $asn_path directory. Nothing done.\n";
      push (@file_errors, "There is no $asn_path directory. Nothing done.\n");
      return(1);
   }

   return (0);
} # create_file_list

##############################################################################
##! @fn int process_file_list
##! @brief Gathers information from all files given in the list.
##!
##! Opens the file that contains the list of files that are of interest.
##! Then it opens each of those files one at a time.
##! 
##! @return This creates two hash tables. One is for the code top item
##! and the other is for enumeration values (of a code top item).
##! A "code top item" is some code element definition. Top means that it
##! is not part of an enumeration.
##! 
##! In both cases, what the hash stores is the doxygen comment associated
##! with it.
##!
##! The key into ct_item is the name of the item fixed up to what will be
##! seen later in bapitypes.
##!
##! The key into enum_item is the name of the enumeration plus the enumeration item
##! fixed up to what will be seen later in bapitypes.
##!
##! @bug None.
##!
##! @ingroup tp_bapi
##############################################################################
sub process_file_list {

   unless (open ( IN_LIST, $_file_list)) {
      print "Cannot open file \"$_file_list\"\n";
      push (@file_errors, "Cannot open file \"$_file_list\"\n");
      exit(1);
   }
   
INCOMING:   while (<IN_LIST>){    # read a line from file into $_
      $in_file = $_;
      $in_file =~ s/\n//;
      printf ("Working on... %s\n", $in_file);
      unless (open ( IN_FILE, $in_file)) {
         print "Cannot open file \"$in_file\"\n";
         push (@file_errors, "Cannot open file \"$in_file\"\n");
         next INCOMING;
      }
      $_temp_in = "";   # reset temporary input buffer
      $_dox_comment = "";
      $_dox_flag = 0;
      $_seq_flag = 0;
      $_enum_flag = 0;
      $_rem_enum = "";
INC_LINE:   while (<IN_FILE>){ # read in entire file
         if (0){
            print "enum=$_enum_flag;seq=$_seq_flag\n";
         }
         if (/^$comment_dox/){
            # pass through doxygen comments
            $_temp_in .= $_;
            # collect doxygen comments for top level item
            $_dox_flag = 1;
            $_temp = $_;
            $_temp =~ s/$comment_dox/$comment_dox_normal/g;
            if (0){
               print "hola =$_temp\n";
            }
            $_dox_comment .= $_temp;
            $_enum_flag = 0;
            $_seq_flag = 0;
         } elsif (/^$comment_ignore/) {
            # ignore regular comments
            next INC_LINE;
         } elsif (!/[\S]+/) {
            # ignore blank lines
            # print "ggg :$_\n";
            next INC_LINE;
         } elsif ((/$code_def_mark/) && ($_seq_flag == 0) && ($_enum_flag == 0)) {
            # starting enumeration stuff
            $_temp_in .= $_;
            # pick-off code item name
            # name is [0]; what it is [1]
            @code_def_line = split (/$code_def_mark/, $_, 2);
            $code_def_line[0] =~ s/[\s]//g;  # white space
            $code_def_line[0] =~ s/\-/\_/g;  # change name
            if ($code_def_line[1] =~ /$code_seq/){
               $_seq_flag = 1;
               $_rem_seq = $code_def_line[0];
               if (0){
                  print ">>$code_def_line[0]\n";
               }
            } elsif ($code_def_line[1] =~ /$code_enum/) {
               $_enum_flag = 1;
               $_rem_enum = $code_def_line[0];
               if (0){
                  print ">>$code_def_line[0]\n";
               }
            }
            if ($_dox_flag == 1) { 
               $ct_item{$code_def_line[0]} = $_dox_comment;
               $_dox_flag = 0;
            }
            $_dox_comment = "";
            if (0){
               print "hola =$_rem_enum\n";
            }
         } elsif (($_enum_flag == 1) && (!/\}/) && (!/\{/)) {
            # Within an enumeration
            # pick-off enumeration item name
            # name is [0]; dox comment is [1]
            @code_def_line = split (/$comment_dox/, $_, 2);
            if ($code_def_line[0] !~ /\S+/) {
               # if we're inside of an enumeration and there is no
               # name to the left of the comment, then this
               # comment should be appended to the previous one.
               $enum_item{$_temp} .= "\n$comment_dox_normal$code_def_line[1]"; 
               if (0) {
                  print "stop $_rem_enum\n$_";
                  print "[0]=$code_def_line[0]=\n";
                  print "[1]=$code_def_line[1]=\n";
                  print "$enum_item{$_temp}\n";
               }
               next INC_LINE;
            }
            $code_def_line[0] =~ s/[\s]//g;  # white space
            $code_def_line[0] =~ s/\-/\_/g;  # change name
            $code_def_line[0] =~ s/\,//g;    # get rid of commas
            $code_def_line[1] =~ s/\n//g;    # get rid of carriage return
            # name is [0]; everything else isn't important
            @enum_item_name = split (/\(/, $code_def_line[0], 2);
            # enumeration item; key = enum+item; content = dox comment
            $_temp = "$_rem_enum$enum_item_name[0]";
            $enum_item{$_temp} = "$comment_dox_normal$code_def_line[1]"; 
         } elsif (($_seq_flag == 1) && (!/\}/) && (!/\{/)) {
            # Within a sequence
            if (0){
               print "REPEAT enum=$_enum_flag;seq=$_seq_flag\n";
            }
            # pick-off sequence item name
            # name is [1]; dox comment is somewhere [2]
            @name_chunk = split (/\s+/, $_, 3);
            $_temp = "$_rem_seq$name_chunk[1]";
            if (0) {
               print "input>>$_";
               print "[0]=$name_chunk[0]=\n";
               print "[1]=$name_chunk[1]=\n";
               print "[2]=$name_chunk[2]=\n";
               print "[3]=$name_chunk[3]=\n";
               print "last\n";
            }
            # nothing is [0]; dox comment is [1]
            @code_def_line = split (/$comment_dox/, $name_chunk[2], 2);
            if (0) {
               print "input>>$name_chunk[2]";
               print "[0]=$code_def_line[0]=\n";
               print "[1]=$code_def_line[1]=\n";
               print "last\n";
            }
            $code_def_line[1] =~ s/\n//g;    # get rid of carriage return
            if (/$comp_of_flag/){
               # If the sequence has components it, then flag them for
               # later processing.
               @get_comp1 = split (/$comp_of_flag/, $_, 2);
               @get_comp2 = split (/\s+/, $get_comp1[1], 2);
               $get_comp2[0] =~ s/\,//g;
               if ($get_comp2[0] =~ /\./){
                  @get_comp3 = split (/\./, $get_comp2[0], 2);
                  $get_comp2[0] = $get_comp3[1]; 
               }
               # Create a place-holder key in the enumeration hash
               # to be resolved later.
               $_temp = "$_rem_seq$comp_of_flag2$get_comp2[0]";
               if (0){
                  print "key=$_temp\n";
               }
               $enum_item{$_temp} = "$comment_dox_normal$code_def_line[1]";
            } else {
               # enumeration item; key = enum+item; content = dox comment
               $enum_item{$_temp} = "$comment_dox_normal$code_def_line[1]"; 
            }
         } elsif (($_enum_flag == 1) && (/\}/)) {
            # At the end of an enumeration
            $_temp_in .= $_;
            $_enum_flag = 0;
         } elsif (($_seq_flag == 1) && (/\}/)) {
            # At the end of an enumeration
            $_temp_in .= $_;
            $_seq_flag = 0;
         } else {
            # pass through code stuff
            $_temp_in .= $_;
         }
      } # while within a given file
      close (IN_FILE);
      
      if (0){
         print $_temp_in;
      }
      $unproc_file{$in_file} = $_temp_in;
      # if possible, fix the list with things we know
      &fix_enum_list();
   } # while getting a specific file
   close (IN_LIST);
   
   if (0) {
      foreach $ddd (sort keys %ct_item){
         print "ct key=$ddd content=$ct_item{$ddd}\n";
      }
      exit(1);
   }
   if (0) {
      foreach $ddd (sort keys %enum_item){
         print "enum key=$ddd content=$enum_item{$ddd}\n";
      }
      exit(1);
   }
   if (0) {
      foreach $ddd (keys %unproc_file){
         print "key=$ddd\n=====\ncontent=$unproc_file{$ddd}\n";
      }
   }
   
   
   return (0);

} # process_file_list

##############################################################################
##! @fn int fix_enum_list
##! @brief Resolves enumeration items that were built from other components.
##!
##! Duplicates entries in the hash table for the equivalent of the component.
##! Removes place-holder entries in the hash that say component.
##!
##! @bug This is called several times so that it can resolve things as new
##! things are learned.
##! @ingroup tp_bapi
##############################################################################
sub fix_enum_list {
   # print "Fixing the enumeration list of nested components.\n";
   $_key_flag = 0; 
     
   # Fix the enum list so that components are resolved.
   foreach $enum_key (sort keys %enum_item){
      if ($enum_key =~ /$comp_of_flag2/){
         @key_part = split (/$comp_of_flag2/, $enum_key, 2);
         # print "$key_part[0] and $key_part[1]\n";
         foreach $key_again (sort keys %enum_item){
            # get all keys that match component key [1] and make
            # them part of the other component key [0].
            if ($key_again =~ /^$key_part[1]/){
               # Only set the flag if the item is resolved.
               $_key_flag = 1;
               # We need to correlate key
               @src_key = split (/$key_part[1]/, $key_again, 2);
               # print "$key_part[0]=$key_part[1]=$key_again=$src_key[0] and $src_key[1]\n";
               $_temp_src = "$key_again";
               $_temp_tgt = "$key_part[0]$src_key[1]";
               # print "$_temp_tgt to be $_temp_src\n";
               $enum_item{$_temp_tgt}=$enum_item{$_temp_src};
            }
         } # for each key_again
         # remove the element from the hash that says "component of"
         if ($_key_flag == 1){
            delete($enum_item{$enum_key});
            $_key_flag = 0;
         }
      } # if enum_key is a component
   } # for each enum_key
   if (0) {
      print "This is enumeration list again\n=====================\n";
      foreach $enum_key (sort keys %enum_item){
         print "enum key=$enum_key content=$enum_item{$enum_key}\n";
      } # for each enum_key
   }
   return (1);

} # fix_enum_list


##############################################################################
##! @fn int read_manipulate_master
##! @brief Reads in the file generated by development tools and appends
##! any dox comments that those tools may have stripped from the source files.
##!
##! @return $bapitypes contains the original information along with comments.
##!
##! @ingroup tp_bapi
##############################################################################
sub read_manipulate_master {

   if (0) {
      print "The master file is $bapitypes_file\n";
   } 
   unless (open ( IN_MASTER, $bapitypes_file)) {
      print "Cannot open file \"$bapitypes_file\"\n";
      return (0);
   }
   $_in_enum = 0;
   $_rem_enum = "";
   # Get the master definitions
   while (<IN_MASTER>){    # entire master file into memory.
      # Figure out what the line says.
      if (/^$td_enum/) {
         # We have a type definition that needs comments
         $_in_enum = 1;
         # Figure out which typedef we've got
         # pick-off code item name
         # nothing is [0]; name is [1]; typedef start is [2]
         @code_def_line = split (/$td_enum|\s/, $_, 2);
         $code_def_line[1] =~ s/\s//g;  # white space
         $code_def_line[1] =~ s/\{//g;  # white space
         $_rem_enum = $code_def_line[1];
         $bapitypes .= $ct_item{$code_def_line[1]};
         if (0){
            print $ct_item{$code_def_line[1]};
            print "what=$code_def_line[1]\n";
         }
         if (1) {
            # Adam E's suggestion about which keyword to eliminate
            $bapitypes .= "\/\/ Next line changed to fake out Doxygen\n\/\/ $_";
            $bapitypes .= "$td_enum \{\n";
         } else {
            $bapitypes .= $_;
         }
      } elsif (/^$td_struct/){
         # We have a type definition that needs comments
         $_in_struct = 1;
         # Figure out which typedef we've got
         # pick-off code item name
         # nothing is [0]; name is [1]; typedef start is [2]
         @code_def_line = split (/$td_struct|\s/, $_, 2);
         $code_def_line[1] =~ s/\s//g;  # white space
         $code_def_line[1] =~ s/\{//g;  # white space
         $_rem_enum = $code_def_line[1];
         $bapitypes .= $ct_item{$code_def_line[1]};
         if (0){
            print $ct_item{$code_def_line[1]};
            print "what=$code_def_line[1]\n";
         }
         $bapitypes .= $_;
      } elsif ( ($_in_enum == 1) && (!/\}/)) {
         # Middle of an enumeration
         # name is [0]; value is [1]
         $_temp_line = $_;
         $_temp_line =~ s/\n//;
         @code_def_line = split (/\=/, $_temp_line, 2);
         $code_def_line[0] =~ s/\s//g;  # white space
         $_temp = "$_rem_enum$code_def_line[0]";
         if (exists($enum_item{$_temp})) {
            if (0){
               print "what=$_temp=$enum_item{$_temp}\n";
            }
            $_temp_line .= " $enum_item{$_temp}\n"; 
         } else {
            $_temp_line .= "\n"; 
         }
         $bapitypes .= $_temp_line;
      } elsif ( ($_in_struct == 1) && (!/\}/)) {
         # Middle of an structure
         # name is [2]; value is [1]
         $_temp_line = $_;
         $_temp_line =~ s/\n//;
         @code_def_line = split (/\s+/, $_temp_line, 3);
         if (0) {
            print "[0]=$code_def_line[0]\n";
            print "[1]=$code_def_line[1]\n";
            print "[2]=$code_def_line[2]\n";
         }
         $code_def_line[2] =~ s/\s//g;  # white space
         $code_def_line[2] =~ s/\;//g;  # semicolin
         $code_def_line[2] =~ s/\n//g;  # carriage return
         $_temp = "$_rem_enum$code_def_line[2]";
         if (exists($enum_item{$_temp})) {
            if (0){
               print "what=$_temp=$enum_item{$_temp}\n";
            }
            $_temp_line .= " $enum_item{$_temp}\n"; 
         } else {
            $_temp_line .= "\n"; 
         }
         $bapitypes .= $_temp_line;
      } elsif (($_in_enum == 1) && (/\}/)) {
         # End of the enumeration
         # We're doing something unique here:
         if (1) {
            # Adam E's suggestion about which keyword to eliminate
            $bapitypes .= $_;
         } else {
            $bapitypes .= "\/\/ Next line changed to fake out Doxygen\n\/\/ $_";
            $bapitypes .= "\} \;\n";
         }
         $_in_enum = 0;
      } elsif (($_in_struct == 1) && (/\}/)) {
         # End of the structure
         $bapitypes .= $_;
         $_in_struct = 0;
      } else {
         $bapitypes .= $_;
      }
      
      
   }
   close (IN_MASTER);
   if (0){
      print $bapitypes;
   }

   return (1);
} # read_manipulate_master

##############################################################################
##! @fn int write_output_file
##! @brief Uses the second argument of the input (master file) and appends
##! ".gen" to the name and outputs the enhanced data.
##!
##! @return A new file named similar to the input file but with a ".gen" extension.
##!
##! @ingroup tp_bapi
##############################################################################
sub write_output_file {
   $file_out = $_[0];
   print "Writing the $file_out output file.\n";
   
   # $bapitypes =~ s/typedef enum/enum/g;
   
   unless (open ( OUT_FILE, ">$file_out")) {
      push (@file_errors, "Cannot open file \"$file_out\"\n");
      print "Cannot open file \"$file_out\"\n";
   }
   print (OUT_FILE "\/\/\! \@file\n");
   print (OUT_FILE "\/\/\! \@brief Slightly modified version of the original header file.\n");
   print (OUT_FILE "\/\/\! \n\/\/\! Comments that were stripped by external proprietary tools\n");
   print (OUT_FILE "\/\/\! have been re-inserted in this version which is only used\n");
   print (OUT_FILE "\/\/\! to generate HTML documentation.\n");
   print (OUT_FILE "\/\/\! \n\/\/\! A couple of other modifications were carried out\n");
   print (OUT_FILE "\/\/\! to make Doxygen happy.\n");
   print (OUT_FILE "\/\/\! \n\/\/\! THIS FILE IS GENERATED!!!\n");
   print (OUT_FILE "\/\/\! \n\/\/\! \@ingroup bapi\n");


   print (OUT_FILE "$bapitypes");
   close (OUT_FILE);

} # write_output_file



##############################################################################
##! @fn int using_asn_bapi
##! @brief What to do when no arguments are given.
##! @param None
##! @return None
##! 
##! @bug None
##! @ingroup tp_bapi
##############################################################################
sub using_asn_bapi {
   print "asn_bapitypes.pl <path to asn files> <path and name of bapitypes>\n";
   print "Outputs <path and name of bapitypes> with a \".gen\" extension.\n";
   print "\nTerminating asn_bapitypes.pl without doing anything.\n";
   return;
}





##############################################################################
##! @fn int END
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_bapi
##############################################################################
sub END {
   undef ($_file_list); # "_file_list";
   undef ($in_file ); #  "";
   undef ($f_type ); #  "asn";
   undef ($asn_path ); #  "./";
   
   undef ($comment_dox ); #  "\-\-\/\/\!";  # --//!
   undef ($comment_dox_normal ); #  "\/\/\!";  # //!
   undef ($comment_ignore ); #  "\-\-\/\/"; # --//
   undef (%unproc_file); # key = $in_file; content = $_temp_in no blank lines   
   undef ($code_enum ); #  "ENUMERATED";
   undef ($code_seq ); #  "SEQUENCE";
   undef ($code_def_mark ); #  "\:\:\=";  # ::=
   undef (%ct_item);  # code top item; key = item name; content = dox comment
   undef (%enum_item); #enumeration item; key = enum+item; content = dox comment

   undef ($td_enum ); #  "typedef enum";    #type definition flag
   undef ($td_struct ); #  "typedef struct";    #type definition flag

   
   ##############################################################################
   ## Memory clean-up.
   ##############################################################################

   print "\n============  Finished asn_bapitypes.pl ==================================\n";
   exit(0);
} # END
