#!/usr/local/bin/perl

#///////////////////////////////////////////////////////////////////////////////
##! @file
##! @brief Input filter for Doxygen for most C/C++ code files.
##! 
##! This changes code comments and other items on-the-fly so that it can be
##! more effectively processed by Doxygen. In particular, our coding standard
##! calls for slash-slash-bang comments. However, these comments are not interpretted as
##! blocks even when placed together. Sometimes comments from a file header
##! were being placed on code items. By changing the comment style to be
##! C-style, a true comment block could be created with no misinterpretation.
##! 
##! Other changes that this file does include:
##! 
##! <ul><li> formatting some auto-generated classes to be correct for extraction.</li>
##! <li>replacing @_bug with @_lim -- defined in the project file -- so that
##! the word "bug" never appears in our output.</li>
##! </ul>
##!
##! @param Input source file.
##! @return Output is file with changes.
##!
##! @ingroup tp_tools tp_dox
##!
##! @author Glenn C. Maxey
##
#// $Id: dox_bug_filter.pl,v 1.9 2002/05/10 16:53:54 gmaxe Exp $
#//
#// 2002 Created by Voyant Technologies, Inc., Westminster, Colorado, USA.
#//
#// Permission to use, copy, modify, and distribute this software and its 
#// documentation under the terms of the GNU General Public License is hereby 
#// granted. No representations are made about the suitability of this software 
#// for any purpose. It is provided "as is" without express or implied warranty. 
#// See the GNU General Public License (http://www.gnu.org/copyleft/gpl.html) 
#// for more details.
#// 
#// Documents produced by this script are derivative works derived from the 
#// input used in their production; they are not affected by this license.
#//
#// $Log: dox_bug_filter.pl,v $
#// Revision 1.9  2002/05/10 16:53:54  gmaxe
#// Added testing for inline comments; broke them out so that they
#// appear on a line by themselves IN FRONT OF their owning code items.
#// PTR#:  1-N8KS
#//
#// Revision 1.8  2002/04/06 01:55:08  gmaxe
#// New files and new general constructs for handling tags. Uses globals better.
#// The html_look* files are designed to handle spider tracing of html systems.
#//
#// Revision 1.7  2002/01/16 19:49:06  gmaxe
#// Removed much of revision history in comments.
#//
#///////////////////////////////////////////////////////////////////////////////

BEGIN {
   local ($comment_count) = 0;  # count if you're in the comments somewhere.
}

NEW_LINE: while (<>) {

   ######### 
   # Begin replacement of //! comments with /** ... **/ comments
   # This purposely does not do NEW_LINE or printing.
   ######### 
   if (/^\/\/\!/) {
      if ($comment_count == 0){
         # first line of a comment block
         $comment_count++;
         $_ =~ s/\/\/\!/\/\*\*/;
      } else {
         # Some line in the middle of a comment block.
         $comment_count++;
         ###
         # Changed to have middle stuff with no asterix.
         # $_ =~ s/\/\/\!/ \*\*/;
         ###
         $_ =~ s/\/\/\!//;
      }
   } elsif ($comment_count > 0){
      # We were in a comment block; need to terminate it.
      $comment_count = 0;
      $_ = " \*\*\/\n" . $_ ;
   }
   ######### End Comment style change.
    


   ######### 
   # Fixing class definitions
   ######### 
   if (/^\s*_CLASSDEF\s*\(\s*([a-zA-Z0-9_]+)\s*\)/)   {
      print "class $1;";
      print "typedef $1 *p$1;";
      print "typedef $1 &r$1;";
      print "typedef $1 *&rp$1;";
      print "typedef const $1 *pc$1;";
      print "typedef const $1 &rc$1;\n";
      next NEW_LINE;
   } 
   if (/^\s*_TYPEDEF\s*\(\s*([a-zA-Z0-9_]+)\s*\)/) {
      print "typedef $1 *p$1;";
      print "typedef $1 &r$1;";
      print "typedef $1 *&rp$1;";
      print "typedef const $1 *pc$1;";
      print "typedef const $1 &rc$1;\n";
      next NEW_LINE;
   } 

   #########
   # Rearrangement of the inline //! comments such
   # that those comments appear on a line before their
   # code item. 
   ######### 
   if (/\/\/\!/) {
      @line_chunk = split (/\/\/\!/, $_, 2);
      print "\/\/\! $line_chunk[1]";
      print "$line_chunk[0]\n";
      next NEW_LINE;
   }


   ######### 
   # Begin replacement of \bug or @bug to @lim
   ######### 
   if (/\\bug|\@bug/) {
      $_ =~ s/\\bug/\@lim/;
      $_ =~ s/\@bug/\@lim/;
      print $_;
      next NEW_LINE;
   } 
   
   ######### 
   # Default Fall through
   ######### 
   print $_;

}  







