#!/usr/local/bin/perl
package xhelp;
##############################################################################
##! @file
##! @brief Generates a file with command definitions and doxygen comments,
##! intended as input to Doxgyen without as much noise.
##! 
##! xhelp_gen.pl input_scope_package root_path_to_files output_file
##!
##! @param input_scope_package A file path and name to a perl package that has the 
##! 2-b-included and 2-b-excluded xhelp commands.
##!
##! @param root_path_to_files must be terminated with a forward (\/) slash.
##! If output file does not have a forward slash (\/) -- an indication
##! of a path --, then the <root path to files> is assumed.
##!
##! @param output_file name of the first output file. This plus another file
##! with a leading underscore are generated. 
##! The file without the underscore is intended as the input to Doxygen.
##!
##! The input scope package is required to have:<ul>
##!
##! <li>The xscope package definition.</li>
##! <li>Routine: <ul>
##!    <li>declare_variable.</li>
##!    <li>Routine: memory_clean_up.</li>
##! </ul></li>
##! <li>Array: <ul>
##!    <li>x_names, which contains all prefixes that could be of interest.</li>
##!    <li>needed_in, which contains a list of xhelp commands that are
##! absolutely needed-to-be-in.</li>
##!    <li>needed_out, which contains a list of commands that we don't
##! want to expose at all if they happen to come through.</li>
##!    <li>include_f_type, which are file types that should be viewed.</li>
##! </ul></li>
##! </ul>
##!
##! The definitions of "clues" in the xscope file is the key to the success
##! of this tool. It needs a hint on what to look for.
##!
##! @note The @ingroup designation was the most useful tag for this tool
##! when standards in Doxygen comments were followed.
##!
##! The intent is that different projects can define their own files that have
##! the xscope definitions of what to look for.
##!
##! This program does divide and conquor. <ol>
##!
##! <li>Using the prefix list, it greps through the code looking at the file
##! types of interest.</li>
##! <li>Using the command list, it greps through the code looking at the file
##! types of interest.</li>
##! <li>The temporary file with grep results is stripped of interesting entries.</li>
##! <li>A file hash is created that is a hash of hashes. <br>
##! <pre>$file_hash {$src_file} {$code_item} {r} {$return_type} = unimportant number </pre>
##! <pre>$file_hash {$src_file} {$code_item} {p} {$prototype} = doxygen </pre>
##! It contains:<ul>
##! <li>{$src_file} the source code files where items of interest were found.</li>
##! <li>{$code_item} code items of interest, each associated with a source file.</li>
##! <li>{r} {$return_type} a hash of return types for each code item; supports
##! overloading.</li>
##! <li>{p} {$prototype} a hash of code definitions for each code item; supports
##! overloading. It contains to the doxygen comment block.</li>
##! </ul></li>
##! <li>The source files from the hash are opened and searched for their respective
##! code item definitions.</li>
##! <li>The source files are searched for comment blocks associated with
##! the code item definitions.</li>
##! <li>The file hash is fleshed out for the prototype definitions and 
##! their respective comment blocks.</li>
##! <li>The file hash is stepped through and output to generated files.
##! The generated files are intended for input to doxygen.
##! <ol>
##! <li>_list_<name>.gen file is for accountability and lists all code items and 
##! whether or not they were accompanied by a doxygen template. This is
##! sorted by source file name. Intended to communicate what is in and out.</li>
##! <li><name>.gen contains a fake function definition with the associated 
##! doxygen template. This list is sorted by code item. Intended for xhelp.
##! <li>class_<name>.gen file contains a fake class definition with the associated doxygen 
##! template. This list is sorted by class name. Intended for APIs implemented
##! through classes (e.g., DSP drivers). Generated when $xscope::sapi==0.</li>
##! <li>dox_template_<name> contains just the associated doxygen template. The
##! list is sorted based on the @fn code item name. Intented for SAPI in conjunction
##! with the vproto.gen that has the function prototypes. Generated when
##! $xscope::sapi==1; this is set in the sapi.pm file.</li>
##! </ol></li>
##! </ol>
##!
##! @ingroup tp_tools tp_xhelp
##!
##! @author Glenn C. Maxey
##
##    $Id: find_extract.pl,v 1.1 2002/12/24 15:05:51 gmaxe Exp $
##
## Copyright 2002 Created by Voyant Technologies, Inc., Westminster, Colorado, USA.
##
## Revision Information:
##
## $Log: find_extract.pl,v $
## Revision 1.1  2002/12/24 15:05:51  gmaxe
## New name and tweaks to add full paths.
##
## Revision 1.13  2002/11/20 15:23:10  gmaxe
## Added exit codes so that wrapper scripts can catch errors properly.
##
## Revision 1.12  2002/07/29 22:46:07  gmaxe
## Enhanced list output to support _TBD_.
##
## Revision 1.11  2002/07/26 18:56:15  gmaxe
## Got rid of old definitions and migrated new structures into all;
## enhanced xhelp output file names and format; now everything is
## alphebetized.
##
## Revision 1.9  2002/06/10 21:45:05  gmaxe
## Updated output to alphebetize things when possible.
## PTR#: 1-09T1
##
## Revision 1.8  2002/04/24 00:08:57  gmaxe
## Improved how items are excluded from xhelp_gen.pl, because
## they were creeping back in when file got processed.
##
## Revision 1.7  2002/04/12 18:01:17  gmaxe
## Various tweaks; work on tag extraction and change_nav;
## more exclused special perl characters from indexer;
## improved support for sapi group pm files in xhelp.
## spider tool now works regularly and as expected.
##
## Revision 1.6  2002/03/28 21:20:51  gmaxe
## Added support for ingroup extraction.
##
## Revision 1.5  2002/03/25 22:50:09  gmaxe
## Fixed another perl special character problem in index;
## Fixed a naming problem in xhelp gen.
##
## Revision 1.4  2002/03/21 23:26:49  gmaxe
## Minor doc changes to indexer;
## xhelp enhanced s.t. it generates class files when needed.
##
## Revision 1.3  2002/02/21 19:54:09  gmaxe
## Improved comments and reliability.
##
## Revision 1.2  2002/02/21 01:39:00  gmaxe
## Working version of xhelp command extractor and prototype generator.
##
##############################################################################

##############################################################################
##! @fn int BEGIN
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_xhelp
##############################################################################
sub BEGIN {
   print "\n============  Starting xhelp_gen.pl ==================================\n";
   $_file_list = "_file_list";
   $in_file = "";
   $root_path = "./";
   $scope_pm = "";
   $no_scope_file = 0;
   $xscope::sapi = 0; # This is set in the generated sapi.pm; When needed, will be overwritten by pm file.
   
   push (@INC, `pwd`);
   push (@INC, '../perl');
   if (0){
      print (@INC, "\n");
   }
   
   # Get scope files if there is one.
   if (@ARGV > 0) {
      $scope_pm = @ARGV[0];
      
      unless (open ( IN_LIST, $scope_pm)) {
         push (@file_errors, "Cannot open file \"$scope_pm\"\n");
         $no_scope_file++;
      }
      close (IN_LIST);
      push (@INC, $scope_pm);
      
      if (!@file_errors) {
         #####
         # All global variables are defined in the following file
         #####
         require $scope_pm;
   
         if (&xscope::declare_variables()) {
            print "Variables initialized from $scope_pm.\n";
         } else {
            push (@file_errors, "Could not initialize variables from $scope_pm.\n");
         }
      } # if not @file_errors
   } else {
      push (@file_errors, "Need to provide a scope file as first argument.\n");
   } # if 1 or more arguments

   # Get path to code files 
   if (@ARGV > 1) {
      $root_path = @ARGV[1];
      if ($root_path =~ /\/$/) {
         print "The path specified is $root_path\n";
      } else {
         push (@file_errors, "The input argument \"$root_path\" requires a forward slash (\/) at the end.\n");
      }
   } # if 1 or more arguments

   if (@ARGV > 2) {
      $gen_proto = @ARGV[2];
      if ($gen_proto !~ /\//) {
         # if the output file does not have a path, then
         # put it in the root location
         if (1) {
            @chunk = split (/\./, $gen_proto);
            $last = pop (@chunk);
            $all = join ("\.", @chunk);
            $gen_list = join ("", $root_path, $all, "_list.", $last);
            $gen_class = join ("", $root_path, $all, "_class.", $last);
            $gen_template = join ("", $root_path, $all, "_dox_template.", $last);
            $gen_proto = "$root_path$gen_proto";
         }
      } else {
         if (1) {
            @chunk = split (/\./, $gen_proto);
            $last = pop (@chunk);
            $all = join ("\.", @chunk);
            $gen_list = join ("", $all, "_list.", $last);
            $gen_class = join ("", $all, "_class.", $last);
            $gen_template = join ("", $all, "_dox_template.", $last);
         }
      }
      print "Potential output files are: \n";
      print "\tlist: \t$gen_list\n";
      print "\tproto:\t$gen_proto\n";
      print "\tclass:\t$gen_class\n";
      print "\tdox:  \t$gen_template\n";
   } else {
      push (@file_errors, "Need to have the output prototype file.\n");
   } # if 3 or more arguments

} # BEGIN
   
   
   

##############################################################################
##! @fn int main
##! @brief The main program.
##!
##! @param file_errors If file errors are encountered before the main
##! routine is hit, it prints them out and exits. 
##!
##! @return None.
##!
##! Controls the program flow. It creates a temporary file with the
##! grep results, processes the temporary file, creates a file hash
##! with source code targets, then outputs files needed for Doxgyen.
##!
##! @bug None
##! @ingroup tp_xhelp
##############################################################################
# sub main {
{
   ##############################################################################
   ## Start of Main Program
   ##############################################################################
   if (@file_errors) {
      # Makes no sense to go on if input parameters are off.
      print "\n============  Summary of errors ==================================\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
      &xhelp::using_xhelp();
      exit(1);
   }

   $_file_list = $root_path . "_eraseme";
   if (1){
      if (&create_xhelp_grep_results($_file_list) == 0) {
         &create_xhelp_hash($_file_list);
      } else {
         print "Cannot create file list\n";
         exit(0);
      }
   } # if test

   $_file_list = $root_path . "_eraseme2";
   if (1){
      if (&create_ingroup_grep_results($_file_list) == 0) {
         &create_ingroup_hash($_file_list);
      } else {
         print "Cannot create file list\n";
         exit(0);
      }
   } # if test

   if (1) {
      &process_ingroup_hash();
   }
   
   &process_file_hash();
   if (1) {
      &write_output_list("$gen_list");
   }
   if (1) {
      if ($xscope::sapi) {
         # This will generate the file only having doxygen templates
         # and no fake function prototypes.
         &write_output_class("$gen_template");
      } else {
         # This will generate a file with fake class and member 
         # function prototypes with associated doxygen templates.
         &write_output_class("$gen_class");
      }
   }
   if (1) {
      if ($xscope::sapi) {
         # The output file is not meaningful for SAPI, because
         # it has class member functions.
      } else {
         &write_output_file($gen_proto);
      }
      
   }
   
   ##############################################################################
   ## Spit out errors
   ##############################################################################
   if (@file_errors) {
      print "\n============  Summary of errors ==================================\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
   }

   ##############################################################################
   ## End of most of work
   ##############################################################################
   exit(0);
}

##############################################################################
##! @fn int create_xhelp_grep_results
##! @brief Creates a temporary file that contains the grep results.
##!
##! @param _file_list A temporary file created from the grep calls.
##!
##! @return The file is created in the directory given by the first input
##! parameter to the program.
##!
##! The xscope::x_names and xscope::needed_in arrays are used to trigger
##! the grep, which starts from the $root_path (input argument) directory.
##!
##! @bug This routine is really snow due to the number of system grep
##! calls it makes.
##!
##! @ingroup tp_xhelp
##############################################################################
sub create_xhelp_grep_results {
   local ($in_file_list) = $_[0];

   system ("echo \"\n\" > $in_file_list");

   if ((1) && (@xscope::x_names)) {
      # Greps for the prefix names over the supported file types.
      print "Grepping for prefix names in code...\n";
      foreach $xword (@xscope::x_names){
         foreach $f_type (@xscope::include_f_type) {
            if (0) {
               print "xword=$xword $f_type in $root_path\n";
            }
            print ("grep \"$xword\" \`find $root_path -name \"\*\.$f_type\" -print\`  >> $in_file_list\n");
            system ("grep \"$xword\" \`find $root_path -name \"\*\.$f_type\" -print\`  >> $in_file_list");
         } # foreach $f_type 
      } # foreach $xword
   } # if (0)
   if ((1) && (@xscope::needed_in)) {
      # Greps for explicit command names that are needed over the supported file types.
      print "Grepping for command names in code...\n";
      foreach $xword (@xscope::needed_in){
         foreach $f_type (@xscope::include_f_type) {
            if (0) {
               print "xword=$xword $f_type in $root_path\n";
            }
            print ("grep \"$xword\" \`find $root_path -name \"\*\.$f_type\" -print\`  >> $in_file_list\n");
            system ("grep \"$xword\" \`find $root_path -name \"\*\.$f_type\" -print\`  >> $in_file_list");
         } # foreach $f_type 
      } # foreach $xword
   } # if (0)
   
   return (0);
} # create_xhelp_grep_results



##############################################################################
##! @fn int create_xhelp_hash
##! @brief Creates a hash of hashes from the grep results that localize
##! target source files and code items to look for in those files.
##!
##! @param _file_list A temporary file created from the grep calls.
##!
##! @return The $file_hash, a hash of hashes, is created.
##! $file_hash {$src_file} {$code_item} {r} {$return_type} = unimportant number
##! $file_hash {$src_file} {$code_item} {p} {$prototype} = doxygen
##!
##! Although a simpler structure could be created, this has the advantage
##! that when a source file is opened, we know exactly what code items to
##! search for in that file.
##!
##! The {r} and {p} hash layer helps separate return type definitions from
##! prototype definitions. Moreover, the subsequent hashes for {$return_type}
##! and {$prototype} support overloading.
##!
##! This routine steps through the output file created by the grep calls.
##! It strips out uninteresting things that grep finds, like hits within
##! comments.
##!
##! @bug This routine fills in the {$return_type} hash, which gets used later
##! to help locate commands in code. The {$prototype} hashes get filled in
##! later.
##!
##! @ingroup tp_xhelp
##############################################################################
sub create_xhelp_hash {
   local ($in_file_list) = $_[0];
   print "Creating command hash from grep results...\n";

   unless (open ( IN_LIST, $in_file_list)) {
      print "Cannot open file \"$in_file_list\"\n";
      push (@file_errors, "Cannot open file \"$in_file_list\"\n");
      exit(1);
   }
   
   $_cnt=0;
   $_xcnt=0;
   X_INCOMING:   while (<IN_LIST>){    # read a line from file into $_
      if ((/^\n/) || (!/\S+/)){
         # get rid of blank lines
         next X_INCOMING;
      }
      if ($_cnt > 4) {
         exit(1);
      }
      @line_chunk = split (/\:/, $_, 2);
      if (0) {
         print "[0]==$line_chunk[0]\n";
         print "[1]==$line_chunk[1]\n";
         $_cnt++;
         # exit(1);
      }
      #####
      # Get rid of comment lines so they don't give us a spurious hit.
      ####
      if (($line_chunk[1] =~ /^\/\//) || ($line_chunk[1] =~ /\*\*/)) {
         # get rid of comment lines
         if (0) {
            print "trashing==$line_chunk[1]";
         }
         next X_INCOMING;
      }
      #####
      # Get rid of things known to be code specific and not prototypes
      # or definitions.
      # 6/10/2002 added /[\s]*\)[\s]*\:/ test
      ####
      if (($line_chunk[1] =~ /switch/) 
         || ($line_chunk[1] =~ /printf/) 
         || ($line_chunk[1] =~ /\;[\s]*$/) 
         || ($line_chunk[1] =~ /[\s]*\=[\s]*/) 
         || ($line_chunk[1] =~ /[\s]+if[\s]*\(/) 
         || ($line_chunk[1] =~ /^[\s]*return/) 
         || ($line_chunk[1] =~ /[\s]*\)[\s]*\:/) 
         || ($line_chunk[1] =~ /^\#/)) {
         # get rid of program lines
         if (0) {
            print "trashing==$line_chunk[1]";
         }
         next X_INCOMING;
      }
      @code_chunk = split (/\(/, $line_chunk[1], 2);
      if ($#code_chunk > 0) {
         foreach $exclude (@xscope::needed_out) {
            if (0) {
               print "exclude me: $exclude\n";
            }
            # Jump over any commands that we are told to specifically exclude.
            if ($code_chunk[0] =~ /$exclude/) {
               if (0) {
                  print "exclude me \"$exclude\" from $code_chunk[0]\n";
               }
               next X_INCOMING;
            }
         }
         if (0) {
            # print "$line_chunk[1]";
            # print "$_";
            print "nnn $code_chunk[0]\n";
         }
         if ($code_chunk[0] =~ /[\S+]/){
            @ode_name = split (/\s/, $code_chunk[0]);
            $_name = $ode_name[$#ode_name];
            if (0){
               print "ooo $_xcnt\:\:$code_chunk[0]>>> $_name\n";
            }
         } else {
            if (0){
               print "ooo2 $_xcnt\:\:$code_chunk[0]\n";
            }
         }
         #####
         # If we made it this far without doing a "next X_INCOMING" then
         # [1]  $line_chunk[0] contains the file of interest;
         # [2]  $code_chunk[0] contains the return type and item name.
         # [3]  $_name is the item name.
         # Let's put the code items into the file hash to speed up our process.
         #
         # Need to handle function overloading, so
         #
         # $file_hash {files} {code items} {r} {return types}
         #
         # We don't care about what it points to.
         #####
         $file_hash{$line_chunk[0]}{$_name}{r}{$code_chunk[0]} = $_xcnt;
         
         $_xcnt++;
      } # if ($#code_chunk > 0)
   } # while X_INCOMING


   ######
   # What is or will be in file hash
   # $file_hash{$src_file}{$code_item}{r}{$return_type} = number
   # $file_hash{$src_file}{$code_item}{p}{$prototype} = doxygen
   ######

   if (0){ # 
      foreach $src_file (sort keys %file_hash){
         foreach $code_item (sort keys %{$file_hash{$src_file}} ){
            foreach $return_type (sort keys %{$file_hash{$src_file}{$code_item}{r}} ) {
               # print "$code_item :: $return_type :: $src_file\n";
               print "$code_item\n";
            }
         }
      }
      if (0){
         exit(1);
      }
   } # if test
   return (1);

} # create_xhelp_hash


##############################################################################
##! @fn int process_file_hash
##! @brief Steps through each source file looking for the code items to
##! be extracted.
##!
##! @return Fleshes out the file hash with information about the prototype.
##!
##!  $file_hash {$src_file} {$code_item} {r} {$return_type} = number
##!  $file_hash {$src_file} {$code_item} {p} {$prototype} = doxygen
##!  
##! {$src_file}  the file of interest;
##! {$code_item}  the item name within the file of interest.
##! {r} flag for return types.
##! {r} {$return_type} hash of return types associated with code item, to
##! handle overloaded functions.
##! {p} flag for prototypes.
##! {p} {$prototype} hash of prototypes associated with code item and handles
##! overloaded functions. It points to associated doxygen comments.
##!
##! The routine opens each source file and reads it into $src_content. 
##! For each {$src_file} {$code_item}, split on the item's {$return_type}.
##! Determine whether it is a definition (ok) or prototype (ignore).
##! Extract the prototype definition.
##! Look for Doxygen comments right before split.
##! Divide and conquor; keep splitting the halves until there are no more
##! {$return_type} in.
##! Create a {p} {$prototype} hash entry that points to the doxygen comments.
##!
##! @bug When this reads in the source file, it changes //! style comments
##! into /** ... */ comment blocks. Also, straight comment lines beginning
##! with // are outright ignored.
##!
##! @ingroup tp_xhelp
##############################################################################
sub process_file_hash {
   local ($src_content) = "";
   
   if (0){ # 
      foreach $src_file (sort keys %file_hash){
         foreach $code_item (sort keys %{$file_hash{$src_file}} ){
            foreach $return_type (sort keys %{$file_hash{$src_file}{$code_item}{r}} ) {
               print "ppp $file_hash{$src_file}{$code_item}{r}{$return_type} :: $src_file + $code_item\n";
            }
         }
      }
      if (0){
         exit(1);
      }
   } # if test

   
   SRC_IN: foreach $src_file (sort keys %file_hash){
      # for each file that has a code item of interest
      unless (open ( IN_SRC, $src_file)) {
         print "Cannot open file \"$src_file\"\n";
         push (@file_errors, "Cannot open file \"$src_file\"\n");
         next SRC_IN;
      }
      if (0) {
         print "Source file is: $src_file\n";
      }

      undef ($src_content);  # start fresh
      # read file into memory
      while (<IN_SRC>) {
         ######### 
         # Read in the source file, but change Doxygen comment syntax
         # so that it is consistent throughout whole file.
         # Begin replacement of //! comments with /** ... **/ comments
         ######### 
         if (/^\/\/\!/) {
            if ($comment_count == 0){
               # first line of a comment block
               $comment_count++;
               $_ =~ s/\/\/\!/\/\*\*/;
            } else {
               # Some line in the middle of a comment block.
               $comment_count++;
               ###
               # Changed to have middle stuff with no asterix.
               # $_ =~ s/\/\/\!/ \*\*/;
               ###
               $_ =~ s/\/\/\!//;
            }
         } elsif ($comment_count > 0){
            # We were in a comment block; need to terminate it.
            $comment_count = 0;
            $_ = " \*\*\/\n" . $_ ;
         }
         ######### End Comment style change.
         
         #########
         # Trash all lines beginning with a comment.
         #########
         if (/^\/\//) {
            # Don't mess with comment lines.
         } elsif ((/^\/\*/) && (/\*\/$/)) {
            # Don't mess with lines that begin and end as a coment
         } else {
            # Remember everything else
            $src_content .= $_;
         }

      } # reading in source file
      
      # for each code item of interest within the file
      CODE_ITEM: foreach $code_item (sort keys %{$file_hash{$src_file}}) {
         OVERLOAD_ITEM: foreach $return_type (sort keys %{$file_hash{$src_file}{$code_item}{r}}) {
            # for each overloaded return code item
      
            $split_string = $return_type;
            
            if (0) { # 06/10/2002 attempt to filter split_string
              # Attempt to get rid of bogus characters that will screw with pattern
              # matching.
              if ($split_string =~ /\)/) {
                # if it has a closing ), then it has to
                # have an opening (, otherwise it is of no
                # interest and will mess with us later.
                if ($split_string !~ /\(/) {
                   next OVERLOAD_ITEM;
                }
              }
            } # attempt to filter split_string
         
            $search_chunk = $src_content;
            $prototype = "NEXT_ONE";
            $prot_cnt = 0;
         
            # Handle overloaded or multiple occurrences.
            SOME_ITEM: while ($search_chunk =~ /$split_string/) {
               @file_piece = split (/$split_string/, $search_chunk, 2);
  
               $prototype = &inside_parenthesis ($file_piece[1]);
               $search_chunk = $file_piece[1];
               if ($prototype =~ /NEXT_ONE/) {
                  # what we have is the prototype and not 
                  # the desired definition
               } else {
                  ######
                  # this is the desired definition
                  ######

                  $prototype = "$split_string$prototype";

                  #####
                  # 4/23/2002 Noticed that excluded items were creeping back in.
                  # This tests for them again and removes them.
                  #####
                  foreach $exclude (@xscope::needed_out) {
                     # Jump over any commands that we are told to specifically exclude.
                     if ($prototype =~ $exclude) {
                        if (0) {
                           print "==============\nExclude key \"$exclude\" of $prototype\n+++++++\n";
                        }
                        next SOME_ITEM;
                     }
                  }
                  
                  $prototype .= "\n\{\n   \/\/ Generated prototypes.\n\}\n";
                  if (0) {
                     print "The definition:\n$prototype=====\n";
                  }

                  ######
                  # Find the doxygen comments immediately prior to definition
                  # Note that file_piece[0] is not necessarily the first half of
                  # the file; it is from either the beginning of the file or
                  # the first prototype.
                  ######
                  $doxygen = &find_doxygen ($file_piece[0]);
                  if ($doxygen =~ /NO_DOX_COMMENT/) {
                     $doxygen = "\/\/ $doxygen\n";
                  } 
                  $file_hash{$src_file}{$code_item}{p}{$prototype} = $doxygen;

               } # NEXT_ONE
               
            } # while $search_chunk;  there is still something that looks interesting
 
            if (0) {
               # print "split=$split_string\n";
               print "prototype=\n$prototype\n";
               print "doxygen=\n$doxygen\n";
            }
         } # foreach OVERLOAD_ITEM 
      } # foreach CODE_ITEM 
      close (SRC_IN);
   } # foreach SRC_IN
} # process_file_hash

##############################################################################
##! @fn int create_ingroup_grep_results
##! @brief Creates a temporary file that contains the grep results.
##!
##! @param _file_list A temporary file created from the grep calls.
##!
##! @return The file is created in the directory given by the first input
##! parameter to the program.
##!
##! The xscope::x_names and xscope::needed_in arrays are used to trigger
##! the grep, which starts from the $root_path (input argument) directory.
##!
##! @bug This routine is really snow due to the number of system grep
##! calls it makes.
##!
##! @ingroup tp_xhelp
##############################################################################
sub create_ingroup_grep_results {
   local ($in_file_list) = $_[0];

   system ("echo \"\n\" > $in_file_list");

   if ((1) && (@xscope::ingroup)) {
      # Greps for the prefix names over the supported file types.
      print "Grepping for prefix names in code...\n";
      foreach $xword (@xscope::ingroup){
         foreach $f_type (@xscope::include_f_type) {
            if (0) {
               print "xword=$xword $f_type in $root_path\n";
            }
            print ("grep \"$xword\" \`find $root_path -name \"\*\.$f_type\" -print\`  >> $in_file_list\n");
            system ("grep \"$xword\" \`find $root_path -name \"\*\.$f_type\" -print\`  >> $in_file_list");
         } # foreach $f_type 
      } # foreach $xword
   } # if (0)
   return (0);
} # create_ingroup_grep_results


##############################################################################
##! @fn int create_ingroup_hash
##! @brief Creates a hash of hashes from the grep results that localize
##! target source files and code items to look for in those files.
##!
##! @param _file_list A temporary file created from the grep calls.
##!
##! @return The $file_hash, a hash of hashes, is created.
##! $file_hash {$src_file} {$code_item} {r} {$return_type} = unimportant number
##! $file_hash {$src_file} {$code_item} {p} {$prototype} = doxygen
##!
##! This looks for @ingroup places in the code comments.
##!
##! Although a simpler structure could be created, this has the advantage
##! that when a source file is opened, we know exactly what code items to
##! search for in that file.
##!
##! The {r} and {p} hash layer helps separate return type definitions from
##! prototype definitions. Moreover, the subsequent hashes for {$return_type}
##! and {$prototype} support overloading.
##!
##! This routine steps through the output file created by the grep calls.
##! It strips out uninteresting things that grep finds, like hits within
##! comments.
##!
##! @bug This routine fills in the {$return_type} hash, which gets used later
##! to help locate commands in code. The {$prototype} hashes get filled in
##! later.
##!
##! @ingroup tp_xhelp
##############################################################################
sub create_ingroup_hash {
   local ($in_file_list) = $_[0];
   print "Creating command hash from grep results...\n";

   unless (open ( IN_LIST, $in_file_list)) {
      print "Cannot open file \"$in_file_list\"\n";
      push (@file_errors, "Cannot open file \"$in_file_list\"\n");
      exit(1);
   }
   
   $_cnt=0;
   $_xcnt=0;
   X_INCOMING:   while (<IN_LIST>){    # read a line from file into $_
      if ((/^\n/) || (!/\S+/)){
         # get rid of blank lines
         next X_INCOMING;
      }
      if ($_cnt > 4) {
         exit(1);
      }
      @line_chunk = split (/\:/, $_, 2);
      if (0) {
         print "[0]==$line_chunk[0]\n";
         print "[1]==$line_chunk[1]\n";
         $_cnt++;
         # exit(1);
      }
      #####
      # Get rid of all non-comment lines so they don't give us a spurious hit.
      ####
      if (($line_chunk[1] =~ /^\/\//) || ($line_chunk[1] =~ /\*\*/)) {
         # These are the ones we want.
         if (0) {
            print "keep1==$line_chunk[1]";
         }
      } else {
         # get rid of non-comment lines
         if (0) {
            print "trashing1==$line_chunk[1]";
         }
         next X_INCOMING;
      }
      #####
      # Get rid of things known to be code specific and not comments
      # containing the @ingroup designation
      ####
      if ($line_chunk[1] =~ /\@ingroup/) {
         # These are the ones we want.
         if (0) {
            print "keep2==$line_chunk[1]";
         }
      } else {
         # get rid of program lines
         if (0) {
            print "trashing2==$line_chunk[1]";
         }
         next X_INCOMING;
      }
      # Remove things from the line of no interest
      $line_chunk[1] =~ s/\/\/\![\s+]\@ingroup[\s+]//;
      if (0) {
         print "keep3==$line_chunk[1]";
      }

      # Get the groups
      
      @group_chunk = split (/\s/, $line_chunk[1]);
      foreach $group (@group_chunk){
         #####
         # If we made it this far without doing a "next X_INCOMING" then
         # [1]  $line_chunk[0] contains the file of interest;
         # Let's put the code items into the file hash to speed up our process.
         #
         #
         # $ingroup_hash {$files} {$group}
         #####
         if ($group =~ /\*/) {
            # /** style doxgyen comments could mess this up, as in
            # ** @ingroup blah
            $group =~ s/\*//g;
         }
         if (0) {
            print "$_xcnt group==$group\n";
         }
         foreach $desired_group (@xscope::ingroup){
            if (0) {
               print "$_xcnt desired group==$desired_group\n";
            }
            if (($desired_group =~ /$group/i) && ($group =~ /$desired_group/i)){
               # double testing assures that I'm not triggering on part of the
               # group name.
               if (0) {
                  print "found==$group ;desired == $desired_group\n";
               }
               if (exists ($group_hash{$line_chunk[0]}{$group})) {
                  $group_hash{$line_chunk[0]}{$group} .= "\; $_xcnt";
               } else {
                  $group_hash{$line_chunk[0]}{$group} = "\; $_xcnt";
               }
               $_xcnt++;
            }
         }
      } # $group
   } # while X_INCOMING

   ######
   # What is or will be in group hash
   # $group_hash{$src_file}{$group} = number
   ######

   if (0){ # 
      foreach $src_file (sort keys %group_hash){
         foreach $group_item (sort keys %{$group_hash{$src_file}} ){
            print "$group_item :: $src_file :: $group_hash{$src_file}{$group_item}\n";
         }
      }
      if (0){
         exit(1);
      }
   } # if test
   return (1);

} # create_ingroup_hash



##############################################################################
##! @fn int process_ingroup_hash
##! @brief Steps through each source file looking for the group items to
##! be extracted.
##!
##! @return Adds things to the file hash for later processing.
##!
##!  $file_hash {$src_file} {$code_item} {r} {$return_type} = number
##!  
##! {$src_file}  the file of interest;
##! {$code_item}  the item name within the file of interest.
##! {r} flag for return types.
##! {r} {$return_type} hash of return types associated with code item, to
##! handle overloaded functions.
##!
##! The routine opens each source file and reads it into $src_content. 
##! For each {$src_file} {$group_item}, split on the item's @ingroup.
##!
##! Find the code definition and get it into the file_hash.
##!
##! @bug When this reads in the source file, it changes //! style comments
##! into /** ... */ comment blocks. Also, straight comment lines beginning
##! with // are outright ignored.
##!
##! @ingroup tp_xhelp
##############################################################################
sub process_ingroup_hash {
   local ($src_content) = "";
   
   if (0){ # 
      foreach $src_file (sort keys %group_hash){
         foreach $group_item (sort keys %{$group_hash{$src_file}} ){
            print "$group_item :: $src_file :: $group_hash{$src_file}{$group_item}\n";
         }
      }
      if (0){
         exit(1);
      }
   } # if test

   
   SRC_IN: foreach $src_file (sort keys %group_hash){
      # for each file that has a code item of interest
      unless (open ( IN_SRC, $src_file)) {
         print "Cannot open file \"$src_file\"\n";
         push (@file_errors, "Cannot open file \"$src_file\"\n");
         next SRC_IN;
      }
      if (0) {
         print "Source file is: $src_file\n";
      }
      undef ($src_content);  # start fresh
      # read file into memory
      while (<IN_SRC>) {
         ######### 
         # Read in the source file, but change Doxygen comment syntax
         # so that it is consistent throughout whole file.
         # Begin replacement of //! comments with /** ... **/ comments
         ######### 
         if (/^\/\/\!/) {
            if ($comment_count == 0){
               # first line of a comment block
               $comment_count++;
               $_ =~ s/\/\/\!/\/\*\*/;
            } else {
               # Some line in the middle of a comment block.
               $comment_count++;
               ###
               # Changed to have middle stuff with no asterix.
               # $_ =~ s/\/\/\!/ \*\*/;
               ###
               $_ =~ s/\/\/\!//;
            }
         } elsif ($comment_count > 0){
            # We were in a comment block; need to terminate it.
            $comment_count = 0;
            if (/^\/\//) {
               # Don't mess with comment lines.
               $_ = " \*\*\/\n" ;
            } else {
               # Not a comment lines.
               $_ = " \*\*\/\n" . $_ ;
            }
         }
         ######### End Comment style change.
         
         #########
         # Trash all lines beginning with a comment.
         #########
         if (/^\/\//) {
            # Don't mess with comment lines.
         } elsif ((/^\/\*/) && (/\*\/$/)) {
            # Don't mess with lines that begin and end as a comment
         } else {
            # Remember everything else
            $src_content .= $_;
         }

      } # reading in source file
      if (0){
         print $src_content;
         exit(1);
      }
      
      # for each @ingroup within the file, find the return value.
      
      if ($src_content =~ /\@ingroup/){
         @dox_ingroup1 = split (/\@ingroup/, $src_content);
         DOX_INCOMING: foreach ($i = 1; $i <= $#dox_ingroup1; $i++) {
            # purposely don't look at the first chunk, because it'll be
            # uninteresting. Hopefully there's not an off-by-one error.
            # Split out the rest of the comments.
            @dox_ingroup2 = split (/\*\//, $dox_ingroup1[$i], 2);
            # dox_ingroup2[0] is remainder of comment
            # dox_ingroup2[1] is code stuff
            $process_g_flag = 0;
            foreach $desired_group (@xscope::ingroup){
               if ($dox_ingroup2[0] =~ /$desired_group/) {
                  # This dox_group needs to be parsed for its return value
                  $process_g_flag = 1;
               }
            } # $desired_group
            
            if ($process_g_flag) {
               # Get the return value and code item name from the 
               # code chunk and build up the file_hash.
               # dox_ingroup2[1] is code stuff
               @code_chunk1 = split (/\{/, $dox_ingroup2[1], 2);
               # $code_chunk1[0] is without implementation
               @code_chunk2 = split (/\(/, $code_chunk1[0], 2);
               if ($#code_chunk2 > 0) {
                  if (1) {
                     #####
                     # 4/23/2002 Notice that the excluded elements were creeping back
                     # in when the actual source code file was processed.
                     # Seeing if we have to exclude them later,
                     # figured might as well not test here to save some time.
                     #####
                     foreach $exclude (@xscope::needed_out) {
                        # Jump over any commands that we are told to specifically exclude.
                        if ($code_chunk2[0] =~ $exclude) {
                           next DOX_INCOMING;
                        }
                     }
                  }
                  # Get rid of any leading carriage returns and slashes
                  $code_chunk2[0] =~ s/^\s*\n+//;
                  if (0) {
                     print "nnn $code_chunk2[0]\n";
                  }
                  if ($code_chunk2[0] =~ /[\S+]/){
                     @ode_name = split (/\s/, $code_chunk2[0]);
                     $_name = $ode_name[$#ode_name];
                     $_name =~ s/^\*//;
                     if (0){
                        print "ooo $_xcnt\:\:$code_chunk2[0]>>> $_name\n";
                     }
                  } else {
                     if (0){
                        print "ooo2 $_xcnt\:\:$code_chunk2[0]\n";
                     }
                  }
                  #####
                  # If we made it this far without doing a "next D_INCOMING" then
                  # [1]  $src_file contains the file of interest;
                  # [2]  $code_chunk2[0] contains the return type and item name.
                  # [3]  $_name is the item name.
                  # Let's put the code items into the file hash to speed up our process.
                  #
                  # Need to handle function overloading, so
                  #
                  # $file_hash {files} {code items} {r} {return types}
                  #
                  # We don't care about what it points to.
                  #####
                  if (exists ($file_hash{$src_file}{$_name}{r}{$code_chunk2[0]})) {
                     $file_hash{$src_file}{$_name}{r}{$code_chunk2[0]} .= "; gp $_xcnt";
                  } else {
                     $file_hash{$src_file}{$_name}{r}{$code_chunk2[0]} = "gp $_xcnt";
                  }
         
                  $_xcnt++;
               } # ($#code_chunk2 > 0)
            } # DOX_INCOMING
         } # $dox_chunk
      } # ($src_content =~ /\@ingroup/)

      close (SRC_IN);
   } # foreach SRC_IN

   ######
   # What is or will be in file hash
   # $file_hash{$src_file}{$code_item}{r}{$return_type} = number
   ######

   if (0){ # 
      $j=0;
      foreach $src_file (sort keys %file_hash){
         foreach $code_item (sort keys %{$file_hash{$src_file}} ){
            foreach $return_type (sort keys %{$file_hash{$src_file}{$code_item}{r}} ) {
               $j++;
               # print "$j $code_item :: $return_type :: $src_file\n";
               print "$j $code_item :: $return_type\n";
            }
         }
      }
      if (1){
         print "stopping here...";
         exit(1);
      }
   } # if test
      if (0){
         print "stopping here...";
         exit(1);
      }

} # process_ingroup_hash


##############################################################################
##! @fn int inside_parenthesis
##! @brief Grabs what's inside the first set of parenthesis and takes into
##! consideration nested parenthesis.
##!
##! @param in_text has a piece of text that has a set of parenthesis 
##! near its beginning.
##!
##! @return If the first non-white space character after the top-level 
##! parenthesis is an open curly brace ({), then the parenthesis are associated
##! with a definition; return everything inside the first set of parenthesis 
##! including the parenthesis. 
##!
##! This handles nested parenthesis within the first set of parentheis.
##!
##! @bug This forces us to determine whether it was a prototype or definition,
##! rather than being generic for returning whatever was in the first set of
##! parenthesis.
##!
##! @ingroup tp_xhelp
##############################################################################
sub inside_parenthesis {
   local ($in_text) = $_[0];
   undef (@in_array);
   undef (@out_array);
   local ($nest_cnt) = 0;
   local ($have_it) = 0;
   
   
   @in_array = split (//, $in_text);
   if (0) {
      # print "GCMin:$#in_array:$in_text\n";
      print "begin:$begin\n";
      print "end:$end\n======\n";
      for ($i=0; $i<200; $i++){
         print "$in_array[$i]";
      }
      print "\nhey======\n";
      # exit(1);
   }
   NEXT_CHAR: foreach $cchar (@in_array) {
      # put together an array that contains the prototype
      # and handles nested parenthesis with the prototype.
      if ($have_it) {
         # If this is set, we have the parenthesis,
         # but keep checking it until we have either a semicolon (;)
         # or an opening curly brace ({);
         if ($cchar =~ /\;/) {
            # If a semicolon follows what was in parenthesis,
            # then we've snagged the prototype which is not of interest.
            # get out of here
            if (0) {
               print "Z1no$nest_cnt:$cchar\n";
            }
            return ("NEXT_ONE");
         } elsif ($cchar =~ /\{/) {
            # If an open curly brace follows what was in parenthesis,
            # then we've snagged the real code item that is of interest.
            if (0) {
               print "Z2yes$nest_cnt:$cchar\n";
            }
            goto REALLY_HAVE;
         } else {
            if (0) {
               print "Z3notyet$nest_cnt:$cchar\n";
            }
            next NEXT_CHAR;
         }
      } elsif (($nest_cnt == 0) && ($cchar =~ /\(/)) {
         if (0) {
            print "A$nest_cnt:$cchar\n";
         }
         push (@out_array, $cchar);
         $nest_cnt++;
         next NEXT_CHAR;
      } elsif (($nest_cnt != 0) && ($cchar =~ /\(/)) {
         if (0) {
            print "B$nest_cnt:$cchar\n";
         }
         push (@out_array, $cchar);
         $nest_cnt++;
         next NEXT_CHAR;
      } elsif (($nest_cnt != 0) && ($cchar =~ /\)/)) {
         push (@out_array, $cchar);
         $nest_cnt--;
         if (0) {
            print "C$nest_cnt:$cchar\n";
         }
         if ($nest_cnt == 0){
            # Mark that we really have it;
            $have_it = 1;
         } 
         next NEXT_CHAR;
         
      } elsif (($nest_cnt == 0) && ($cchar !~ /\(/)) {
         if (0) {
            print "D$nest_cnt:$cchar\n";
         }
         push (@out_array, $cchar);
         next NEXT_CHAR;
      } elsif ($nest_cnt != 0) {
         if (0) {
            print "E$nest_cnt:$cchar\n";
         }
         push (@out_array, $cchar);
         next NEXT_CHAR;
      } # nest_cnt loop
   } # foreach $cchar
   
   REALLY_HAVE:
   $out = join ("", @out_array);
   undef (@in_array);
   undef (@out_array);
   if (0) {
      print "what's coming out=\n$out\n======\n";
      # exit(1);
   }
   return ($out);
} # inside_parenthesis


##############################################################################
##! @fn int find_doxygen
##! @brief Looks for doxygen comments in test passed in.
##!
##! @param piece_to_search a chunk of text that may have zero or more
##! doxygen comment blocks.
##!
##! @return The last doxygen comment block if it exists and doesn't have
##! code in between the block and the very end. If there is code inbetween
##! the block and the end or if there are no doxygen comment blocks, it
##! returns NO_DOX_COMMENT.
##!
##! Code between the comment block is spotted by open parenthesis "(" or
##! open curly braces "{" or semicolons ";". 
##! It means the comment block found doesn't belong with the code item in
##! question, otherwise it'd be the last thing.
##! 
##! @bug This assumes the the Doxgyen comments have been flagged with 
##! a leading /**.
##!
##! @ingroup tp_xhelp
##############################################################################
sub find_doxygen {
   local ($piece_to_search) = $_[0];
   local ($dox_flag_b) = "\/\*\*";
   local ($dox_flag_e) = "\*\/";
   local ($out) = "";
   
   if ($piece_to_search =~ /\/\*\*/) {
      # Doxygen in piece that could potentially be processed
      @dox_pieces = split ( /\/\*\*/, $piece_to_search);
      if (0) {
         print "Maybe there is some doxygen (\/\*\*) ($#dox_pieces) to process.\n";
         print "\/\*\*$dox_pieces[$#dox_pieces]";
      }
      # reduce to just doxygen stuff
      @dox_pieces2 = split (/\*\//, $dox_pieces[$#dox_pieces], 2);
      
      # ERROR CHECKING
      # The leftover piece $dox_pieces2[1] is between the doxygen comment
      # and the code definition.
      # If there are parenthesis "(" or curly braces in the left over piece,
      # then the doxygen comment does not belong with this code item.
      if (($dox_pieces2[1] =~ /\(/) || ($dox_pieces2[1] =~ /\{/) || ($dox_pieces2[1] =~ /\;/)) {
         $out = "NO_DOX_COMMENT";
      } else {
         $out = "\/\*\*";
         $out .= $dox_pieces2[0];
         $out .= "\*\/";
      }
   } else {
      if (0) {
         print "No doxygen to process.\n";
      }
      $out = "NO_DOX_COMMENT";
   }
   return ($out);
   
} # find_doxygen


##############################################################################
##! @fn int write_output_file
##! @brief Creates a file that has a doxygen comment block, code definition,
##! and empty body for each item of interest from the file hash.
##!
##! @param file_out the path and filename of the output file.
##!
##! @return The generated output file.
##!
##! This routine steps through the file hash and writes to the file_out
##! all relevant information, in addition to a header and white space.
##!
##! The file hash has already been established:
##!
##!  $file_hash {$src_file} {$code_item} {p} {$prototype} = doxygen
##!  
##! {$src_file}  the file of interest;
##! {$code_item}  the item name within the file of interest.
##! {p} flag for prototypes.
##! {p} {$prototype} hash of prototypes associated with code item and handles
##! overloaded functions. It points to associated doxygen comments.
##!
##! @ingroup tp_xhelp
##############################################################################
sub write_output_file {
   $file_out = $_[0];
   undef (%s_code_item);   # sorted hash for code item
   print "Writing the $file_out output file.\n";
   
   unless (open ( OUT_FILE, ">$file_out")) {
      push (@file_errors, "Cannot open file \"$file_out\"\n");
      print "Cannot open file \"$file_out\"\n";
   }
   print (OUT_FILE "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n");
   if (1) {
      # Non-doxygen disclaimer at top of file.
      print (OUT_FILE "\/\/ Generated file with the appropriate commands.\n");
      print (OUT_FILE "\/\/ \n\/\/ Contains blank function declaration that\n");
      print (OUT_FILE "\/\/ are preceded by their appropriate Doxygen comments\n");
      print (OUT_FILE "\/\/ if they exist.\n");
      print (OUT_FILE "\/\/ \n\/\/ Its purpose is to reduce what Doxygen looks at\n");
      print (OUT_FILE "\/\/ to get rid of the noise in the output.\n");
      print (OUT_FILE "\/\/ \n\/\/ THIS FILE IS GENERATED!!!\n");
   } else {
      print (OUT_FILE "\/\/\! \@file\n");
      print (OUT_FILE "\/\/\! \@brief Generated file with the appropriate commands.\n");
      print (OUT_FILE "\/\/\! \n\/\/\! Contains blank function declaration that\n");
      print (OUT_FILE "\/\/\! are preceded by their appropriate Doxygen comments\n");
      print (OUT_FILE "\/\/\! if they exist.\n");
      print (OUT_FILE "\/\/\! \n\/\/\! Its purpose is to reduce what Doxygen looks at\n");
      print (OUT_FILE "\/\/\! to get rid of the noise in the output.\n");
      print (OUT_FILE "\/\/\! \n\/\/\! THIS FILE IS GENERATED!!!\n");
      # print (OUT_FILE "\/\/\! \n\/\/\! \@ingroup xhelp\n");
   }
   print (OUT_FILE "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n");
   print (OUT_FILE "\n\n\n");


if (1) { # new way of doing it so that it gets sorted
   foreach $src_file (sort keys %file_hash){
      if (0) {
         print "$src_file is source file\n";
      }
      foreach $code_item (sort keys %{$file_hash{$src_file}} ){
         if (0) {
            print "$code_item is code item\n";
         }
         PPROTO: foreach $prototype (sort keys %{$file_hash{$src_file}{$code_item}{p}} ){
            if (0) {
               print "$prototype\n";
            }
            $process_g_flag = 1;
            #####
            # Exclude certain doxygen comment blocks;
            # In particular, if the block is the file header, it may
            # be of no interest.
            #####
            foreach $undesired_tag (@xscope::dox_exclude){
               if ($file_hash{$src_file}{$code_item}{p}{$prototype} =~ /$undesired_tag/) {
                  # This item does not belong in the output
                  if (1) {
                     print"\"$code_item\" not in output because of \"$undesired_tag\"\n";
                  }
                  $process_g_flag = 0;
               }
            } # $undesired_tag

            if ($process_g_flag) {
               if ($xscope::sapi) {    # This is set in the generated sapi.pm
                  $command_name = &get_sapi_name ($file_hash{$src_file}{$code_item}{p}{$prototype});
                  if (0) {
                     print "debug:> $code_item\n";
                     # print "before the call with $src_file, $code_item, $prototype\n";
                     # print "$file_hash{$src_file}{$code_item}{p}{$prototype}\n";
                     print "c_name:> $command_name\n";
                  }
                  if ($command_name =~ /_NONE_/) {
                     # if we don't have a command name, no sense putting anything out.
                     next PPROTO;
                  }
               } else {  # for non-SAPI commands, like xhelp
                  $command_name = $code_item;
               }
               if (0) {
                  print "debug:> $code_item\n";
                  print "c_name:> $command_name\n";
               }
               if (exists ($s_code_item{$code_item})) {
                  $temp_string = "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n";
                  $temp_string .= "$file_hash{$src_file}{$code_item}{p}{$prototype}\n";
                  $temp_string .= "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n";
                  $temp_string .= "$prototype\n\n";
                  # if it exists, append this information to it.
                  $s_code_item{$command_name} .= $temp_string;
               } else {
                  $temp_string = "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n";
                  $temp_string .= "$file_hash{$src_file}{$code_item}{p}{$prototype}\n";
                  $temp_string .= "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n";
                  $temp_string .= "$prototype\n\n";
                  # if it doesn't exists, write this information to it.
                  $s_code_item{$command_name} = $temp_string;
               }
            }  # $process_g_flag
         }
      }
   }
   if (0) {
      print "\n\nHi Glenn\n=====\n";
      foreach $command_name (sort keys %s_code_item){
         print "$command_name\n";
         # print "$s_code_item{$command_name}";
      }
   }
  
   foreach $command_name (sort keys %s_code_item){
      print (OUT_FILE "$s_code_item{$command_name}");
   }
   
   undef (%s_code_item);
} # if (1) # new way of doing it so that it gets sorted




if (0) { # old way of doing it
   foreach $src_file (sort keys %file_hash){
      foreach $code_item (sort keys %{$file_hash{$src_file}} ){
         foreach $prototype (sort keys %{$file_hash{$src_file}{$code_item}{p}} ){
            $process_g_flag = 1;
            #####
            # Exclude certain doxygen comment blocks;
            # In particular, if the block is the file header, it may
            # be of no interest.
            #####
            foreach $undesired_tag (@xscope::dox_exclude){
               if ($file_hash{$src_file}{$code_item}{p}{$prototype} =~ /$undesired_tag/) {
                  # This item does not belong in the output
                  $process_g_flag = 0;
               }
            } # $undesired_tag

            if ($process_g_flag) {
               print (OUT_FILE "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n");
               print (OUT_FILE "$file_hash{$src_file}{$code_item}{p}{$prototype}\n");
               print (OUT_FILE "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n");
               print (OUT_FILE "$prototype\n\n");
            }
         }
      }
   }
} # if (0) { # old way of doing it, no sorting done
   close (OUT_FILE);
} # write_output_file

##############################################################################
##! @fn int get_sapi_name
##! @brief parses out the name of the sapi or bapi function from the doxygen
##! comments.
##! 
##! @param $src_file
##! @param $code_item
##!
##! @return The name of the command (with arguments) or _NONE_.
##!
##############################################################################
sub get_sapi_name  {
   local ($all_doc) = $_[0];
   local (@all_line);
   if (0) {
      print "you are here with:) \n$all_doc\n";
   }
   
   @all_line = split (/\n/, $all_doc);
   
   DOCLINE: foreach $line (@all_line) {
     if ($line =~ /\@fn/){
        @fn_chunk = split (/\@fn[\s]*/, $line, 2);
        @rtn_chunk = split (/[\s]+/, $fn_chunk[1]);
        if (0) {
           print "returning with \"$rtn_chunk[1]\"\n";
        }
        return ($rtn_chunk[1]);  # this contains the sapi function name
        break;
     }
   } # foreach $line
   if (0) {
     print "returning with \"_NONE_\"\n";
   }
   return ("_NONE_");
   
} # get_sapi_name

##############################################################################
##! @fn int write_output_list
##! @brief Creates a file that has a list of commands found, their location
##! found in the source code, and an indication of existing doxygen comments.
##!
##! @param file_out the path and filename of the output file.
##!
##! @return The generated output file.
##!
##! This routine steps through the file hash and writes to the file_out
##! all relevant information, in addition to a header and white space.
##!
##! The file hash has already been established:
##!
##!  $file_hash {$src_file} {$code_item} {p} {$prototype} = doxygen
##!  
##! {$src_file}  the file of interest;
##! {$code_item}  the item name within the file of interest.
##! {p} flag for prototypes.
##! {p} {$prototype} hash of prototypes associated with code item and handles
##! overloaded functions. It points to associated doxygen comments.
##!
##! @ingroup tp_xhelp
##############################################################################
sub write_output_list {
   $file_out = $_[0];
   print "Writing the $file_out output file.\n";
   
   unless (open ( OUT_FILE, ">$file_out")) {
      push (@file_errors, "Cannot open file \"$file_out\"\n");
      print "Cannot open file \"$file_out\"\n";
   }
   print (OUT_FILE "This file contains a list of the commands\n");
   print (OUT_FILE "and their location.\n");
   print (OUT_FILE "The letter in parenthesis refers to existing Doxygen comments.\n");
   print (OUT_FILE "(n) has no comment block.\n");
   print (OUT_FILE "(y) has a comment block filled in.\n");
   print (OUT_FILE "(_TBD_) has a comment block still needing information.\n");
   print (OUT_FILE "==============================================================\n\n");

   foreach $src_file (sort keys %file_hash){
      foreach $code_item (sort keys %{$file_hash{$src_file}} ){
         foreach $prototype (sort keys %{$file_hash{$src_file}{$code_item}{p}} ){
            if ($file_hash{$src_file}{$code_item}{p}{$prototype} =~ /NO_DOX_COMMENT/) {
               print (OUT_FILE "\(n\) $code_item :: $src_file\n");
            } elsif ($file_hash{$src_file}{$code_item}{p}{$prototype} =~ /_TBD_/) {
               print (OUT_FILE "\(_TBD_\) $code_item :: $src_file\n");
            } else {
               print (OUT_FILE "\(y\) $code_item :: $src_file\n");
            }
         }
      }
   }
   close (OUT_FILE);
   
   if (0) {
      foreach $src_file (sort keys %file_hash){
         foreach $code_item (sort keys %{$file_hash{$src_file}} ){
            print "$code_item\n";
         }
      }
      print "hola glenn\n";
      exit(1);
   }
   if (0) {
      foreach $src_file (sort keys %file_hash){
         foreach $code_item (sort keys %{$file_hash{$src_file}} ){
            foreach $prototype (sort keys %{$file_hash{$src_file}{$code_item}{p}} ){
               print "$prototype";
            }
         }
      }
      print "hola glenn\n";
      exit(1);
   }
} # write_output_list

##############################################################################
##! @fn int write_output_class
##! @brief Creates a file that has a list of commands found, their location
##! found in the source code, and an indication of existing doxygen comments.
##!
##! @param file_out the path and filename of the output file.
##!
##! @return The generated output file.
##!
##!
##! @ingroup tp_xhelp
##############################################################################
sub write_output_class {
   $file_out = $_[0];
   print "Writing the $file_out output file.\n";
   
   unless (open ( OUT_FILE, ">$file_out")) {
      push (@file_errors, "Cannot open file \"$file_out\"\n");
      print "Cannot open file \"$file_out\"\n";
   }

   print (OUT_FILE "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n");
   if (1) {
      print (OUT_FILE "\/\/ brief Generated file with the appropriate commands.\n");
      print (OUT_FILE "\/\/ \n\/\/ Contains blank function declaration that\n");
      print (OUT_FILE "\/\/ are preceded by their appropriate Doxygen comments\n");
      print (OUT_FILE "\/\/ if they exist.\n");
      print (OUT_FILE "\/\/ \n\/\/ Its purpose is to reduce what Doxygen looks at\n");
      print (OUT_FILE "\/\/ to get rid of the noise in the output.\n");
      print (OUT_FILE "\/\/ \n\/\/ THIS FILE IS GENERATED!!!\n");
   } else {
      print (OUT_FILE "\/\/\! \@file\n");
      print (OUT_FILE "\/\/\! \@brief Generated file with the appropriate commands.\n");
      print (OUT_FILE "\/\/\! \n\/\/\! Contains blank function declaration that\n");
      print (OUT_FILE "\/\/\! are preceded by their appropriate Doxygen comments\n");
      print (OUT_FILE "\/\/\! if they exist.\n");
      print (OUT_FILE "\/\/\! \n\/\/\! Its purpose is to reduce what Doxygen looks at\n");
      print (OUT_FILE "\/\/\! to get rid of the noise in the output.\n");
      print (OUT_FILE "\/\/\! \n\/\/\! THIS FILE IS GENERATED!!!\n");
      # print (OUT_FILE "\/\/\! \n\/\/\! \@ingroup xhelp\n");
   }
   print (OUT_FILE "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n");
   print (OUT_FILE "\n\n\n");
   
   #####
   # Sort the output by actual code item
   # We just need keys into the other structure; we don't need all of the data
   # of the other structure.
   #####
   undef (%list_of_items);
   if ($xscope::sapi) {  # This is set in the generated sapi.pm
      # Then the list of $xscope::sapi_names is available
      # We still need to step through and match things up from the doxygen comment
      foreach $src_file (sort keys %file_hash){
         S_CODE: foreach $code_item (sort keys %{$file_hash{$src_file}} ){
            S_PROTO: foreach $prototype (sort keys %{$file_hash{$src_file}{$code_item}{p}} ){
               @sm_dox = split (/\n/, $file_hash{$src_file}{$code_item}{p}{$prototype});
               for ($i=0; $i <= $#sm_dox; $i++) {
                  if ($sm_dox[$i] =~ /\@fn/) {
                     $match_line = $sm_dox[$i];
                     if (0) {
                        print "match:$match_line\n";
                     }
                     break;   # get out of this inner most loop
                  }
               }
               S_NAME: foreach $names (@xscope::sapi_names){
                  ####
                  # This is intended to match our sapi.pm generated list of names
                  # with prototypes that are near the beginning of the doxygen comment.
                  ####
                  # if ($file_hash{$src_file}{$code_item}{p}{$prototype} =~ /$names[\s]*\(/ ) {
                  if ($match_line =~ /$names[\s]*\(/ ) {
                     # something like "@fn int bogus_name ( blah, blah)" 
                     push ( @{$list_of_items{$names}{src}}, $src_file);
                     push ( @{$list_of_items{$names}{code}}, $code_item);
                     push ( @{$list_of_items{$names}{proto}}, $prototype);
                     next S_PROTO;
                  }
               }
            }
         }
      }
      
   } else { # if ($xscope::sapi)
      # Then the $file_hash{$src_file}{$code_items} are the names we need
      foreach $src_file (sort keys %file_hash){
         foreach $code_item (sort keys %{$file_hash{$src_file}} ){
            foreach $prototype (sort keys %{$file_hash{$src_file}{$code_item}{p}} ){
               push ( @{$list_of_items{$code_item}{src}}, $src_file);
               # duplicates it, but allows for a parallel structure as above
               # which gives us the needed keys into the original structure
               push ( @{$list_of_items{$code_item}{code}}, $code_item);
               push ( @{$list_of_items{$code_item}{proto}}, $prototype);
            }
         }
      }
   } # if ($xscope::sapi)
   
   if (0) { # Debug
      foreach $item (sort keys %list_of_items) {
         for ($i=0; $i <= $#{$list_of_items{$item}{src}}; $i++) {
            # print "$i\=\=$item\t\=\=$list_of_items{$item}{src}[$i]\n";
            print "$i\=\=$item\t\=\=$list_of_items{$item}{code}[$i]\n";
         }
      }
      if (0) {
         print "hola glenn\n";
         exit(1);
      }
   } 
   

   #####
   # The problem this solves is that doxygen's output is organized
   # in the order that it receives it.
   # I end up sorting all temporary files before running
   # them through doxygen. Then my HTML will be organized.
   #####
   
   #####
   # We remove the following, which is the old access method and unsorted.
   # It is replaced with what is sorted.
   # We need to match braces.
   #####
   # foreach $src_file (sort keys %file_hash){
   #    foreach $code_item (sort keys %{$file_hash{$src_file}} ){
   #       foreach $prototype (sort keys %{$file_hash{$src_file}{$code_item}{p}} ){
   #####
   foreach $item (sort keys %list_of_items) {
      for ($i=0; $i <= $#{$list_of_items{$item}{src}}; $i++) {
         $src_file = $list_of_items{$item}{src}[$i];
         $code_item = $list_of_items{$item}{code}[$i];
         $prototype = $list_of_items{$item}{proto}[$i];
            
         { # $prototype matching parenthesis
   #####
   # Replacement loop above
   #####

            $process_g_flag = 1;
            #####
            # Exclude certain doxygen comment blocks;
            # In particular, if the block is the file header, it may
            # be of no interest.
            #####
            foreach $undesired_tag (@xscope::dox_exclude) {
               if ($file_hash{$src_file}{$code_item}{p}{$prototype} =~ /$undesired_tag/) {
                  # This item does not belong in the output
                  $process_g_flag = 0;
               }
            } # $undesired_tag

            if ($process_g_flag) {
               # if ($prototype =~ /\"\)\;\}/){
               if ($prototype =~ /\"/){
                  # The construct above is giving us problems in the output,
                  # so let's not put it in the output.
                  delete ($file_hash{$src_file}{$code_item}{p}{$prototype});
               } elsif ($prototype =~ /\:\:/){
                  @sm_proto = split (/\{/, $prototype);
                  $sm_proto[0] =~ s/\n//g;
               
                  @class_ch1 = split (/\:\:/, $prototype);
                  @class_ch2 = split (/[\s+]/, @class_ch1[0]);
                  if (exists ($class_hash{$class_ch2[$#class_ch2]})) {
                     # $class_hash{$class_ch2[$#class_ch2]} .= "      virtual $sm_proto[0]\;\n";
                     # $class_hash{$class_ch2[$#class_ch2]} .= "      $sm_proto[0]\;\n";
                  } else {
                     # $class_hash{$class_ch2[$#class_ch2]} = "      virtual $sm_proto[0]\;\n";
                     # $class_hash{$class_ch2[$#class_ch2]} = "      $sm_proto[0]\;\n";
                     $class_hash{$class_ch2[$#class_ch2]} = "   \/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n\n";
                  }
                  #####
                  # code prototype is $prototype while
                  # doxygen template is in $file_hash{$src_file}{$code_item}{p}{$prototype}
                  #####
                  if (!($xscope::sapi)) {  # This is set in the generated sapi.pm
                     # Means we're working on something that wants to have
                     # the class definitions come through.
                     # We need the class definition.
                     $class_hash{$class_ch2[$#class_ch2]} .= "   \/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n";
                     $class_hash{$class_ch2[$#class_ch2]} .= "$file_hash{$src_file}{$code_item}{p}{$prototype}\n";
                     $class_hash{$class_ch2[$#class_ch2]} .= "   \/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n";
                     $class_hash{$class_ch2[$#class_ch2]} .= "      $sm_proto[0]\;\n";
                  } else { # $xscope::sapi
                     # when working on sapi, we don't need to be so anal about
                     # creating class stuff; 
                     # we just need the doxygen comment output so that
                     # the @fn can map to the sapi_gen prototypes later.
                     if ($file_hash{$src_file}{$code_item}{p}{$prototype} =~ /\@fn/) {
                        # only output a sapi @fn doxygen template;
                        print (OUT_FILE  "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n");
                        print (OUT_FILE  "$file_hash{$src_file}{$code_item}{p}{$prototype}\n");
                     }
                  } # $xscope::sapi
               }
            } # if process flag
         } # $prototype matching parenthesis
      } # $i
   } # $item

   if (!($xscope::sapi)) {  # This is set in the generated sapi.pm
      # Means we're working on something that wants to have
      # the class definitions come through.
      # We need the class definition.
      foreach $class (sort keys %class_hash){
         print (OUT_FILE "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n");
         print (OUT_FILE "\/\/ begin $class\n");
         print (OUT_FILE "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n");
         print (OUT_FILE "class $class\n");
         print (OUT_FILE "{\n");
         print (OUT_FILE "   public\:\n");
         print (OUT_FILE "$class_hash{$class}\n");
         print (OUT_FILE "}\;\n");
         print (OUT_FILE "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n");
         print (OUT_FILE "\/\/ end $class\n");
         print (OUT_FILE "\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\n");
         print (OUT_FILE "\n\n\n\n");
      }
   }
   close (OUT_FILE);
} # write_output_class


##############################################################################
##! @fn int using_xhlp_bapi
##! @brief What to do when no arguments are given.
##! @param None
##! @return None
##! 
##! @bug None
##! @ingroup tp_xhelp
##############################################################################
sub using_xhelp {
   print "============  Syntax  ==================================\n";
   print "xhelp_gen.pl <input scope package> <root path to files> <output file path \& name> \n\n";
   print "- <input scope package> is file path and name to a perl package that ";
   print "has lists of functions to be included and to be excluded.\n\n";
   print "- <root path to files> is the code location to start looking ";
   print "and must be terminated with a forward (\/) slash. ";
   print "If output file does not have a forward slash (\/) -- an indication ";
   print "of a path --, then the <root path to files> is assumed.\n\n";
   print "- <output file path \& name> plus another file ";
   print "with a leading underscore are generated. ";
   print "The file without the underscore is intended as the input to Doxygen.\n\n";
   print "============  Terminate  ==================================\n";
   print "Terminating xhelp_gen.pl without doing anything.\n";
   return(1);
}


##############################################################################
##! @fn int END
##! @brief Code to execute when terminated nicely.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_xhelp
##############################################################################
sub END {
   undef ($_file_list); # "_file_list";
   undef ($in_file ); #  "";
   undef ($root_path ); #  "./";
   undef (%file_hash);

   
   ##############################################################################
   ## Memory clean-up.
   ## Perform this if it was defined.
   ##############################################################################
   if ($no_scope_file > 0){
      &xscope::memory_clean_up();
   }
   
   print "\n============  Finished xhelp_gen.pl ==================================\n";
} # END
