#!/usr/local/bin/perl
package gen_nav;

##############################################################################
##! @file
##! @brief Looks into HTML files that have been provided by other sources
##! and parses them for useful information.
##! 
##! It extracts appropriate information needed for a entries in an index
##! and table of contents.
##!
##! It starts at the files listed in @xscope::top_files array. For every
##! HTML file there, it finds their hyperlinks. It successively traces through
##! the hyperlinks creating a data structure. This is the data structure used
##! in the TOC. The hyperlinks that it traces are also used directly in the 
##! table of contents.
##!
##! Certain files can be excluded from creating children or tracing further.
##! All top files in particular are added to this list so that it does not
##! loop continuously through things it already knows, or more realistically,
##! so that it doesn't build a top-levels data structure under some other
##! top-level data.
##!
##! This has several routines (spider_trace, index_token_generation, script_generation
##! etc.) that call themselves recursively. The key to stop the recursion is when
##! the owning files for the children (hyperlinks) have already been visited.
##!
##! @ingroup tp_tools tp_nav
##!
##! @author Glenn C. Maxey
##!
##    $Id: html_look_integrate.pl,v 1.5 2002/11/20 15:23:02 gmaxe Exp $
##
## 2002 Created by Voyant Technologies, Inc., Westminster, Colorado, USA.
##
## Permission to use, copy, modify, and distribute this software and its 
## documentation under the terms of the GNU General Public License is hereby 
## granted. No representations are made about the suitability of this software 
## for any purpose. It is provided "as is" without express or implied warranty. 
## See the GNU General Public License (http://www.gnu.org/copyleft/gpl.html) 
## for more details.
## 
## Documents produced by this script are derivative works derived from the 
## input used in their production; they are not affected by this license.
##
##    Revision Information:
##
##    $Log: html_look_integrate.pl,v $
##    Revision 1.5  2002/11/20 15:23:02  gmaxe
##    Added exit codes so that wrapper scripts can catch errors properly.
##
##    Revision 1.4  2002/07/26 18:56:10  gmaxe
##    Got rid of old definitions and migrated new structures into all;
##    enhanced xhelp output file names and format; now everything is
##    alphebetized.
##
##    Revision 1.3  2002/04/12 18:01:13  gmaxe
##    Various tweaks; work on tag extraction and change_nav;
##    more exclused special perl characters from indexer;
##    improved support for sapi group pm files in xhelp.
##    spider tool now works regularly and as expected.
##
##    Revision 1.2  2002/04/10 00:35:27  gmaxe
##    Tracing tool now works;
##    indexer ignores more special characters;
##    pm file is the first version; contains the proper data structures
##    to handle bsp/vxworks
##
##    Revision 1.1  2002/04/06 01:55:10  gmaxe
##    New files and new general constructs for handling tags. Uses globals better.
##    The html_look* files are designed to handle spider tracing of html systems.
##
##
##
##############################################################################


##############################################################################
##! @fn int BEGIN
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub BEGIN {
   print "\n============  Starting html_look_integrate.pl ==================================\n";
   $_file_list = "_file_list";
   $_toc_file = "tree.html";
   $_index_file = "_index_list";
   $_arg_inc = 0;
   
   $no_scope_file = 0;
   $scope_pm = "globe.pm";  # first time through; other scope stuff passed in.
   
   $in_file = "tree.js"; # default

   push (@INC, `pwd`);
   push (@INC, '../perl');
   if (0){
      print (@INC, "\n");
   }
   #####
   # All global variables are defined in the following file
   #####
   unless (open ( IN_LIST, $scope_pm)) {
      unless (open ( IN_LIST, "../perl/$scope_pm")) {
         push (@file_errors, "Cannot open file \"$scope_pm\" or \"../perl/$scope_pm\"\n");
         $no_scope_file++;
      }
   }
   close (IN_LIST);
   push (@INC, $scope_pm);
   push (@INC, "../perl/$scope_pm");

   if (!@file_errors) {
      #####
      # All global variables are defined in the following file
      #####
      require $scope_pm;
   
      if (&globe::declare_variables()) {
         print "Variables initialized from $scope_pm.\n";
      } else {
         push (@file_errors, "Could not initialize variables from $scope_pm.\n");
      }
   } # if not @file_errors
   
   

   # Get scope files if there is one.
   if (@ARGV > $_arg_inc) {
      $scope_pm = @ARGV[$_arg_inc];
      
      unless (open ( IN_LIST, $scope_pm)) {
         push (@file_errors, "Cannot open file \"$scope_pm\"\n");
         $no_scope_file++;
      }
      close (IN_LIST);
      push (@INC, $scope_pm);
      
      if (!@file_errors) {
         #####
         # All global variables are defined in the following file
         #####
         require $scope_pm;
   
         if (&xscope::declare_variables()) {
            print "Variables initialized from $scope_pm.\n";
         } else {
            push (@file_errors, "Could not initialize variables from $scope_pm.\n");
         }
      } # if not @file_errors
   } else {
      push (@file_errors, "ERROR: Need to provide a scope file as first argument.");
   } # if 1 or more arguments
   $_arg_inc++;

   # Get path to code files
  if (0) { # opt out
   if (@ARGV > $_arg_inc) {
      $root_path = @ARGV[$_arg_inc];
      $globe::path = @ARGV[$_arg_inc];
      if ($root_path =~ /\/$/) {
         print "The path specified is $root_path\n";
      } else {
         push (@file_errors, "The input argument \"$root_path\" requires a forward slash (\/) at the end.\n");
      }
   } else {
      push (@file_errors, "ERROR: root path is required.");
   } # if 1 or more arguments
   $_arg_inc++;
  } # opt out
   
   if (@ARGV > $_arg_inc) {
      $in_file = @ARGV[$_arg_inc];
      $globe::master_nav_file = @ARGV[$_arg_inc];
      print "The master file is $in_file.\n";
      unless (open ( IN_MASTER, "$in_file")) {
         push (@file_errors, "Cannot open file \"$in_file\"\n");
      }
   } else {
      push (@file_errors, "ERROR: master HTML file is required.");
   } # if 2 or more arguments
   $_arg_inc++;

   # Optional: a fully qualified path of where to start.
   if (@ARGV > $_arg_inc) {
      undef (@globe::top_files);
      push (@globe::top_files, @ARGV[$_arg_inc]);
   } # if 3 or more arguments
   $_arg_inc++;


  if (0) { # opt out
   if (@ARGV > $_arg_inc) {
      $gen_proto = @ARGV[$_arg_inc];
      if ($gen_proto !~ /\//) {
         # if the output file does not have a path, then
         # put it in the root location
         $gen_list = "_$gen_proto";
         $gen_class = "c_$gen_proto";
         $gen_proto = "$root_path$gen_proto";
         $gen_list = "$root_path$gen_list";
         $gen_class = "$root_path$gen_class";
      } else {
         @chunk = split (/\//, $gen_proto);
         $chunk[$#chunk] = "_$chunk[$#chunk]";
         $gen_list = join ("\/", @chunk);
         @chunk = split (/\//, $gen_proto);
         $chunk[$#chunk] = "c_$chunk[$#chunk]";
         $gen_class = join ("\/", @chunk);
      }
      print "The output files are $gen_proto and $gen_list\n";
   } else {
      push (@file_errors, "Need to have the output prototype file.\n");
   } # if 3 or more arguments
   $_arg_inc++;
  } # opt out

   
} # BEGIN

##############################################################################
##! @fn int main
##! @brief The main program.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
# sub main {
{
   ##############################################################################
   ## Program start
   ##############################################################################

   if (0){
      print "=== Definitions 3 \n";
      exit(1);
   }

   if (@file_errors) {
      # Makes no sense to go on if input parameters are off.
      print "\n============  Summary of errors =================================a\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
      &using_voy_nav();
      exit(1);
   }

   #####
   # MASTER FILE DEFINITIONS
   # Get the master definitions
   #####
   while (<IN_MASTER>){    # entire master file into memory.
      $globe::master_nav .= $_;
   }
   close (IN_MASTER);
   &globe::get_master_nav_info;
   if (!exists ($globe::m_info{order})) {
      if (1) {
         push (@file_errors, "ERROR: The master file \"$globe::master_tree_file\" does not have the ordering.");
         print "Ordering begins with \"$globe::m_define{order}[0]\" \nand ends with \"$globe::m_define{order}[1]\"\n";
         print "Inbetween on a line by themselves is the directory doc_publish subdirectory name.\n";
      }
      exit(1);
   } else {
      if (0) {
         print "Got to here...\n$globe::m_info{order}\n======\n";
         for ($i = 0; $i <= $#{$globe::m_info{_array_order}}; $i++) {
            foreach $k (keys %{$globe::m_info{_array_order}[$i]}) {
               print "$i $k $globe::m_info{_array_order}[$i]{$k}\n";
            }
         }
         # exit(1);
      }
   }
   #####

   #####
   # This is the main loop that does a spider trace through the documentation.
   #####
   if (1) { # Turn on/off spider default 1 -- mandatory
      foreach $in_file (@xscope::top_files){
         @_name = split (/\//, $in_file);
         pop (@_name);  # This is the filename, which we don't want.
         $globe::org{$in_file}{path} = join ("\/", @_name, "");
         if (0){
            print "Starting file = $in_file\n";
         }
         if (0){
            print "root path is $globe::org{$in_file}{path}\n";
         }
         #####
         # Read in file and get its child hyperlinks
         #####
         if (&get_input_file ($in_file)) {
            # good
            # Level is 2 for get_hyperlinks, because 1 is already the main guy.
            # when using nested scripts, main guy is already handled; 
            # start numbering with 1
            if (!&get_hyperlinks ($in_file, 1)) {
               # bad; did not succeed getting hyperlinks from file
               push (@file_errors, "No hyperlinks found in $in_file.");
            }
         } else {
            # bad
            # Turned off because trapped elsewhere
            if (0) {
               push (@file_errors, "Could not find \"$in_file\". :)");
            }
         } # get_input_file
         if (0) {
            print "Title is \"$globe::org{$in_file}{title}\"\n";
         }
         if (1) {
            print "Starting spider trace...\n";
            &spider_trace ($in_file, 1);  # base level is 1, the starting point.
         }
         if (0) {
            print "\nEnding spider trace...\n";
         }
      } # for each top_file
   }  # Turn on/off spider

   undef (%globe_all_files);
   $globe_file_cnt=0;
   #####

   #####
   # Test routines that are recursively called.
   # Turned off, but when on can help debug what is being collected.
   # leave this off
   #####
   if (0) { # debug if default 0
      local ($globe::_last_level) = 1;
      # local ($globe::_last_level) = 0;
      print "Testing the structure...\n";
      print "Start last level = $globe::_last_level\n";
      &testing_structure ($xscope::top_files[0]);
      $globe::_last_level = 1;
      # $globe::_last_level = 0;
      print "======\nSecond time through...\n";
      print "Start last level = $globe::_last_level\n";
      &testing_structure ($xscope::top_files[0]);
      if (0){
         foreach $in_file (sort keys %{globe_all_files}) {
            print "input $in_file\n";
         }
      }
      exit(1);
   } # debug if
   #####
   
   #####
   # Used to see what's in the data structures.
   # Leave this off
   #####
   if (0) { # debug if default 0
      foreach $in_file (sort keys %{globe::org}) {
         print "file $in_file\n";
         print "  g title $globe::org{$in_file}{title}\n";
         if (1){
            print "  g title $globe::org{$in_file}{title}\n";
            print "  g path $globe::org{$in_file}{path}\n";
         }
         if (0){
            foreach $k (sort keys %{$globe::org{$in_file}}) {
               print "  $k \$globe::org{\$in_file}{$k} = $globe::org{$in_file}{$k}\n";
            }
         }
      }
   } # debug if
   #####

   undef (%globe_all_files);
   $globe_file_cnt=0;
   
   #####
   # Creates the output scripts for each top file.
   #####
   if (1) { # turn on/off default 1
      foreach $in_file (@xscope::top_files) {
         # $globe::org{$in_file}{title}
         ###
         # Create a file name from the title.
         # Replace characters.
         ###
         $tree = $globe::org{$in_file}{title};
         $tree =~ s/[\s+]/_/g;
         $tree =~ s/[\'\"\`\~\!\@\#\$\%\^\&\*\(\)\=\+\{\}\[\]\?\/\\\,\.\?\>\<\|]/_/g;
         $tree = "tree_$tree";
         $globe::org{$in_file}{script_local} = $tree;
         $m_tree = "m_$tree";
         $globe::org{$in_file}{script_master} = $m_tree;
         $tree = "$globe::path$tree";
         $m_tree = "$globe::path$m_tree";
         if (0) {
            print "Old title \"$globe::org{$in_file}{title}\"\n  new \"$tree\"\n\nmaster \"$m_tree\"\n";
         }
         
         # $tree = "$globe::path";
         # $tree .=  "tree_$q";
         if ((0) && (&output_structure_script ($tree, $in_file, "local"))) {
            # assuming successful
         }
         if ((1) && (&output_structure_script ($m_tree, $in_file, "master"))) {
            # assuming successful
         }
      } # foreach $in_file
   } # turn on/off
   #####

   ###################
   # special case a script file for the Top Most master file.
   ###################
   if (1) { # turn on/off
      $tree = $globe::path;
      $tree .= "tree_start_here";
      if ((0) && (&starting_point_script ($tree, "script_local", "local"))){
      } # starting_point_script
      $tree = $globe::path;
      $tree .= "m_tree_start_here";
      if ((1) && (&starting_point_script ($tree, "script_master", "master"))) {
      } # starting_point_script
   } # turn on/off
   #####



   undef (%globe_all_files);
   $globe_file_cnt=0;
   

   ###################
   # Take care of index tokens
   ###################
   if (1) { # turn on/off default 1
      print "Generating index tokens $_index_file...\n";
      $_in_index_file = "$globe::path$_index_file";

      $q = 0;
      foreach $in_file (@xscope::top_files) {
         if (&handle_index_tokens ($in_file)) {
            # assuming successful
         }
         $q++;
      } # foreach $in_file
      
      if (@globe::index_info) {
         print "Initial index token count: $#globe::index_info\n";
         if (1) { # 4/8/2002 Get rid of duplicate entries
            @sm_index = sort (@globe::index_info);
            @globe::index_info = @sm_index;
            undef (@sm_index);
            $remember = $globe::index_info[$i];
            push (@sm_index, $globe::index_info[0]);
            #####
            # Loop to get rid of duplicates
            #####
            for ($i=1; $i <= $#globe::index_info; $i++) { # not starting on first 
               if ($remember ne $globe::index_info[$i]) {
                  push (@sm_index, $globe::index_info[$i]);
               }
            }
            @globe::index_info = @sm_index;
            undef (@sm_index);
            
         }
         print "Final index token count: $#globe::index_info\n";
         
         # clean up memory
         
         
         unless (open ( OUT_INDEX_LIST, ">$_in_index_file")) {
            push (@file_errors, "Cannot open file \"$_in_index_file\"");
            exit(1);
         }
         if (1) { # 4/5/2002 one way of doing index tokens
            for ($i=0; $i<@globe::index_info; $i++){
               # print "$i = $globe::index_info[$i]\n";
               print (OUT_INDEX_LIST "$globe::index_info[$i]\n");
            }
         }
         close (OUT_INDEX_LIST);
      }
   } # turn on/off
   #####


   ###################
   # Take care of navigation and copyright in the HTML file
   ###################
   if (1) { # turn on/off default 1
      print "Updating the individual HTML files...\n";
      undef (%globe_all_files);
      $globe_file_cnt=0;

      $q = 0;
      foreach $in_file (@xscope::top_files) {
         if (&html_file_update ($in_file)) {
            # assuming successful
         } else {
            # happens when you can't write the file out
         }
         $q++;
      } # foreach $in_file
      
   } # turn on/off
   #####
   
   
   if (0) {
      print "got this far...\n";
      exit(0);
   }

   ##############################################################################
   ## End of Program
   ##############################################################################
   exit(0);
} # the main routine

##############################################################################
##! @fn int using_voy_nav  ( )
##! @brief What to do when no arguments are given.
##! @param None
##! @return None
##! 
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub using_voy_nav {
   print "\nhtml_look_integrate.pl operates on text files that indicate heirarchical \n";
   print "structure of the system. Needed arguments:\n";
   print "- appropriate scope pm; required. \n";
   print "- name of master file. \n";
   print "- [optional] fully qualified path and name of starting HTML file for spider trace. \n";
   print "    If not here, the starting HTML needs to be specified in the scope file. \n";
   # print "- output toc script file; optional. \n";

   return;
}


##############################################################################
##! @fn int get_input_file
##! @brief Grabs the information from the input file and puts into an array
##! of hash elements.
##!
##! @param file_incoming This is the file to open.
##!
##! Reads in the file into memory for later processing to find anchors.
##! While it's here, it tries to find the title for the file for later display
##! and file naming purposes.
##!
##! @ingroup tp_nav
##############################################################################
sub get_input_file {
   local ($file_incoming) = $_[0];
   undef ($globe::entire_file);
   
   if (0) {
      print "get_input_file with $file_incoming.\n";
   }
   
   unless (open ( IN_COMING, "$file_incoming")) {
      push (@file_errors, "Cannot open file \"$file_incoming\" <<<");
      return (0);
   }
   while (<IN_COMING>){
      $globe::entire_file .= $_;
   }
   close (IN_COMING);
   if (0) {
      print "$globe::entire_file\nglenn\n";
   }
   $not_critical = 0;
   $case_in = 1;

   # Look for an h1
   ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                               "<h1", ">", 
                               $not_critical, $case_in);
   if ($piece) {
      # rebuilt accurate test string
      $test = "<h1$piece>";
      # Redo this on entire file now that complete first tag is known.
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                            $test, "</h1", 
                            $not_critical, $case_in);
      if ($piece) {
         # Strip out html tags within this piece.
         while ($piece =~ /\</) {
            ($before, $throw_away, $after) = &globe::get_tag_chunk( $piece,
                            "<", ">", 
                            $not_critical, $case_in);
            $piece = join ("", $before, $after);
         } # while temp title.
          
         $piece =~ s/\n//g;  # strip out carriage return
         $piece =~ s/^[\s+]//g;  # strip out leading/trailing space
         $piece =~ s/[\s+]$//g;  # strip out leading/trailing space
         if (0) {
            print "h1 title is \"$piece\"\n";
         }
         $globe::org{$file_incoming}{title} = $piece;
      } else {
         print "Could find no h1.\n";
      }
   } else {
   
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                               "<title>", "</title>", 
                               $not_critical, $case_in);
      if (($piece) && ($piece =~ /[\w+]/)) {
         $piece =~ s/\n//g;  # strip out carriage return
         while ($piece =~ /^[\s+]/){
            $piece =~ s/^[\s+]//;  # strip out leading/trailing space
         }
         while ($piece =~ /[\s+]$/){
            $piece =~ s/[\s+]$//;  # strip out leading/trailing space
         }
         if (0) {
            print "<title> is \"$piece\"\n";
         }
         $globe::org{$file_incoming}{title} = $piece;

       } else {
         print "Could find no h1 or title.\n";
       }
   }
   if (0) {
      print "title from get_input \"$globe::org{$file_incoming}{title}\" for $file_incoming\n";
   }
   
   if (1) {
      # Remove the voyant tags so that we don't get extra links to mess us up.
      &remove_voyant_tags;
   }
   
   return (1);   
} # get_input_file

##############################################################################
##! @fn int get_hyperlinks
##! @brief Parses through the information in the globe::entire_file buffer
##! for anchors.
##!
##! @param _file_4_links contains the fully qualified name of the file.
##! @param _level contains the level, or distance from, the root for use in
##! assigning to the hyperlinks discovered.
##!
##! All qualified anchors represent children of the _file_4_links. This
##! routine creates the children's url, display text, and level.
##!
##! This routine calls the routine verify_link, which has criteria to test 
##! against. The criteria is intended to make the qualification of the link 
##! fail. In other words, when it returns from verify_link with a failed (0) criteria,
##! this get_hyperlinks routine knows that it should not add a child for
##! that hyperlink. 
##!
##! @bug Assumes that $globe::entire_file has an HTML file in it.
##! @ingroup tp_nav
##############################################################################
sub get_hyperlinks {
   local ($_file_4_links) = $_[0];
   local ($_level) = $_[1];
   local ($before) = "";
   local ($piece) = 0;
   local ($after) = "";
   local ($b_href) = "";
   local ($href) = 0;
   local ($keep_href) = "";
   local ($a_href) = "";
   local ($d);
   $_l_cnt = 0;
   
   if (0){
      print "get_hyperlinks from $_file_4_links.\n";
   }
   
   @pot_link_chunk = split (/\<a[\s+]/i, $globe::entire_file);
   if ($#pot_link_chunk < 1) {
      # file doesn't have any hyperlinks
      return (0);
   }
   POTENTIAL_L: for ($d = 1; $d <= $#pot_link_chunk; $d++){
      # Purposely skipping the zero element; has no link into it.
      #####
      # Work on hyperlink
      #####
      $piece = 0;
      ($before, $piece, $after) = &globe::get_tag_chunk( $pot_link_chunk[$d],
                               "href", ">", 
                               $not_critical, $case_in);
      if ($piece) {
         # This contains the information inside the anchor
         ($b_href, $href, $a_href) = &globe::get_tag_chunk( $piece,
                               "\"", "\"", 
                               $not_critical, $case_in);
         if ($href) {
            #####
            # If this has valid information, then let's first strip out
            # and further test what the link is with verify_link.
            #####
            $keep_href = &verify_link ($href, $_file_4_links);
            if ($keep_href) {
               #####
               # This is for valid hyperlink references.
               #####
               $href = $keep_href;
               if (0) {
                  # old way; 
                  local ($path) = $globe::org{$_file_4_links}{path};
                  $globe::org{$_file_4_links}{child}[$_l_cnt]{url}= "$path$href";
               }
               if (0) {
                  ####
                  # Don't assign this until we're sure that we have valid
                  # text to go with it.
                  # If we wait, we won't get empty index tokens or
                  # TOC entries
                  ####
                  $globe::org{$_file_4_links}{child}[$_l_cnt]{url}= "$href";
               }
               if (0) {
                  print "... link = $globe::org{$_file_4_links}{child}[$_l_cnt]{url}\n";
               }
            } else {
               # if verify_link returns 0, then it is not a link we want to hold on to.
               next POTENTIAL_L;
            }
         } else {
            # Did not contain an href; but if it has a name, it
            # still might be useful to have in script.
            # TBD: implement this later if needed.
            # print "Did not contain a valid href for me.\n";
            
            next POTENTIAL_L;
         }         
      } else {
         # This anchor does not have a hyperlink
         next POTENTIAL_L;
      }
      #####
      # Work on link text
      # $after is still valid from above; contains what followed starting 
      # anchor designation <a ....>
      #####
      ($b_anc_end, $a_anc_end) = split (/\<\/a>/i, $after, 2);
      undef ($remember);
      # Get rid of any html that might be in link text;
      while ($b_anc_end =~ /\</) {
         # while html still in b_anc_end
         ($before, $piece, $after) = &globe::get_tag_chunk( $b_anc_end,
                               "<", ">", 
                               $not_critical, $case_in);
         if (($piece =~ /^img/i) && ($piece =~ /alt[\s*]\=/i)){
            # This might have some valid information to remember
            @chunks = split (/alt[\s*]\=/i, $piece, 2);
            ($b2, $p2, $a2) = &globe::get_tag_chunk( $chunks[1],
                               "\"", "\"", 
                               $not_critical, $case_in);
            $remember=$p2;
            if (1) {
                print "Alt text = \"$remember\" for $globe::org{$_file_4_links}{child}[$_l_cnt]{url}\n";
            }
         }
         
         # rebuild and retry
         $b_anc_end = join ("", $before, $after);
         $b_anc_end =~ s/\n/ /g;
         $remove = "&nbsp;";
         $b_anc_end =~ s/$remove/ /g;
         $b_anc_end =~ s/&nbsp;/ /g;
         $b_anc_end =~ s/  / /g;
         $b_anc_end =~ s/  / /g;
         $b_anc_end =~ s/  / /g;
         $b_anc_end =~ s/[\s+]$//;
      } # while html still in b_anc_end
      
      #####
      # Get rid of portion of entries that begin with numbers or punctuation
      #####
      if (1) {
         # Get rid of topics that are all numbers.
         if ($b_anc_end !~ /[\D+]/) {
            if (0) {
               print "removing \"$b_anc_end\"\n";
            }
            next POTENTIAL_L;
         }
         if ($b_anc_end =~ /^[\d+][\.]/) {
            while ($b_anc_end =~ /^[\d+][\.]/){
               @_remove = split (/^[\d+][\.]/, $b_anc_end, 2);
               $remove = shift (@_remove);
               $b_anc_end = join ("", @_remove);
            }
            while ($b_anc_end =~ /^[\s+]/) {
               $b_anc_end =~ s/^[\s]//;
            }
            if (0) {
               print "has (had) digits \"$b_anc_end\"\n";
            }
         }
         if ($b_anc_end =~ /^[\d+][\s+]/) {
            while ($b_anc_end =~ /^[\d+][\s+]/){
               @_remove = split (/^[\d+][\s+]/, $b_anc_end, 2);
               $remove = shift (@_remove);
               $b_anc_end = join ("", @_remove);
            }
            while ($b_anc_end =~ /^[\s+]/) {
               $b_anc_end =~ s/^[\s]//;
            }
            if (0) {
               print "has (had) digits \"$b_anc_end\"\n";
            }
         }
         if ($b_anc_end =~ /^[\d+]$/) {
            if (0) {
               print "has (had) digits \"$b_anc_end\"\n";
            }
            while ($b_anc_end =~ /^[\d+]$/){
               @_remove = split (/^[\d+]/, $b_anc_end, 2);
               $remove = shift (@_remove);
               $b_anc_end = join ("", @_remove);
            }
            while ($b_anc_end =~ /^[\s+]/) {
               $b_anc_end =~ s/^[\s]//;
            }
            if (0) {
               print "has (had) digits \"$b_anc_end\"\n";
            }
         }
         if ($b_anc_end =~ /^[A-Z]\.[\d]/) {
            if (0) {
               print "has (had) digits \"$b_anc_end\"\n";
            }
            while (($b_anc_end =~ /^[A-Z]\.[\d]/) 
               || ($b_anc_end =~ /^[\d+]\./) 
               || ($b_anc_end =~ /^\.[\d+]/)  ){
               if ($b_anc_end =~ /^[A-Z]\.[\d]/) {
                  @_remove = split (/^[A-Z]\./, $b_anc_end, 2);
               } elsif ($b_anc_end =~ /^[\d+]\./) {
                  @_remove = split (/^[\d+]\./, $b_anc_end, 2);
               } else {
                  @_remove = split (/^\.[\d+]/, $b_anc_end, 2);
               }
               
               $remove = shift (@_remove);
               $b_anc_end = join ("", @_remove);
            }
            while ($b_anc_end =~ /^[\s+]/) {
               $b_anc_end =~ s/^[\s]//;
            }
            if (0) {
               print "has (had) digits \"$b_anc_end\"\n";
            }
         }
         if ($b_anc_end =~ /^[\d+][\,]/) {
            while ($b_anc_end =~ /^[\d+][\,]/){
               @_remove = split (/^[\d+][\,]/, $b_anc_end, 2);
               $remove = shift (@_remove);
               $b_anc_end = join ("", @_remove);
            }
            while ($b_anc_end =~ /^[\s+]/) {
               $b_anc_end =~ s/^[\s]//;
            }
            if (0) {
               print "has (had) digits \"$b_anc_end\"\n";
            }
         }
         if ($b_anc_end =~ /^[\d+]&nbsp;/) {
            while ($b_anc_end =~ /^[\d+]/){
               @_remove = split (/^[\d+]&nbsp;/, $b_anc_end, 2);
               $remove = shift (@_remove);
               $b_anc_end = join ("", @_remove);
            }
            while ($b_anc_end =~ /^[\s+]/) {
               $b_anc_end =~ s/^[\s]//;
            }
            if (0) {
               print "has (had) digits \"$b_anc_end\"\n";
            }
         }
         if ($b_anc_end =~ /^[\s+]/) {
            $b_anc_end =~ s/^[\s+]//g;
            if (0) {
               print "has (had) digits \"$b_anc_end\"\n";
            }
         }
      }
      
      #####
      # If it doesn't have text, try to give it some alternate 
      # text or return with 0;
      #####
      if ($b_anc_end !~ /[\S+]/){
         # See if the $piece contains alt text or an image
         if ((0) && ($remember =~ /[\w+]/)){
            $b_anc_end = $remember;
         } else {
            # There was no valid text, so go and get next hyperlink
            next POTENTIAL_L;
         }
      }

      if (0) {
         print "Link text = $b_anc_end\n";
      }


      # clean up the text from things like &nbsp
      $b_anc_end =~ s/\&nbsp\;/ /g;
      $b_anc_end =~ s/&nbsp;/ /g;
      $b_anc_end =~ s/^\s+//;
      $b_anc_end =~ s/\s+$//;
      
      $globe::org{$_file_4_links}{child}[$_l_cnt]{text}=$b_anc_end;
      
      #####
      # The children own their level rather than the owning file,
      # so that they'll be displayed properly when
      # referenced from multiple locations.
      #####
      $globe::org{$_file_4_links}{child}[$_l_cnt]{level} = $_level;
      
      #####
      # URL comes from above.
      #####
      $globe::org{$_file_4_links}{child}[$_l_cnt]{url}= "$href";

      if (0) {
         print "Link text = $globe::org{$_file_4_links}{child}[$_l_cnt]{text}\n";
         print "   Link url = $globe::org{$_file_4_links}{child}[$_l_cnt]{url}\n";
         print "   Link level = $globe::org{$_file_4_links}{child}[$_l_cnt]{level}\n";
         print "   Link parent = $_file_4_links\n";
      }

      #####
      # Should be the last thing; increment counter
      #####
      $_l_cnt++;
   } # for
   
   if (0) {
      print "org file: $_file_4_links with children\n";
      for ($d=0; $d <= $#{$globe::org{$_file_4_links}{child}}; $d++){
            print "  child $d $globe::org{$_file_4_links}{child}[$d]{url}\n";
      }
      print "  org file above: $_file_4_links\n";
      # print "ending get_hyperlinks\n";
      # exit(1);
   }
   return (1);

} # get_hyperlinks 


##############################################################################
##! @fn int verify_link
##! @brief Tests the potential link against various criteria to validate
##! whether the link is of value.
##! 
##! @param potential_link Contains a fully qualified hyperlink.
##! @param contains the owning file.
##! 
##! @return If a criteria is matched, this routine returns 0 (meaning
##! uninteresting link. If not of the criteria sets off a flag, this then
##! returns the potential link as validated.
##! 
##! Criteria that makes the link uninteresting are things like:
##! linking to itself, up linking to a top-level entity, linking to
##! an html topic that we want to exclude as being a child, linking to
##! other children at the same level, etc.
##!
##! @ingroup tp_nav
##############################################################################
sub verify_link {
   local ($potential_link) = $_[0];
   local ($owning_file) = $_[1];
   local ($_href) = "";


   if (0) {
      print "Entering verify_link with\n\t$potential_link\n\t$owning_file\n";
   }
   { # saves removing indents
      { # saves removing indents
         if ($potential_link) {
            # strip out what we don't want, like leading ./
            local ($_rel) = "\.\/";
            while ($potential_link =~ /^$_rel/){
            if (0) { print "\tbefore $potential_link "; }
               $potential_link =~ s/^$_rel//;
               if (0) { print "after $potential_link \n"; }
            }
            ####
            # VxWorks had some strange file names with ( and ), but
            # escaped out with the characters below in the links.
            ####
            if ($potential_link =~ /\%28/){
               if (0) { print "==== before $potential_link "; }
               $potential_link =~ s/\%28/\(/;
            }
            if ($potential_link =~ /\%29/){
               $potential_link =~ s/\%29/\)/;
               if (0) { print "after $potential_link \n"; }
            }
            if (0) {
               print "Hyperlink = $potential_link\n";
            }
            #####
            # Test for if link within same page
            # If so, we don't want it
            #####
            if ($potential_link =~ /\#/){
               # See if there is any meat to this anchor
               @temp_sides = split (/\#/, $potential_link, 2);
               if ($temp_sides[0] !~ /[\w+]/){
                  # If the left side has no meat, it means it is a link
                  # within the page. We can safely ignore.
                  return (0);
               }
            }
            if (($potential_link =~ /mailto\:/i)
              || ($potential_link =~ /http\:\/\//i)
              || ($potential_link =~ /www\./i))
            {
               # We can safely ignore.
               return (0);
            }
            if (0){
               print "href $potential_link and \n\t path $globe::org{$owning_file}{path}\n";
               # exit(1);
            }
            
            
            #####
            # get rid of relative path markings
            #####
            local ($path) = $globe::org{$owning_file}{path};
            if (0) { 
               print "\n$potential_link and path before = $path \n";
            }
            local ($_href) = $potential_link;
            local ($_rel) = "../";
            if ($_href =~ /^$_rel/){
               if (0) { 
                  print "\n$potential_link and path before = $path \n";
               }
               while ($_href =~ /^$_rel/){
                  $_href =~ s/$_rel//;
                  @_ath = split (/\//, $path);
                  pop (@_ath);
                  $path = join ("\/", @_ath);
                  if ($path !~ /\/$/){
                     $path .= "\/";
                  }
               } # while
               if (0) { 
                  print "$_href at ppath after = $path\n";
                  # exit(1);
               } 
               $potential_link = $_href; 
            } # if
            #####
            # Test for targets and removing links that are to the
            # same file
            #####
         if (1) {
            if ($potential_link =~ /\#/) {
               @_ath = split (/\#/, $potential_link, 2);
               $_href = $_ath[0];
            } else {
               $_href = $potential_link;
            }
            
            if (($owning_file =~ /$_href/) && ($owning_file =~ /$potential_link/)) {
               # Don't add hyperlinks that take you back to the same file
               # However, if the link has targets, then it is okay
               if (0) {
                  print "test1 this=$_href against that=$owning_file \n";
               }
               return (0);
            }
         }  # if (0)
            #####
            # Rebuild a true link file name
            #####
            $potential_link = "$path$potential_link";
            
            #####
            # Don't add hyperlinks back to itself.
            #####
            if ($globe::org{$owning_file}{url} =~ /$potential_link/){
               return (0);
            }
            
            #####
            # Don't add duplicate hyperlinks from other children at that level.
            #####
            for ($r = 0; $r <= $#{$globe::org{$owning_file}{child}}; $r++) {
               if ($globe::org{$owning_file}{child}[$r]{url} =~ /$potential_link/){
                  if (0){
                     print "RAN INTO SOME DUPLICATION; elimination\n";
                     exit(1);
                  }
                  return (0);
               }
            }

            #####
            # Don't add hyperlinks that go to files with these names
            #####
            foreach $_href (@xscope::gen_ex_child){
               if (0) {
                  print "verifying against $_href,\n\t $potential_link\n";
               }
               if (($_href =~ /$potential_link/) || ($potential_link =~ /$_href/)) {
                  if (0){
                     print "Not adding child link to top level $potential_link.\n";
                     print "Not validating $potential_link\n";
                  }
                  return (0);
               }
            } # foreach #_href


            
            #####
            # Don't add hyperlinks that go to one of the top-level to-be-excluded files
            #####
            foreach $_href (@xscope::ex_as_child){
               if (0) {
                  print "verifying against $_href,\n\t $potential_link\n";
               }
               if (($_href =~ /$potential_link/) || ($potential_link =~ /$_href/)) {
                  if (0){
                     print "Not adding child link to top level $potential_link.\n";
                     print "Not validating $potential_link\n";
                  }
                  return (0);
               }
            } # foreach #_href
            if (0) {
               print "verified $potential_link\n";
               # exit(1);
            }
            
            #####
            # If we made it this far, then we can return the modified potential_link
            #####
            return ($potential_link);
            
         } # if for potential_link

      } # saves removing indents
   } # saves removing indents            
} # verify_link 



##############################################################################
##! @fn int spider_trace
##! @brief Traces the child hyperlinks of a starting file and places them
##! into the data structure.
##!
##! @param start_file The starting HTML file to trace.
##!
##! This calls routines to open and read the children HTML files and to locate the
##! appropriate hyperlinks within those files.
##!
##! @bug This assumes that the data structure has already been started by
##! having a root file, by having already opened it and acquired its children.
##!
##! @ingroup tp_nav
##############################################################################
sub spider_trace {
   local ($start_file) = $_[0];
   local ($base_level) = $_[1];
   local ($new_level) = 0;
   local ($_f_cnt) = 1; # set so that loop will start
   local ($new_file_entry);
   local (%to_do);
   
   if (1) {
      print "spider trace $start_file\n";
   } else {
      print "$globe_file_cnt...";
   }
   $globe_file_cnt++;

   #####
   # New level should be one more than current level
   #####
   $new_level = $base_level + 1;
       


   if (exists ($globe::org{$start_file}{o_level})) {
      $globe::org{$start_file}{o_level} .= "...$base_level";
   } else {
      $globe::org{$start_file}{o_level} = $base_level;
   }
   { # FILLER BRACKET
      for ($i = 0; $i <= $#{$globe::org{$start_file}{child}}; $i++) {
         # strip out any targets inside of a file...
         if ($globe::org{$start_file}{child}[$i]{url} =~ /\#/){
            @_name = split (/\#/, $globe::org{$start_file}{child}[$i]{url}, 2);
            # $new_file_entry .= @_name[0];
            $new_file_entry = @_name[0];
         } else {
           # child url name is good to go
           $new_file_entry = $globe::org{$start_file}{child}[$i]{url};
         }
      
         if (0) {
            print "Child url $globe::org{$start_file}{child}[$i]{url}\n";
            print "Child1 $new_file_entry\n";
         }
         #####
         # Add this child to a list for later creation if it doesn't exist.
         #####
         if ((1) && (!(exists ($globe::org{$new_file_entry})))) {
            #####
            # The entry contains the path without any targets.
            # The url contains the path with targets.
            #####
            $to_do{$new_file_entry} = $globe::org{$start_file}{child}[$i]{url};

            #####
            # Set up the path to be used when generating links
            #####
            @_name = split (/\//, $new_file_entry);
            pop (@_name);  # This is the filename, which we don't want in path.
            $globe::org{$new_file_entry}{path} = join ("\/", @_name, "");
            #####
            # The entry contains the path without any targets.
            # The url contains the path with targets.
            #####
            $globe::org{$new_file_entry}{url} = $to_do{$new_file_entry}; # should have the url
            
            #####
            # Read in file and get its child hyperlinks
            #####
            if (&get_input_file ($new_file_entry)) {
               # good
               if (0) {
                  print "after input title = $globe::org{$new_file_entry}{title}\n";
               }
               if (!&get_hyperlinks ($new_file_entry, $new_level)) {
                  # bad; did not succeed getting hyperlinks from file
                  push (@file_errors, "No hyperlinks found in $new_file_entry.");
                  print ("\nNo hyperlinks found in $new_file_entry\n");
               }
               if (0) {
                  print "base $base_level new $new_level for all children\n";
                  print "\tafter hyperlink title = $globe::org{$new_file_entry}{title}\n";
                  print "\t$new_file_entry\n";
               }
            } else {
               # bad
               # Messaging turned off because trapped elsewhere.
               if (0) {
                  push (@file_errors, "Could not find \"$new_file_entry\" !!");
                  print ("\nCould not find $new_file_entry !!!?\n");
               }
            } # get_input_file
               
          } # if ((1) && (!(exists ($globe::org{$new_file_entry}))))
       } # for ($i = 0;...

       if (0) {
          print "======== to be traced\n";
          foreach $new_file_entry (keys %to_do) {
              print "key: $new_file_entry\n";
              print "value: $to_do{$new_file_entry}\n";
          }
       }
       

       
       foreach $new_file_entry (keys %to_do) {
          if ($to_do{$new_file_entry}) {
            #####
            # if new entry doesn't exist; make it exist
            # Handle its title and path
            #####
            
            #####
            # call this routine again with that new file as the starting point
            # RECURSIVE CALL
            #####
            if (0) {
               print "Recursive call.\n  owner = $start_file\n  with $new_file_entry\n";
               print "   path $globe::org{$new_file_entry}{path}\n  title = $globe::org{$new_file_entry}{title}; \n";
            }
            &spider_trace($new_file_entry, $new_level);
            if (0) {
               print "  return from spider with $new_file_entry\n";
               print "  title = $globe::org{$new_file_entry}{title}; \n";
            }
            
            $to_do{$new_file_entry} = 0;
          }  # if (!($to_do{$new_file_entry}))
            
          
       }  # foreach $new_file_entry

   }   # FILLER BRACKET   
         
   undef (%to_do);   
   
   if (0) {
      foreach $in_file (sort keys %{globe::org}) {
         print "debug file $in_file\n";
         print "  g title $globe::org{$in_file}{title}\n";
      }
   }
   
   return (1);

} # spider_trace


##############################################################################
##! @fn int testing_structure
##! @brief Traces through the created data structure.
##! 
##! @param begin_file is the starting point to begin tracing.
##! 
##! This assumes that the begin_file is present in the data structure and
##! has children to trace through.
##! 
##! @note This is a RECURSIVE algorithm and can be used as a template for other
##! recursive things involving the same data structure.
##!
##! @ingroup tp_nav
##############################################################################
sub testing_structure {
   local ($begin_file) = $_[0];
   local ($child);
   local ($loc_title);
   local ($f);
   if (0) {
      print "test_file $begin_file\n$globe::org{$begin_file}{title}\n";
   }
   
   for ($f = 0; $f <= $#{$globe::org{$begin_file}{child}}; $f++){
      $child = $globe::org{$begin_file}{child}[$f]{url};
      if ($child =~ /\#/){
         @_name = split (/\#/, $child, 2);
         $child = $_name[0];
      }
      if (1) {
         print "  child level=$globe::org{$begin_file}{child}[$f]{level} $globe::org{$begin_file}{child}[$f]{text}\n";
         print "      p level=$globe::org{$begin_file}{o_level} $globe::org{$begin_file}{title}\n";
         # print "     parent $begin_file\n";
         # print "  $globe::org{$begin_file}{child}[$f]{url}\n";
      }
      if ($globe::org{$begin_file}{child}[$f]{level} > $globe::_last_level + 1) {
         print "DANGER WILL ROBINSON!!! DANGER!!!\n";
         print "child $globe::org{$begin_file}{child}[$f]{url}\n";
         print "parent $begin_file\n";
         print "parent level $globe::org{$begin_file}{o_level} \nchild level $globe::org{$begin_file}{child}[$f]{level}\n";
         print "parent has children, who are:\n";
         for ($b = 0; $b <= $#{$globe::org{$begin_file}{child}}; $b++) {
            print "\t $b: $globe::org{$begin_file}{child}[$b]{text}";
            print "; level $globe::org{$begin_file}{child}[$b]{level}\n";
         }
         print "===============\n";
         
         #####
         # Fixing the level
         #####
         # $globe::org{$begin_file}{child}[$f]{level}--;
         print "Fixing the level\n";
         print "parent level $globe::org{$begin_file}{o_level} \nchild level $globe::org{$begin_file}{child}[$f]{level}\n";
         print "===============\n";
         
         # exit(1);
      } 
      $globe::_last_level = $globe::org{$begin_file}{child}[$f]{level};
      if (! (exists ($globe_all_files{$child}) )) {
         $globe_all_files{$child} = 1;
         $globe_file_cnt++;
         #####
         # Recursive Call
         #####
         &testing_structure ($child);
      }
      
   } # for each child

   return (1);

} # testing_structure

##############################################################################
##! @fn int output_structure_script
##! @brief The starting point for outputting the TOC script files.
##!
##! @param _in_file / htree_file the prefix of the name of the script file to output.
##! @param trace_start_file starting point in the data structure that has
##! children to trace through.
##! @param globe_path_purge the key into the $globe::nav_path_purge hash,
##! which contains paths to purge from the output hyperlinks in the TOC, etc.
##! 
##! @note This calls script_structure which is a recursive routine that traces through
##! the data structure.
##!
##! @ingroup tp_nav
##############################################################################
sub output_structure_script {
   local ($_in_file) = $_[0];
   local ($htree_file) = $_[0];
   local ($trace_start_file) = $_[1];
   local ($globe_path_purge) = $_[2]; # used globally by subsequent routines
   undef (%globe_found_files);  # Used in a global way
   local ($s_path) = "";
   $globe_file_cnt = 0;
   
   if (($htree_file =~ /(\.html)$/i) || ($htree_file =~ /(\.htm)$/i)){
      @_name = split (/\.htm/i, $htree_file, 2);
      $htree_file = $_name[0];
   }
   $htree_file .= ".script";
   
   if (1) {
      print "The output TOC script file is \"$htree_file\".\n";
      print "The portion of the path to purge from hyperlinks is \"$globe::nav_purge_path{$globe_path_purge}\".\n";
   }

   unless (open(OUT_SCRIPT, ">$htree_file")) {
      push (@file_errors, "Cannot open file \"$htree_file\".");
      return(0);
   }
   ##################################################
   # Root level of tree.
   ##################################################
   ($okay, $s_path) = &purge_full_path ($trace_start_file,  $globe::nav_purge_path{$globe_path_purge});
   if (0) {
      print "before $trace_start_file\nokay=$okay after $s_path\n";
   }

   { # filler bracket
      #####
      # This top title is part of the master and does not need to be
      # part of this.
      #####
      print (OUT_SCRIPT "Item level=1 ");
      print (OUT_SCRIPT "image=nav_doc.gif ");
      print (OUT_SCRIPT "url=$s_path,basefrm ");
      print (OUT_SCRIPT "selected=YES ");
      print (OUT_SCRIPT "text=$globe::org{$trace_start_file}{title}\r\n");
   } # filler bracket
   
   ######
   # Starting routine that can get called recursively
   ######
   print "$globe_file_cnt Starting the building of scripts...\n";
   &script_structure($trace_start_file);
   print "\n  $globe_file_cnt Ending the building of scripts...\n";



   ##################################################
   # Clean up navigation.
   ##################################################
   close (OUT_SCRIPT);
   undef (%globe_found_files); 
   return (1);
} # output_structure_script


##############################################################################
##! @fn int purge_full_path
##! @brief Changes the file name by purging a portion of its path.
##! 
##! @param _filename original filename.
##! @param part2rm portion of the path in the filename to remove.
##! 
##! @return 1 and the updated filename if successful; 0 and the unmodified
##! filename if unsuccessful.
##!
##! @ingroup tp_nav
##############################################################################
sub purge_full_path {
   local ($_filename) = $_[0];
   local ($part2rm) = $_[1];
   
   if ($_filename =~ /$part2rm/){
      $_filename =~ s/$part2rm//;
      return (1, $_filename);  
   } else {
      return (0, $_filename);  
   }
} # purge_full_path


##############################################################################
##! @fn int starting_point_script
##! @brief Creates the top level TOC script that uses subscript references.
##! 
##! @param htree_file The prefix of the script file to be output.
##! @param which_script The key into the hash for whether we're creating a
##! "script_local" or "script_master" script. This had the appropriate file names.
##! @param which_purge_path The key into the hash for the appropriate purge
##! path depending upon whether or not this is a "local" or "master" version.
##! 
##! This knows about the top level entries and what they should be named.
##! The names have been added to the appropriate hash. It can generate the
##! master script that calls subscripts depending upon which version we're
##! doing.
##! 
##! The difference between a local and master version are the paths for links.
##! Normally, the other tools would handle this for me. However, because this
##! is starting out already as a nested version, it does not integrate well.
##! It is better to have this take care of all entries appropriately so that
##! the other tools only has to integrate the top level script (through
##! a master_nav file).
##! 
##! @ingroup tp_nav
##############################################################################
sub starting_point_script {
   local ($htree_file) = $_[0];
   local ($which_script) = $_[1];
   local ($which_purge_path) = $_[2];
   
   if (($htree_file =~ /(\.html)$/i) || ($htree_file =~ /(\.htm)$/i)){
      @_name = split (/\.htm/i, $htree_file, 2);
      $htree_file = $_name[0];
   }
   $htree_file .= ".script";

   if ((0) && ($htree_file =~ /\//)) {
      @_name = split (/\//, $htree_file);
      $_name[$#_name] = "tree_$_name[$#_name]";
      $htree_file = join ("\/", @_name);
   }
   
   print "The new output TOC script file is \"$htree_file\".\n";

   unless (open(OUT_SCRIPT, ">$htree_file")) {
      push (@file_errors, "Cannot open file \"$htree_file\".");
      return(0);
   }

   if ($which_purge_path =~ /master/) {
      # The master level doesn't need a top level
   } else {
      # $globe::top_most_level_title
      ($okay, $strip_p) = &purge_full_path ($globe::top_most_level{url}, $globe::nav_purge_path{$which_purge_path});
      print (OUT_SCRIPT "Item level=1 ");
      print (OUT_SCRIPT "image=nav_folderclosed.gif ");
      # print (OUT_SCRIPT "url=$strip_p ");
      print (OUT_SCRIPT "text=$globe::top_most_level{title}\r\n");
   }
   

   $i = 0;
   foreach $_in_file (@xscope::top_files){
      $globe::org{$_in_file}{title} =~ s/^[\s+]//;  # get rid of leading spaces
      if ($which_purge_path =~ /master/) {
         print (OUT_SCRIPT "Item level=1 ");
      } else {
         print (OUT_SCRIPT "Item level=2 ");
      }
      print (OUT_SCRIPT "image=nav_folderclosed.gif ");
      print (OUT_SCRIPT "subscript=$globe::org{$_in_file}{$which_script}.script ");
      print (OUT_SCRIPT "selected=YES ");
      print (OUT_SCRIPT "text=$globe::org{$_in_file}{title}\r\n");
      if (0) {
         print "$i $_in_file\n\t$globe::org{$_in_file}{title}\n";
      }
      $i++;
   }
   close (OUT_SCRIPT);

} # starting_point_script





##############################################################################
##! @fn int script_structure
##! @brief Traces through the data structure following the children and
##! creates entries into the script file for the TOC.
##! 
##! @param begin_file is the point to start in the data structure.
##! 
##! For every begin_file, this outputs entries into the already open
##! OUTSCRIPT file for each of its children.
##! 
##! Whenever a child is encountered, it is tested (using globe_found_files)
##! to see if it has already been processed. This prevents it looping forever.
##! 
##! If a given child has not been traced, this calls itself with that child.
##! 
##! It was deemed desireable to still print out children entries even if
##! we weren't going to trace them (because they already had been traced.)
##! 
##! @note This is a RECURSIVE routine. It stops when all children at
##! that level have had appropriate entries made and none have
##! been flagged to be traced.
##!
##! @bug $globe_found_files{$begin_file} is used in a global way. 
##! $globe::nav_purge_path is also a global hash. The key into the hash
##! ($globe_path_purge) was defined a level higher and used globally.
##! 
##! @ingroup tp_nav
##############################################################################
sub script_structure {
   local ($begin_file) = $_[0];
   # $globe_path_purge is global from calling routine, so we don't have
   # to remember it.
   local ($child);
   local ($strip_p) = "";
   local ($trace) = 0;
   local ($f);
   local ($_lev);
   
   if (0) {
      print "script_structure $begin_file\n$globe::org{$begin_file}{title}\n";
   } else {
      # purposely not terminated; shows that it is working
      print "$globe_file_cnt.";
   }
   
   
   for ($f = 0; $f <= $#{$globe::org{$begin_file}{child}}; $f++){
      ($okay, $strip_p) = &purge_full_path ($globe::org{$begin_file}{child}[$f]{url}, $globe::nav_purge_path{$globe_path_purge});
      
      # okay to still print out the children even if we're not going to 
      # trace them.
      #####
      # We were getting wrong levels; now we assure that children are always only
      # incremented one from the parent printing them out.
      #####

      print (OUT_SCRIPT "Item level=$globe::org{$begin_file}{child}[$f]{level} ");
      print (OUT_SCRIPT "image=nav_doc.gif ");
      print (OUT_SCRIPT "url=$strip_p,basefrm ");
      print (OUT_SCRIPT "selected=YES ");
      print (OUT_SCRIPT "text=$globe::org{$begin_file}{child}[$f]{text}\r\n");

      $child = $globe::org{$begin_file}{child}[$f]{url};
      if ($child =~ /\#/){
         @_name = split (/\#/, $child, 2);
         $child = $_name[0];
      }
      if (! (exists ($globe_found_files{$child}) )){
         $globe_found_files{$child} = 1;
         $globe_file_cnt++;
         #####
         # Recursive Call
         #####
         &script_structure ($child);
      }
      
   } # for each child

   return (1);

} # script_structure



##############################################################################
##! @fn int handle_index_tokens
##! @brief Traces through the data structure and outputs appropriate entries
##! for the index.
##! 
##! @param begin_file The starting point in the structure.
##! 
##! @note This is a RECURSIVE routine. It creates appropriate entries for each
##! of the children into the globe::index_info data structure, which later
##! gets output to a file.
##!
##! @bug Does a convoluted thing to get the path to be truly integrated
##! into the system. Also, it creates a helluva lot of entries for BSP (4022)
##! before any word-chunking might be considered.
##!
##! @ingroup tp_nav
##############################################################################
sub handle_index_tokens {  

   local ($begin_file) = $_[0];
   local ($child);
   local ($f);

   if (0) {
      print "handle_index_tokens $begin_file\n$globe::org{$begin_file}{title}\n";
   }
   
   HONEY_CHILE: for ($f = 0; $f <= $#{$globe::org{$begin_file}{child}}; $f++){
      $child = $globe::org{$begin_file}{child}[$f]{url};
      if ($child =~ /\#/){
         @_name = split (/\#/, $child, 2);
         $child = $_name[0];
      }
      if (0) {
         print "  child $globe::org{$begin_file}{child}[$f]{text}";
         print "  $globe::org{$begin_file}{child}[$f]{url}\n";
      }
      #####
      # 
      #####
      $path_temp = $globe::nav_purge_path{local};
      if ($globe::org{$begin_file}{child}[$f]{url} =~ /^\//) {
         # if the child begins with a slash, then remove the slash
         # from the path temp before combining the two.
         $path_temp = s/\/$//;
      }
      $path_temp .= $globe::org{$begin_file}{child}[$f]{url};
      ($okay, $strip_p) = &purge_full_path ($path_temp, $globe::nav_purge_path{"master"});
      
      $n_temp = $globe::org{$begin_file}{child}[$f]{text};

      $n_temp .= $globe::word_url_boundary;
      $n_temp .= "<a href=\"";
      $n_temp .= $strip_p;
      $n_temp .= "\" target=\"basefrm\">";

      #####
      # This is an attempt at efficiency.
      # It adds duplicates to the list, but they can be removed later,
      # rather than constantly looping within this already expensive
      # loop and recursive routine.
      #####
      push (@globe::index_info, $n_temp);
      
      if (! (exists ($globe_all_files{$child}) )) {
         $globe_all_files{$child} = 1;
         $globe_file_cnt++;
         #####
         # Recursive Call
         #####
         &handle_index_tokens ($child);
      }
      
   } # for each child
   return (1);
   
} # handle_index_tokens   



##############################################################################
##! @fn int html_file_update
##! @brief Traces through the data structure, removes tagged sections, and
##! inserts new tagged sections with navigation information and copyright
##! stuff.
##! 
##! @param begin_file The starting point in the structure.
##! 
##! @note This is a RECURSIVE routine. It opens each HTML file and writes it
##! back out.
##!
##! @ingroup tp_nav
##############################################################################
sub html_file_update {  
   local ($begin_file) = $_[0];
   local ($child);
   local ($f);

   if (1) {
      # print "Updating HTML file $begin_file\n   $globe::org{$begin_file}{title}\n";
      print "Updating HTML file $begin_file\n";
   }

   unless (open ( HTML_FILE, "$begin_file")) {
       push (@file_errors, "Cannot open file \"$begin_file\"");
       exit(1);
   }
   
   undef ($globe::entire_file);
   # read in entire file
   while (<HTML_FILE>){
      $globe::entire_file .= $_;
   }
   close (HTML_FILE);
   
   # Change the tags inside the file
   if (&change_nav){
      # it assumes $globe::entire_file
      # assume success for now
   }

   # Write out entire file after changing the tags inside the file.
   unless (open ( HTML_FILE, ">$begin_file")) {
       push (@file_errors, "Cannot open file \"$begin_file\"");
       return (0);
   }
   print (HTML_FILE "$globe::entire_file");
   close (HTML_FILE);
   
   for ($f = 0; $f <= $#{$globe::org{$begin_file}{child}}; $f++){
      $child = $globe::org{$begin_file}{child}[$f]{url};
      if ($child =~ /\#/){
         @_name = split (/\#/, $child, 2);
         $child = $_name[0];
      }
      if (0) {
         print "  lil chile $globe::org{$begin_file}{child}[$f]{text}";
         print "  $globe::org{$begin_file}{child}[$f]{url}\n";
      }
      
      if (! (exists ($globe_all_files{$child}) )) {
         $globe_all_files{$child} = 1;
         $globe_file_cnt++;
         #####
         # Recursive Call
         #####
         &html_file_update ($child);
      }
   } # for each child

   return (1);
   
} # html_file_update   

##############################################################################
##! @fn int change_nav
##! @brief Removes existing tags in the file and inserts new tags with the
##! content from the master files.
##! 
##! @return This updates the globe::entire_file with new information.
##! 
##! @note This is different than the change_nav in the voyant_nav.pl
##! program.
##!
##! @ingroup tp_nav
##############################################################################
sub change_nav {  
   if (0){
      print "... changing the navigation.\n";
   }
   
   #####
   # Make sure our input HTML file is well formed
   #####
   $globe::entire_file =~ s/\<\/Head\>/\<\/head\>/g;
   $globe::entire_file =~ s/\<\/HEAD\>/\<\/head\>/g;
   $globe::entire_file =~ s/\<Head\>/\<head\>/g;
   $globe::entire_file =~ s/\<HEAD\>/\<head\>/g;
   $globe::entire_file =~ s/\<\/Body/\<\/body/g;
   $globe::entire_file =~ s/\<\/BODY/\<\/body/g;
   $globe::entire_file =~ s/\<Body/\<body/g;
   $globe::entire_file =~ s/\<BODY/\<body/g;
   $globe::entire_file =~ s/\<Link rel\=/\<link rel\=/g;
   $globe::entire_file =~ s/\<Link Rel\=/\<link rel\=/g;
   $globe::entire_file =~ s/\<LINK REL\=/\<link rel\=/g;
   $globe::entire_file =~ s/\<Link/\<link/g;
   $globe::entire_file =~ s/\<LINK/\<link/g;
   
   
   # Remove any style sheets that might be hanging out.
   # get rid of any old tags.

   while ($globe::entire_file =~ /\<link/){
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              "<link", ">", 0, 1);
      if ($piece) {
         $globe::entire_file = join ("", $before, $after);
      }
   }
   while ($globe::entire_file =~ /\<body/){  
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              "<body", ">", 0, 1);
      if (1) {
         $globe::entire_file = join ("", $before, $after);
      }
   }
   if ($globe::entire_file =~ /\<\/head>/){  # insert a <body> tag
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              "</head", ">", 0, 1);
      if (1) {
         $globe::entire_file = join ("", $before, "</head><body>",$after);
      }
   }
   
   
   # Remove any style sheets that might be hanging out.
   # get rid of any old tags.
   &remove_voyant_tags;

   
   if (0) {
      print "Finishing up here... premature.\n";
      # print $globe::entire_file;
      print (OUT_HTML "$globe::entire_file");
      close (OUT_HTML);
      system ("cp $out_file $in_file");
      return (0);
   }

   
   ##################
   # voyant header needs to go before </head>
   ##################
   $not_critical = 1;
   $reinsert_tags = 0;
   $def_type = "header";
   if (! &file_chunk_change ($def_type, $not_critical, "</head>", 0, $reinsert_tags, $in_file) ) {
      # Could not put proper chunk in; write out what we have an abort.
      # print (OUT_HTML "$globe::entire_file");
      close (OUT_HTML);
      if (0){ # debug
          print "glenn3\n$globe::entire_file";
          exit(1);
      } # debug
      return(0);
   } # voyant_header

   if (0){ # debug
      print "glenn4\n$globe::entire_file";
      print "oh $def_type $globe::m_info{$def_type}\noh again\n";
      exit(1);
   } # debug
 

   ##################
   # voyant footer needs to go before </body>
   ##################
   $not_critical = 1;
   $reinsert_tags = 0;
   $def_type = "footer";
   if ($globe::entire_file =~ /\<\/body\>/) { 
      &file_chunk_change ($def_type, $not_critical, "<\/body>", 0, $reinsert_tags, $in_file);
      # Could not put proper chunk in; write out what we have an abort.
   } else {
      # Does not have proper body
      push (@file_errors, "ERROR: Does not have </body> in $in_file.");
      return (0);
   } # voyant footer


   ##################
   # voyant common top navigation bar should go after <body>
   # Do AFTER FM or Doxygen navigation
   ##################
   $not_critical = 0;
   $reinsert_tags = 0;
   $def_type = "nav_common";

   if (! &file_chunk_change ($def_type, $not_critical, "<body>", 1, $reinsert_tags, $in_file) ) {
      # Could not put proper chunk in; 
      # Not necessarily a problem.
   } # Common Navigation

   if (0){ # debug
      print "glenn4\n$globe::entire_file";
      print "oh $def_type $globe::m_info{$def_type}\noh again\n";
      exit(1);
   } # debug

   
   
   ##################
   # Handle the PDF file names
   ##################
   $temp = "$globe::rel_path_to_start_point$globe::pdf_dir$globe::master_order_pdf[$globe::master_order_key]";
   $globe::entire_file =~ s/$globe::this_pdf_flag/$temp/g;

   ##################
   # Handle the relative path
   # glenn
   ##################
   # $globe::rel_path_to_start_point_def = "##rp2start##";
   # $globe::rel_path_to_start_point = "./";
   if (0) {
      print "getting rid of path $begin_file\n";
   }
   $_t_path = &purge_full_path ($begin_file, $globe::nav_purge_path{master});
   @_a_path = split (/\//, $_t_path);
   pop (@_a_path); # because last element is file name
   $globe::rel_path_to_start_point = "";
   for ($t=0; $t<=$#_a_path; $t++) {
      $globe::rel_path_to_start_point .= "..\/";
   }
   undef (@_a_path);
   undef ($_t_path);
   if (0) {
      print "new $_t_path\n$globe::rel_path_to_start_point\n";
   }

   $globe::entire_file =~ s/$globe::rel_path_to_start_point_def/$globe::rel_path_to_start_point/g;
   
   ##################
   # Handle the manual or group names
   ##################
   $temp = $globe::master_order_title[$globe::master_order_key];
   $temp =~ s/[\s]+/\&nbsp\;/g;
   
   if ($globe::entire_file =~ /$globe::xmanual/){
      $globe::entire_file =~ s/$globe::xmanual/$temp/g;
   } 
   if ($globe::entire_file =~ /$globe::xgroup/){
      $globe::entire_file =~ s/$globe::xgroup/$temp/g;
   }
   $globe::entire_file =~ s/$globe::fix_path_flag/$globe::fix_path_to/g;
   
   #####
   # Take care of current file name which might be in header,
   # as specified in template and needed in tree applet.
   #####
   if ($globe::entire_file =~ /$globe::m_var{define}{curr2}/){
      $globe::entire_file =~ s/$globe::m_var{define}{curr2}/$rel_to_file/g;
   } 
   

   ##################
   # Handle the document number
   ##################
   if ($globe::entire_file =~ /$globe::voy_variable_doc_num/){
      $globe::entire_file =~ s/$globe::voy_variable_doc_num/$globe::m_info{_array_order}[$globe::master_order_key]{num}/;
   } 
   
   
   #################################
   # Remove certain specified HTML tags
   #################################
   if ((@globe::voy_html_zap < 1) || ($globe::voy_html_zap[0] =~ /none/)){
      # Do nothing
      # print "$in_file has no tags to zap.\n";
   } else {
      # Remove the tags in question
      for ($i=0; $i<@globe::voy_html_zap; $i++){
         # Determine start and Stop tags
         @zap_tag = split ( /\,/, $globe::voy_html_zap[$i]);
         # clean up tag
         for ($j=0; $j <@zap_tag; $j++){
            $zap_tag[$j] =~ s/^[\s]*\"//;
            $zap_tag[$j] =~ s/\"[\s]*$//;
            # print "Zap this =$zap_tag[$j]\n";
         }
         $t_cnt=0;
         while ($globe::entire_file =~ /$zap_tag[0]/i){  # case insensitive test
             $t_cnt++;
             print "$t_cnt: Removing everything between \"$zap_tag[0]\" and \"$zap_tag[1]\"\n";
             # First tag part A
             @pre_chunks = split ( /$zap_tag[0]/, $globe::entire_file, 2);
             # First tag part B
             @post_chunks = split ( /$zap_tag[1]/, $pre_chunks[1], 2);
             if (@zap_tag > 2){
                # Second tag
                # Hit the first occurrence of the ending tag
                $post_chunks[1] =~ s/$zap_tag[2]//i; # do it only once, case insensitive
             }
             # rebuild
             $globe::entire_file = join ("", $pre_chunks[0], $post_chunks[1]);
         } # while it still is infested with the tag
      } # for all zap tags
      # print "$in_file was tested for tags to zap.\n";
   } # if zap tags to remove
   undef(@pre_chunks);
   undef(@post_chunks);
   #################################

} # change_nav



##############################################################################
##! @fn int remove_voyant_tags
##! 
##! 
##! 
##############################################################################
sub remove_voyant_tags {

   # Remove any style sheets that might be hanging out.
   # get rid of any old tags.
   while ($globe::entire_file =~ $globe::m_define{header}[0]){
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              $globe::m_define{header}[0], 
                              $globe::m_define{header}[1],
                              0, 1);
      if ($piece) {
         $globe::entire_file = join ("", $before, $after);
      }
   }
   while ($globe::entire_file =~ $globe::m_define{nav_book}[0]){
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              $globe::m_define{nav_book}[0], 
                              $globe::m_define{nav_book}[1],
                              0, 1);
      if ($piece) {
         $globe::entire_file = join ("", $before, $after);
      }
   }
   while ($globe::entire_file =~ $globe::m_define{nav_common}[0]){
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              $globe::m_define{nav_common}[0], 
                              $globe::m_define{nav_common}[1],
                              0, 1);
      if ($piece) {
         $globe::entire_file = join ("", $before, $after);
      }
   }
   while ($globe::entire_file =~ $globe::m_define{footer}[0]){
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              $globe::m_define{footer}[0], 
                              $globe::m_define{footer}[1],
                              0, 1);
      if ($piece) {
         $globe::entire_file = join ("", $before, $after);
      }
   }



} # remove_voyant_tags




##############################################################################
##! @fn int file_chunk_change
##! @brief Replaces the information for the given chunk. (Not complete).
##! 
##! @param _type_define The key into the m_define and m_info hashes.
##! @param _criticality specifies whether or not to record error messages.
##! @param _where_flag The tag where this is supposed to go.
##! @param _after if 0, set before, otherwise after
##! @param _in_file Name of file being worked on.
##!
##! @retval success 1 if found tags, 0 if missing.
##! 
##! @bug Uses the $globe::entire_file
##! @ingroup tp_toc
##############################################################################
sub file_chunk_change {
   $_type_define = $_[0];
   $_criticality = $_[1];
   $_where_flag = $_[2];
   $_after = $_[3];
   $_insert_tags = $_[4];  # pass through
   $_in_file = $_[5];

   if (0) {
      print "file_chunk_change $_type_define $globe::m_define{$_type_define}[0]\n$_in_file\n";
   }
   if (( $globe::entire_file =~ /$globe::m_define{$_type_define}[1]/ ) && ( $globe::entire_file =~ /$globe::m_define{$_type_define}[0]/ ) ) {
      ($okay, $globe::entire_file) = &globe::replace_tag_chunk ( $globe::entire_file,
                $globe::m_define{$_type_define}[0], 
                $globe::m_define{$_type_define}[1], 
                $globe::m_info{$_type_define}, $_insert_tags, $_criticality);
      if (!$okay) {
         push (@file_errors, "$_type_define definition messed up in $in_file.");
      }
      if (0){
         print "file_chunk_change ok=$okay\n";
      }

   } else {
      # Figure out where the $_where_flag should go
      if ( $globe::entire_file =~ /$_where_flag/ ) {
         @pre_chunks = split ( /$_where_flag/, $globe::entire_file, 2);
         if ($_after) {
            $globe::entire_file = sprintf("%s%s%s%s", 
               $pre_chunks[0], 
               $_where_flag,
               $globe::m_info{$_type_define}, 
               $pre_chunks[1]);
         } else {
            $globe::entire_file = sprintf("%s%s%s%s", 
               $pre_chunks[0], 
               $globe::m_info{$_type_define}, 
               $_where_flag,
               $pre_chunks[1]);
         }
      } else {
         push (@file_errors, "WARNING: No $_where_flag defined in $_in_file.");
         # Do only if no head/body tags
         print "WARNING: No $_where_flag defined.\n";
         return (0);
      } # figuring out where it should go
   }
   return (1);

} # file_chunk_change






##############################################################################
##! @fn int html_exist
##! @brief Tests that the URL to a file exists.
##!
##! @param test_file File to test to see if it exists.
##! 
##! Discovered that doxygen's installdox does not update the tree.js files,
##! which leads to entries in the TOC that 404. Test it and return an 
##! appropriate response.
##!
##! @ingroup tp_nav
##############################################################################
sub html_exist {
   # Test the damn URL to make sure the file even exists.
   $test_file = $_[0];

   unless (open(TEST, "$globe::path$test_file")) {
      if (0){
         print "$globe::path$test_file doesn't exist.\n";
      }
      return(0);
   }
   close (TEST);
   return (1);


} # html_exist


##############################################################################
##! @fn int END
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub END {
   &globe::do_errors;
   undef ($_file_list);     # "_file_list";
   undef ($_index_file); # = "_index_list";
   undef ($in_file);     # "";
   undef ($f_type);     # "htm";

   
   ##############################################################################
   ## Memory clean-up.
   ##############################################################################
   if ($no_scope_file > 0){
      &xscope::memory_clean_up();
      &globe::memory_clean_up();
   }

   if (@file_errors) {
      print "\n============  Summary of errors =================================b\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
   }

   print "\n============  Finished html_look_integrate.pl =================================\n";
} # END

