#!/usr/local/bin/perl
package scheck;
##############################################################################
##! @file
##! @brief Checks the generated SAPI files for completeness.
##! 
##! This uses the "sapi.pm" file that is generated by the "vproto.pl" tool.
##! It can be named something else, but is the $scope_pm input variable
##! to this program.
##! 
##! The "sapi.pm" file was intended for use with the xhelp tool for reducing noise.
##! It has several data structures defined and then undefined. They were defined
##! because they could have been useful in locating the proper methods and
##! doxygen comments in the code. (It turns out that the ingroup is more
##! effective at that.)
##! 
##! Those data structures were undefined (with "if (1)" enclosing an
##! "undef") so that they wouldn't result in unnecessary grep operations.
##! 
##! Those data structures, however, are useful for testing the integrity of
##! the system in terms of missing commands.
##! 
##! This tool expects you to turn off ("if (0)") these undefinition within the 
##! the sapi.pm file. Run this tool and look at the output.
##!
##! @ingroup tp_tools tp_xhelp
##!
##! @author Glenn C. Maxey
##
##    $Id: sapi_check.pl,v 1.2 2002/11/20 01:49:13 gmaxe Exp $
##
## Copyright 2002 Created by Voyant Technologies, Inc., Westminster, Colorado, USA.
##
## Revision Information:
##
## $Log: sapi_check.pl,v $
## Revision 1.2  2002/11/20 01:49:13  gmaxe
## Minor changes to exit codes.
##
## Revision 1.1  2002/04/12 22:52:25  gmaxe
## New tool to check on whether or not all SAPI commands are present.
## Uses the files generated by the vproto_gen.pl tool.
##
##
##############################################################################

##############################################################################
##! @fn int BEGIN
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_xhelp
##############################################################################
sub BEGIN {
   print "\n============  Starting sapi_check.pl ==================================\n";
   $_file_list = "_file_list";
   $in_file = "";
   $root_path = "./";
   $scope_pm = "";
   $no_scope_file = 0;
   $arg_cnt=0;
   
   push (@INC, `pwd`);
   push (@INC, '../perl');
   if (0){
      print (@INC, "\n");
   }
   
   # Get scope files if there is one.
   if (@ARGV > $arg_cnt) {
      $scope_pm = @ARGV[$arg_cnt];
      
      unless (open ( IN_LIST, $scope_pm)) {
         push (@file_errors, "Cannot open file \"$scope_pm\"");
         $no_scope_file++;
      }
      close (IN_LIST);
      push (@INC, $scope_pm);
      
      if (!@file_errors) {
         #####
         # All global variables are defined in the following file
         #####
         require $scope_pm;
   
         if (&xscope::declare_variables()) {
            print "Variables initialized from $scope_pm.\n";
         } else {
            push (@file_errors, "Could not initialize variables from $scope_pm.");
         }
      } # if not @file_errors
   } else {
      push (@file_errors, "Need to provide a scope file as first argument.");
   } # if 1 or more arguments
   $arg_cnt++;

  if (0) { # not used
   # Get path to code files 
   if (@ARGV > $arg_cnt) {
      $root_path = @ARGV[$arg_cnt];
      if ($root_path =~ /\/$/) {
         print "The path specified is $root_path\n";
      } else {
         push (@file_errors, "The input argument \"$root_path\" requires a forward slash (\/) at the end.\n");
      }
   } # if 1 or more arguments
   $arg_cnt++;
  } # not used

   if (@ARGV > $arg_cnt) {
      $gen_proto = @ARGV[$arg_cnt];
      print "The prototype input file is $gen_proto \n";
   } else {
      push (@file_errors, "Need to have the input prototype file.");
   } # if 3 or more arguments
   $arg_cnt++;


   if (@ARGV > $arg_cnt) {
      $gen_class_dox = @ARGV[$arg_cnt];
      print "The class methods with doxygen comments is $gen_class_dox \n";
   } else {
      push (@file_errors, "Need to have the input class doxygen file.");
   } # if 3 or more arguments
   $arg_cnt++;

   if (@ARGV > $arg_cnt) {
      $gen_error_out = @ARGV[$arg_cnt];
      print "The output file will be $gen_error_out \n";
   } else {
      push (@file_errors, "Need to have an output file for the results.");
   } # if 3 or more arguments
   $arg_cnt++;


} # BEGIN
   
   
   

##############################################################################
##! @fn int main
##! @brief The main program.
##!
##! @param file_errors If file errors are encountered before the main
##! routine is hit, it prints them out and exits. 
##!
##! @return None.
##!
##! If the needed data structures in $scope_pm have not been enabled,
##! this program errors out immediately.
##! 
##! Controls the program flow. It creates a temporary file with the
##! error results.
##!
##! @bug None
##! @ingroup tp_xhelp
##############################################################################
# sub main {
{
   if (!@xscope::x_names) {
      print "\nThe \"$scope_pm\" file should have been generated with \"\@x_names\" completed.\n";
      print "Then \"\@x_names\" was probably turned off inside of an \"if (1) {\" statement\"\n";
      print "that \"undef(\@x_names)\".\n\n";
      print "If this is the case, edit \"$scope_pm\" and change turn this off with\n";
      print "\"if (0) {\"\n";
      push (@file_errors, "Edit \"$scope_pm\" to enable \"\@x_names\".");
   }
   if (!@xscope::needed_in) {
      print "\nThe \"$scope_pm\" file should have been generated with \"\@needed_in\" completed.\n";
      print "Then \"\@needed_in\" was probably turned off inside of an \"if (1) {\" statement\"\n";
      print "that \"undef(\@needed_in)\".\n";
      print "\nIf this is the case, edit \"$scope_pm\" and change turn this off with\n";
      print "\"if (0) {\"\n";
      push (@file_errors, "Edit \"$scope_pm\" to enable \"\@needed_in\".");
   }
   ##############################################################################
   ## Start of Main Program
   ##############################################################################
   if (@file_errors) {
      # Makes no sense to go on if input parameters are off.
      print "\n============  Summary of errors ==================================\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
      &scheck::using_scheck();
      exit(1);
   }
   
   if ((0) && (!(&read_and_check($gen_proto, "proto")))){
      print "$gen_proto failed some things.\n";
   } 
   if ((1) && (!(&read_and_check($gen_class_dox, "class")))){
      print "$gen_class_dox failed some things.\n";
   }
   
   if (1) {
      &write_output_list($gen_error_out);
   }
   
   ##############################################################################
   ## Spit out errors
   ##############################################################################
   if (@file_errors) {
      print "\n============  Summary of errors ==================================\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
   }

   ##############################################################################
   ## End of most of work
   ##############################################################################
   exit(0);
}

##############################################################################
##! @brief Reads in the entire generated file and kicks off a routine to
##! to the actual test.
##! 
##! @param src_file The file to open and read in.
##! @param which_test Specifies which other test to run. Presently either
##! "proto" or "class".
##! 
##! @bug When reading in the file and the later test is "proto", this doesn't
##! bother to read in comment lines.
##############################################################################
sub read_and_check {
   local ($src_file) = $_[0];
   local ($which_test) = $_[1];
   undef ($globe::entire_file);
   unless (open ( IN_SRC, $src_file)) {
      print "Cannot open file \"$src_file\"\n";
      push (@file_errors, "Cannot open file \"$src_file\"\n");
      $ret_val = 0;
      return (0);
   }
   while (<IN_SRC>) {
      if ($which_test =~ /proto/){
         if (($_ !~ /\/\//) && ($_ !~ /\/\*/)){
            # Only read in non-comment lines.
            $globe::entire_file .= $_;
         }
      } else {
         $globe::entire_file .= $_;
      }
   }
   if ($which_test =~ /proto/){
      $ret_val = &check_prototype($src_file);
   } elsif ($which_test =~ /class/){
      $ret_val = &check_class_dox($src_file);
   } else {
      print "The test specified, \"$which_test\" is invalid.\n";
      $ret_val = 0;
   }
   close (IN_SRC);
   if (0) {
      print $globe::entire_file;
      exit(1);
   }
   return ($ret_val);
} # read_and_check

##############################################################################
##! @brief Checks the file that contains prototypes against the data structure
##! created in an earlier process and part of the $scope_pm file.
##! 
##! @param in_file A notation of the file that the information was checked
##! against.
##! 
##! This loops through the (enabled) data structure to make sure that those
##! keywords appear somewhere in the file. It is assumed that their occurrence
##! means that they have been defined.
##! 
##! If the keyword from the data structure does not appear in the file,
##! this logs and error to @file_error for later output.
##! Not all errors are true errors.
##!
##! This checks for both the occurrence of the expected SAPI command (originally
##! defined in vfuncs.inp).
##############################################################################
sub check_prototype {
   local ($in_file) = $_[0];
   push (@file_errors, "$in_file: Checking for prototypes in $in_file...");
   print "======\nChecking for prototypes in $in_file\n=====\n";
   foreach $item (@xscope::x_names){
      if ($globe::entire_file !~ /$item/){
         if (1) {
            print "$in_file: Missing prototype for \"$item\".\n";
         }
         push (@file_errors, "$in_file: Missing prototype for \"$item\".");
      }
      if (0) {
         print "$item\n";
      }
   }
   print "======\n\n";
   return (1);
} # check_prototype

##############################################################################
##! @brief Checks the file that contains class definitions and doxygen
##! comments and compares against the data structure
##! created in an earlier process and part of the $scope_pm file.
##! 
##! @param in_file A notation of the file that the information was checked
##! against.
##! 
##! This loops through the (enabled) data structure to make sure that those
##! keywords appear somewhere in the file. It is assumed that their occurrence
##! means that they have been defined.
##! 
##! If the keyword from the data structure does not appear in the file,
##! this logs and error to @file_error for later output.
##! Not all errors are true errors.
##!
##! This checks for both the occurrence of the expected SAPI command (originally
##! defined in vfuncs.inp) and the associated class method that implements
##! the command.
##############################################################################
sub check_class_dox {
   local ($in_file) = $_[0];
   push (@file_errors, "\n\n>>>>> Checking for doxygen comment blocks in $in_file...");
   print "======\nChecking for doxygen comment blocks in $in_file\n=====\n";
   foreach $item (@xscope::x_names){
      if ($globe::entire_file !~ /$item/){
         if (1) {
            print "$in_file: Missing doxygen comment block for \"$item\".\n";
         }
         push (@file_errors, "$in_file: Missing doxygen comment block for \"$item\".");
      }
      if (0) {
         print "$item\n";
      }
   }
   print "======\n\n";


   push (@file_errors, "\n\n>>>>> Checking for C++ class definitions in $in_file...");
   print "======\nChecking for C++ class definitions in $in_file\n=====\n";
   foreach $item (@xscope::needed_in){
      if ($globe::entire_file !~ /$item/){
         if (1) {
            print "$in_file: Missing C++ class definitions for \"$item\".\n";
         }
         push (@file_errors, "$in_file: Missing C++ class definitions for \"$item\".");
      }
      if (0) {
         print "$item\n";
      }
   }
   print "======\n\n";


   return (1);
} # check_class_dox


##############################################################################
##! @fn int write_output_list
##! @brief Creates a file lists all of the potential errors discovered in
##! the generated files.
##!
##! @param file_out The name of the file to output.
##!
##! @ingroup tp_xhelp
##############################################################################
sub write_output_list {
   $file_out = $_[0];
   print "Writing the $file_out output file.\n";
   
   unless (open ( OUT_FILE, ">$file_out")) {
      push (@file_errors, "Cannot open file \"$file_out\"\n");
      print "Cannot open file \"$file_out\"\n";
   }
   print (OUT_FILE "This file contains a list of potential errors in the\n");
   print (OUT_FILE "SAPI generation process.\n\n");
   print (OUT_FILE "Not all of them are true errors, because it is known\n");
   print (OUT_FILE "that certain functions beginning with \":\" or \"__\"\n");
   print (OUT_FILE "are NOT really errors.\n\n");
   print (OUT_FILE "This list is just to serve as a starting point for\n");
   print (OUT_FILE "potentially missing items and maybe the class that\n");
   print (OUT_FILE "is supposed to define it.\n\n");
   print (OUT_FILE "==============================================================\n\n");

   for ($i=0; $i<@file_errors; $i++){
      print (OUT_FILE "$i = $file_errors[$i]\n");
   }
   close (OUT_FILE);
} # write_output_list


##############################################################################
##! @fn int using_scheck
##! @brief What to do when no arguments are given.
##! @param None
##! @return None
##! 
##! @bug None
##! @ingroup tp_xhelp
##############################################################################
sub using_scheck {
   print "============  Syntax  ==================================\n";
   print "sapi_check.pl <input scope package> <input prototype> <input class> <output file path \& name> \n\n";
   print "- <input scope package> is file path and name to a perl package that ";
   print "has lists of functions to be included and to be excluded.\n\n";
   print "- <input prototype> is file path and name the generated sapi prototype ";
   print "file that came from the vproto_gen.pl tool.\n\n";
   print "- <input prototype> is file path and name the generated sapi class ";
   print "file that came from the vproto_gen.pl tool; it has class member function ";
   print "definitions and their associated doxygen comments.\n\n";
   print "- <output file path \& name> is file path and name of error file ";
   print "of things that might be missing. \n\n";
   print "============  Terminate  ==================================\n";
   print "Terminating sapi_check.pl without doing anything.\n";
   return(1);
}




##############################################################################
##! @fn int END
##! @brief Code to execute when terminated nicely.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_xhelp
##############################################################################
sub END {
   undef ($_file_list); # "_file_list";
   undef ($in_file ); #  "";
   undef ($root_path ); #  "./";
   undef (%file_hash);

   
   ##############################################################################
   ## Memory clean-up.
   ## Perform this if it was defined.
   ##############################################################################
   if ($no_scope_file > 0){
      &xscope::memory_clean_up();
   }
   
   print "\n============  Finished sapi_check.pl ==================================\n";
   
} # END
