#!/usr/local/bin/perl
##############################################################################
##
##    $Id: voyant_indexer.pl,v 1.29 2002/12/24 15:05:54 gmaxe Exp $
##
##! @file
##! @brief Creates a comprehensive index from temporary index files that
##! were generated from another program.
##!
##! @param $globe::path location to find the index files. 
##! The name should be terminated with a slash (\). 
##! Although the directory_name is the first command line parameter,
##! the true input are the index files contained within that directory.
##! Index files must begin with "index_". 
##!
##! @param $globe::master_tree_file [optional path and] filename for the HTML file
##! to use as a template for all index files to be generated.
##! Ideally, this should contain navigation tools to get between the
##! generated index files [a-z] and other parts of the system, such as
##! the table of contents. This file has several specially flagged
##! HTML comment sections that are required.
##!
##! @param $globe::ignore_terms_file [optional path and] filename for a text file that
##! contains words to ignore in the word-chunking process.
##!
##! @return This creates a series of HTML files that begin with "m_idx_".
##! Generally, what follows in the name is the first character of the
##! first word within the file. All index entries beginning with that character
##! are in that file. These files are created in $globe::path.
##! 
##! <ol><li>The $globe::master_tree_file is viewed first to make sure that it has all
##! required information.</li>
##! 
##! <li>This program issues a system call to create a list of candidate input
##! index_ files in the $globe::path directory.</li>
##! 
##! <li> Then it steps through each of
##! those files and concatentates their input into $globe::master_raw.</li>
##!
##! <li> $globe::master_raw is turned into a hash table $idx_struct. The
##! key into the hash table is the index entry. What gets stored is both the
##! display text and an array of URLs.</li><ul>
##!
##! <li>$entry = compacted, clean, lower-case key from display text for sorting. </li>
##! <li>$subentry = compacted, clean, lower-case key from display text for sorting.</li>
##! <li>$idx_struct{$entry}{display} = display text.</li>
##! <li>$idx_struct{$entry}{url}[] = array of URL's for the $entry.</li>
##! <li>$idx_struct{$entry}{sub}{$subentry}{display} = display text
##!              for the $entry's $subentry.</li>
##! <li>$idx_struct{$entry}{sub}{$subentry}{url}[] = array of URL's 
##!              for the $entry's $subentry.</li></ul>
##!  
##! 
##! <li>Word-chunking is performed on each element in the $globe::master_index. 
##! Natural boundaries (spaces, dashes, underscores, changes in case in the middle of a word) 
##! are used to create additional two-level index entries that contain the word-chunk 
##! followed by where it came from. The $globe::ignore_terms_file is used to
##! eliminate unuseful word-chunked entries (such as "the", "a", "to", etc.)
##! The additional useful entries are appended to
##! the contents of the hash $globe::master_raw using the $globe::division_mult_entry
##! separator only if the new entry is not a duplicate.
##! Word-chunking is particular useful for API documentation so that the reader
##! does not have to remember the exact name of a code item in order to find
##! it. An initial index token of "api_GetMovie-list" could be found not just under
##! its name in the "A's", but under "get", "movie", and "list".</li>
##! 
##! <li>The expanded list is sorted.</li>
##! 
##! <li>Exact duplicates in terms of index entry and URL are removed from the
##! $globe::master_raw hash.</li>
##! 
##! <li>The sorted list is output to a series of m_idx_ files.
##! New m_idx_ files are created whenever a new character starts a word in the list.</li>
##! 
##! <li>When an index entry is referenced by multiple URLs, the additional references
##! appear in the output as small document icons next to the first reference in 
##! plain text.</li>
##! </ol> 
##!
##! More information about the true input "index_" files.
##! These files are an unsorted running list of index tokens that were 
##! extracted from the HTML files in a directory. The tokens have two 
##! parts: the index entry and its URL. The separator is :,: (($globe::word_url_boundary).
##! Additionally, the index entry can have two levels. In such
##! cases, the seperator is :;: ($globe::word_c_boundary).
##! Finally, a given index entry can represent multiple references or URLs.
##! In such cases, the multiple entries are separated by :;;;: ($globe::division_mult_entry)
##! 
##! If the separators are changed in the generator program (voyant_nav.pl),
##! they need to be changed here, too. The variable names in both programs are the same.
##! The separators themselves were chosen because they were deemed never to occur 
##! in an index entry or URL and aren't Perl special characters.
##! 
##! More information about the master_file. Aside from serving as a template for
##! all generated index files, this file chunks information using specially tagged
##! HTML comments in order to simplify locating where generated information is to
##! be placed. In addition, some tags contain information critical to the proper operation
##! of the indexer.
##! 
##! The critical tags given by Perl variable and HTML syntax are:
##! <ul><li>($globe::m_define{order}[0]) = "< !-- begin voy_order --"</li>
##! <li>($globe::m_define{order}[1]) = "!-- end voy_order -->"</li>
##! <li>($globe::m_define{structure}[0]) = "< !-- begin voy_structure -->"</li>
##! <li>($globe::m_define{structure}[1]) = "< !-- end voy_structure -->"</li>
##! <li>($globe::m_define{topic_list}[0]) = "< !-- begin voy_topic_list -->"</li>
##! <li>($globe::m_define{topic_list}[1]) = "< !-- end voy_topic_list -->"</li></ul>
##! 
##! The HTML syntax can be changed. However, the voy order sections are identical
##! for various programs and their master_files which facilitates propogating information.
##! 
##! @bug This does not support index entries that might be or have Perl special
##! characters. These are often eliminated early in the process.
##!
##! The input index_ files cannot have ".htm" as part of the name.
##! This assumes that input information was of the proper format with an
##! index entry, $globe::word_url_boundary, and URL. If any of the input index_
##! files did not have this, this can cause problems.
##!
##! Rather than passing in variables which can create copies in memory, many items
##! use global variables that are defined in globe.pm. When a
##! variable is known to be global, its name begins with "$globe::". The intent is
##! to facilitate maintenance by having all user-defined tags in one place outside 
##! of the program.
##!
##! Many debug statements are left in the code, although commented out or programmed
##! out with if (0){...}. On occassion, a statement is copied, commented out, and then
##! the copy modified in order to keep old techniques around while verifying new techniques.
##! Old techniques were not always purged once the new one worked.
##!
##! @ingroup tp_tools tp_idx
##!
##! @author Glenn C. Maxey
##
## 2002 Created by Voyant Technologies, Inc., Westminster, Colorado, USA.
##
## Permission to use, copy, modify, and distribute this software and its 
## documentation under the terms of the GNU General Public License is hereby 
## granted. No representations are made about the suitability of this software 
## for any purpose. It is provided "as is" without express or implied warranty. 
## See the GNU General Public License (http://www.gnu.org/copyleft/gpl.html) 
## for more details.
## 
## Documents produced by this script are derivative works derived from the 
## input used in their production; they are not affected by this license.
##
##    Revision Information:
##
##       $Log: voyant_indexer.pl,v $
##       Revision 1.29  2002/12/24 15:05:54  gmaxe
##       New name and tweaks to add full paths.
##
##       Revision 1.28  2002/11/20 15:23:05  gmaxe
##       Added exit codes so that wrapper scripts can catch errors properly.
##
##       Revision 1.27  2002/07/26 18:56:11  gmaxe
##       Got rid of old definitions and migrated new structures into all;
##       enhanced xhelp output file names and format; now everything is
##       alphebetized.
##
##       Revision 1.26  2002/07/22 21:51:31  gmaxe
##       Created separate configuration file; allow for two inputs to
##       the voyant_latex.pl. commented out old variables in globe.pm.
##
##       Revision 1.25  2002/07/09 21:37:27  gmaxe
##       Changed the data structure to make maintenance, input, and output easier.
##       Improved the efficiency of the word chunking, html output, script output,
##       and duplicate entries. It handles multiple URLs for a given index entry
##       better by putting it into an array inside the hash. However, the
##       elimination of ignore terms is still expensive.
##
##       Revision 1.24  2002/07/03 15:53:58  gmaxe
##       Commented out ignore_terms loops within expensive word-chunking loops.
##       Ignore terms are dealt with later. Reduces time by over 50%.
##
##
##############################################################################


##############################################################################
##! @fn int BEGIN
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub BEGIN {
   print "\n============  Starting voyant_indexer.pl ==================================\n";

   $_index_file_list = "_index_file_list";
   $in_file = "";
   $f_type = "index_";
   $_arg_inc = 0;

   $no_scope_file = 0;
   $scope_pm = "globe.pm";  # first time through; other scope stuff passed in.


   push (@INC, `pwd`);
   push (@INC, '/rtfm/techpubs/perl');
   if (0){
      print (@INC, "\n");
   }
   #####
   # All global variables are defined in the following file
   #####
   unless (open ( IN_LIST, $scope_pm)) {
      $scope_pm = "/rtfm/techpubs/perl/$scope_pm";
      unless (open ( IN_LIST, $scope_pm)) {
         push (@file_errors, "Cannot open file \"$scope_pm\"");
         $no_scope_file++;
      }
   }
   close (IN_LIST);
   push (@INC, $scope_pm);
   if (!@file_errors) {
      #####
      # All global variables are defined in the following file
      #####
      require $scope_pm;

      if (&globe::declare_variables()) {
         print "Variables initialized from $scope_pm.\n";
      } else {
         push (@file_errors, "Could not initialize variables from $scope_pm.\n");
      }
   } # if not @file_errors


   # Get path if there is one.
   if (@ARGV > $_arg_inc) {
      $globe::path = @ARGV[$_arg_inc];
      if ($globe::path =~ /\//) {
         print "The path specified is $globe::path\n";
      
         @path_chunks = split ( /\//, $globe::path);
         if (@path_chunks < 2) {
            $globe::rel_path_to_start_point = "./";
            $globe::pdf_name_from_dir = "book_cit_dtoss.pdf";
         } else {
            $globe::rel_path_to_start_point = "../";
            $globe::pdf_name_from_dir = "$path_chunks[@path_chunks - 1].pdf";
         }
      } else {
         push (@file_errors, "The input argument \"$globe::path\" requires a forward slash (\/) at the end.\n");
      }
   } else {
      push (@file_errors, "ERROR: root path is required.");
   } # if 1 or more arguments
   $_arg_inc++;
   
   if (@ARGV > $_arg_inc) {
      $globe::master_tree_file = @ARGV[$_arg_inc];
      print "The master file is $globe::master_tree_file\n";
      unless (open ( IN_MASTER, "$globe::master_tree_file")) {
         push (@file_errors, "Cannot open file \"$globe::master_tree_file\".");
      }
   } # if 2 or more arguments
   $_arg_inc++;

   if (@ARGV > $_arg_inc) {
      $globe::ignore_terms_file = @ARGV[$_arg_inc];
      print "The ignore terms file is $globe::ignore_terms_file\n";
      if (open ( IN_IGNORE, "$globe::ignore_terms_file")) {
         while (<IN_IGNORE>){
            $temp = $_;
            $temp =~ s/\n//;
            push (@globe::ignore_terms, $temp);
         }
         if (0) {
            print "Terms to ignore\n";
            foreach $term (@globe::ignore_terms){
               print "$term\n";
            }
         }
      } else {
         push (@file_errors, "Cannot open file \"$globe::master_tree_file\".");
      }
   } # if 3 or more arguments
   $_arg_inc++;

   if (@ARGV > $_arg_inc) {
      $globe::word_chunk = 1;
      $globe::word_chunk = @ARGV[$_arg_inc];
      if ($globe::word_chunk =~ /^no/i){
         $globe::word_chunk = 0;
      } else {
         $globe::word_chunk = 1;
      }
   } else {
      $globe::word_chunk = 1;
   } # if 3 or more arguments
   print "The flag for word-chunking is $globe::word_chunk; 1, do; 0, don't.\n";
   $_arg_inc++;


}

##############################################################################
##! @fn int main
##! @brief The main program.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
# sub main {
{
   ##############################################################################
   ## Program start
   ##############################################################################

   if (0){
      print "=== Definitions 3 \n";
      exit(1);
   }

   if (@file_errors) {
      # Makes no sense to go on if input parameters are off.
      print "\n============  Summary of errors =================================a\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
      &using_indexer();
      exit(1);
   }


   # Get the master definitions
   while (<IN_MASTER>){    # entire master file into memory.
      $globe::master_index_html .= $_;
   }
   
   
   #####
   # IMPORTANT!!!
   # Master ordering not used
   #####
   &globe::get_master_nav_info();

   $_index_file_list = "$globe::path$_index_file_list";

   if (system ("ls $globe::path$f_type* > $_index_file_list")) {
      print "There is no $globe::path directory. Nothing done.\n";
      return(1);
   }
   
   ##########
   # Create a master index list
   ##########
   &process_index_files();
   print "Total entries incoming = $globe::total_entry_in\n";
   &sort_raw_master();
   &create_index_files();
   &create_index_script();

   if (@file_errors) {
      print "\n============  Summary of errors ==================================\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
   }
   
   
   ##############################################################################
   ## End of all the work
   ##############################################################################
   exit(0);
}


##############################################################################
##! @fn process_index_files
##! @brief Gets a list of index_ files and hands them off one-at-a-time to
##! &get_raw_master.
##!
##! @param $_index_file_list The name of the temporary file that contains
##! a list of potential index_ files.
##!
##! @return Nothing.
##!
##! A system call was made prior to running this routine which generated
##! the list of files in $_index_file_list. This routine 
##!
##! <ol><li>Opens that temporary file.</li>
##! <li> Picks off each file from the list.</li>
##! <li>Passes the index_ file to get_raw_master.</li></ol>
##! 
##! @bug If the index_ file from the list has ".htm" as part of its name,
##! it is ignored.
##!
##! @ingroup tp_idx
##############################################################################
sub process_index_files {
   local ($in_file);
   undef (@path_chunk);
   
   print "Entering process_index_files\n";

   unless (open ( IN_LIST, "$_index_file_list")) {
      print "Cannot open file \"$_index_file_list\"\n";
      exit(1);
   }
   TOKEN_FILE: while (<IN_LIST>){    # read a line from file into $_
      $in_file = $_;
      $in_file =~ s/\n//;
      
      if ($in_file =~ /\.htm/){
         # print "gcm1 $in_file\n";
         next TOKEN_FILE;
      } else {
         print "$in_file\n";
         &get_raw_master($in_file);
      }
      $globe::relative_path = $in_file;
      $globe::relative_path =~ s/index_//;
      $globe::relative_path =~ s/\.html//;
      $globe::relative_path =~ s/\.htm//;
      @path_chunk = split ( /\//, $globe::relative_path, 2);
      if (@path_chunk =~ /2/) {
         $globe::relative_path = "$path_chunk[1]/";
      } else {
         $globe::relative_path .= "\./";
      }
      # print "relative path = $globe::relative_path\n";
   } # while IN_LIST
   close (IN_LIST);

   # print "Master list:\n$globe::master_raw\n";
} #process_index_files



##############################################################################
##! @fn get_raw_master
##! @brief Opens the input file and appends its information to $globe::master_raw
##! before closing and returning.
##!
##! @param $ind_file is the $in_file from process_index_files.
##! @return $globe::master_raw has information from the input file appended to
##! it.
##! 
##! @bug Does no checking on the information from within the input file
##! to verify that it is correct. Garbage in and this may bomb completely.
##! @ingroup tp_idx
##############################################################################
sub get_raw_master {
   local ($ind_file) = $_[0];
   # print "Entering get_raw_master\n";

   unless (open ( IND_FILE, "$ind_file")) {
      print "Cannot open file \"$ind_file\"\n";
      exit(1);
   }
   while (<IND_FILE>){    # read a line from file into $_
      $globe::master_raw .= $_;
      $globe::total_entry_in++;
   } # while IND_FILE
   close (IND_FILE);
} #get_raw_master


##############################################################################
##! @fn sort_raw_master
##! @brief Creates a $idx_struct hash table from the raw index material.
##!
##! @param $globe::master_raw is used to get its information, although not
##! explicitly passed in. To free up memory, this gets undefined.
##!
##! @retval $idx_struct is a hash that uses the index entry as the key
##! and whose contents are the title plus URL.
##! @retval $globe::master_raw is cleared after its information is put into the
##! hash.
##! 
##! Parses the raw index information into a key (the index entry) and its
##! associated title plus URL. The title is the same as the key. It is retained
##! in the contents of the hash to help distinguish between entries and to
##! help eliminate duplicates.
##!
##! This routine calls the word_chunking routine to create additional entries from
##! the items.
##! 
##! Uses @globe::ignore_terms to eliminate entries.
##!
##! When finished, $idx_struct is a big list of index entries.
##!
##! @bug Assumes that input information was of the proper format with an
##! index entry, $globe::word_url_boundary, and URL. If any of the input index_
##! files did not have this, this can cause problems.
##!
##! Eliminates special characters and can purge entries that have Perl 
##! special characters.
##!
##! Word-chunking should have been made an input option. In order to turn off,
##! the code has to be modified: 
##!
##! $entry = compacted and clean display text for sorting.
##! $subentry = compacted and clean display text for sorting.
##! $idx_struct{$entry}{display} = display text
##! $idx_struct{$entry}{url}[] = array of URL's for the $entry.
##! $idx_struct{$entry}{sub}{$subentry}{display} = display text
##!              for the $entry's $subentry.
##! $idx_struct{$entry}{sub}{$subentry}{url}[] = array of URL's 
##!              for the $entry's $subentry.
##!  
##! @ingroup tp_idx
##############################################################################
sub sort_raw_master {
   print "Entering sort_raw_master\n";

   @unchunked_list = split (/\n/, $globe::master_raw);
   TOKEN: for ($i=0; $i<=$#unchunked_list; $i++){
      if ($unchunked_list[$i] =~ /$globe::word_url_boundary/){
         $unchunked_list[$i] =~ s/  / /g;  # some title have 2 spaces
         $unchunked_list[$i] =~ s/\r//g;
         $unchunked_list[$i] =~ s/\n//g;
         @entry_chunk = split (/$globe::word_url_boundary/, $unchunked_list[$i], 2);
         $entry_chunk[0] =~ s/\,[\s]*/$globe::word_c_boundary/;  
         if ($entry_chunk[0] =~ $globe::word_c_boundary){
            #####
            # Indicates that FM had multi-level index token
            # This will not have any word chunking done
            #####
            @title_chunk = split (/$globe::word_c_boundary/, $entry_chunk[0], 2);
            $entry1 = &trash_special_characters ($title_chunk[0]);
            $entry1 =~ s/ //g; # get rid of all spaces
            $entry1 =~ tr/A-Z/a-z/;
            if ((0) && (&ignore_item($entry_chunk[0]))) {
               #####
               # This is turned off because it is assumed that if it came
               # from FM, then the token is valid
               #####
               next TOKEN;
            }
            &add_to_index_struct ($entry1, $title_chunk[0], $entry_chunk[1]);
            
            $entry2 = &trash_special_characters ($title_chunk[1]);
            $entry2 =~ s/ //g; # get rid of all spaces 
            $entry2 =~ tr/A-Z/a-z/;
            if (!(&add_to_lev2_index_struct ($entry1, $entry2, $title_chunk[1], $entry_chunk[1]))) {
               print "Tried adding a second level without the primary level.\n";
            }
         } else { # ($entry_chunk[0] =~ $globe::word_c_boundary)
            #####
            # This will not have any word chunking done
            #####
            $entry1 = &trash_special_characters ($entry_chunk[0]); 
            $entry1 =~ s/ //g; # get rid of all spaces 
            $entry1 =~ tr/A-Z/a-z/;
            if ((0) && (&ignore_item($entry_chunk[0]))) {
               #####
               # This means that the entry is not one of interest
               #####
               next TOKEN;
            }
            &add_to_index_struct ($entry1, $entry_chunk[0], $entry_chunk[1]);
            #####
            # the following does word chunking
            #####
            if ((1) && ($globe::word_chunk)) {
               &word_chunking( $entry_chunk[0], $entry_chunk[1]);
            } # globe::word_chunk
         } # ($entry_chunk[0] =~ $globe::word_c_boundary)
      } else { # TOKEN
         # token does not have URL; don't process
         next TOKEN;
      } # TOKEN
   
   } # $i unchunked_list
   
   $_cnt=0;
   if (0){ # debug loop
      foreach $entry (sort keys %idx_struct){
         $_cnt++;
         print "$_cnt \"$entry\" = \"$idx_struct{$entry}{display}\"\n";
         if (($_cnt < 12) && (1)){
            foreach $subentry (sort keys %{$idx_struct{$entry}{sub}}) {
               print "== $_cnt \"$subentry\" = \"$idx_struct{$entry}{sub}{$subentry}{display}\" \n";
            }
         }
         if ((0)){
            foreach $subentry (sort keys %{$idx_struct{$entry}{sub}}) {
               print "== $_cnt \"$subentry\" = \"$idx_struct{$entry}{sub}{$subentry}{display}\" \n";
            }
         }
         if (($_cnt > 12) && (1)){
            exit(1);
         }
      }
      print "Stopping...\n";
      exit(1);
   } # debug loop
   
} # sort_raw_master




##############################################################################
##! @fn ignore_item
##! @brief Compares the input term to a list of ignore terms.
##!
##! @param $term_to_test Term to test.
##!
##! @return Returns 1 if the term has matched an ignore term;
##! Otherwise it returns 0.
##!
##! It tests for fragments twice, so that "get" doesn't match on
##! "together".
##!
##! @bug If the term has a perl special character in it, it is
##! sent back immediately and left alone.
##!  
##! @ingroup tp_idx
##############################################################################
sub ignore_item {
   $term_to_test = $_[0];

   #####
   # Don't mess with entries that have perl special characters, like +, [
   # Otherwise, it will fail in the testing below; not important enough to
   # purge from the output; probably won't be in the ignore file anyway.
   #####
   if (($term_to_test =~ /\?/)
         || ($term_to_test =~ /\+/)
         || ($term_to_test =~ /\*/)
         || ($term_to_test =~ /\/\./)
         || ($term_to_test =~ /\(/)
         || ($term_to_test =~ /\)/)
         || ($term_to_test =~ /\[/)
         || ($term_to_test =~ /\$/)
         || ($term_to_test =~ /\#/))
   {
      return (0);
   }

   #####
   # Remove index items that are in the ignore file
   #####
   foreach $r_term (@globe::ignore_terms){
      $r2_term = $r_term;
      $r2_term =~ s/ //g;
      #####
      if (0){ #debug
         $stop = 0;
         if ($term_to_test =~ /\+/){
            print "gcm0 $term_to_test\n";
            $stop = 1;
         }
         if ($r_term =~ /\+/){
            print "gcm1 $r_term\n";
            $stop = 1;
         }
         if ($r2_term =~ /\+/){
             print "gcm2 $r2_term\n";
             $stop = 1;
         }
         if ($stop){
             exit(1);
         }
      } #debug
      #####
      #####
      # Need to make sure that "get" doesn't match on "together".
      #####
      if ((($term_to_test =~ /$r_term/i) || ($term_to_test =~ /$r2_term/i)) 
           && (($r_term =~ /$term_to_test/i) || ($r2_term =~ /$entry/i))) {
         if (0){
            print "r_term = $r_term; entry = $term_to_test\n";
         }
         return (1);
      }
   }
   # default fall through; 0 means good.
   return (0);
} # ignore_item


##############################################################################
##! @fn add_to_index_struct
##! @brief Adds an element to the complicated hash table.
##!
##! @param $in_entry the compacted entry into the hash
##! @param $in_title the display text to show for the item
##! @param $in_url the anchor to add to the URL array.
##! If you send in $globe::word_c_boundary for the $in_url,
##! then it won't add the URL to the list.
##!
##! @retval Always returns  1.
##!
##!
##! $entry = compacted and clean display text for sorting.
##! $subentry = compacted and clean display text for sorting.
##! $idx_struct{$entry}{display} = display text
##! $idx_struct{$entry}{url}[] = array of URL's for the $entry.
##! $idx_struct{$entry}{sub}{$subentry}{display} = display text
##!              for the $entry's $subentry.
##! $idx_struct{$entry}{sub}{$subentry}{url}[] = array of URL's 
##!              for the $entry's $subentry.
##!  
##! @ingroup tp_idx
##############################################################################
sub add_to_index_struct {
   $in_entry = $_[0];
   $in_title = $_[1];
   $in_url   = $_[2];
   
   $donothave = 1;
   if ($in_url =~ /$globe::word_c_boundary/) {
      $donothave = 0;
   }
   if (exists ($idx_struct{$in_entry})) {
      # We already have display text;
      # need URL.
      # weed out duplicates from the beginning.
      for ($j = 0; $j <= $#{$idx_struct{$in_entry}{url}}; $j++) {
        if ($idx_struct{$in_entry}{url}[$j] =~ /$in_url/) {
           $donothave = 0;
        }
      } # for $j
      if ($donothave) {
         push (@{$idx_struct{$in_entry}{url}}, $in_url);
      } 
   } else {
      # we don't have anything; need to add it.
      $idx_struct{$in_entry}{display} = $in_title;
      if ($donothave) {
         push (@{$idx_struct{$in_entry}{url}}, $in_url);
      } 
   }
   return (1);
} # add_to_index_struct


##############################################################################
##! @fn add_to_lev2_index_struct
##! @brief Adds an element to the second level of the complicated hash table.
##!
##! @param $in_entry the compacted entry into the hash.
##! @param $in_sub the compacted subentry into the hash.
##! @param $in_title the display text to show for the item.
##! @param $in_url the anchor to add to the URL array.
##!
##! @retval Always returns  1.
##!  
##!
##! $entry = compacted and clean display text for sorting.
##! $subentry = compacted and clean display text for sorting.
##! $idx_struct{$entry}{display} = display text
##! $idx_struct{$entry}{url}[] = array of URL's for the $entry.
##! $idx_struct{$entry}{sub}{$subentry}{display} = display text
##!              for the $entry's $subentry.
##! $idx_struct{$entry}{sub}{$subentry}{url}[] = array of URL's 
##!              for the $entry's $subentry.
##!
##! @ingroup tp_idx
##############################################################################
sub add_to_lev2_index_struct {
   # assumes that $in_entry already exists
   $in_entry = $_[0];
   $in_sub   = $_[1];
   $in_title = $_[2];
   $in_url   = $_[3];
   
   if (!(exists ($idx_struct{$in_entry}))) {
      return (0);
   }
   if (exists ($idx_struct{$in_entry}{sub}{$in_sub})) {
      # We already have display text;
      # need URL.
      # weed out duplicates from the beginning.
      $donothave = 1;
      for ($j = 0; $j <= $#{$idx_struct{$in_entry}{sub}{$in_sub}{url}}; $j++) {
        if ($idx_struct{$in_entry}{sub}{$in_sub}{url}[$j] =~ /$in_url/) {
           $donothave = 0;
        }
      } # for $j
      if ($donothave) {
         push (@{$idx_struct{$in_entry}{sub}{$in_sub}{url}}, $in_url);
      } 
   } else {
      # we don't have anything; need to add it.
      $idx_struct{$in_entry}{sub}{$in_sub}{display} = $in_title;
      push (@{$idx_struct{$in_entry}{sub}{$in_sub}{url}}, $in_url);
   }
   return (1);
} # add_to_lev2_index_struct

##############################################################################
##! @fn word_chunking
##! @brief Performs word-chunking on the passed in entries that was extracted
##! from the $globe::master_raw.
##!
##! @param $unproc_title the unprocessed title passed in $entry_chunk[0]. 
##! 
##! @param $assoc_t_data the associated title and data given by $entry_chunk[1].
##!
##! @return Updated entries in the hash $globe::master_index. If a word-chunk
##! already is available as a key into the hash, then this appends its information
##! to the contents of the key using $globe::division_mult_entry.
##!
##! Word-chunking is performed on the $unproc_title. 
##! Natural boundaries (spaces, dashes, underscores, changes in case in the middle of a word) 
##! are used to create additional two-level index entries that contain the word-chunk 
##! followed by where it came from. 
##!
##! The $globe::ignore_terms_file is used to
##! eliminate unuseful word-chunked entries (such as "the", "a", "to", etc.)
##!
##! The additional useful entries are appended to
##! the contents of the hash $globe::master_raw using the $globe::division_mult_entry
##! separator only if the new entry is not a duplicate.
##!
##! Word-chunking is particular useful for API documentation so that the reader
##! does not have to remember the exact name of a code item in order to find
##! it. An initial index token of "api_GetMovie-list" could be found not just under
##! its name in the "A's", but under "get", "movie", and "list".
##!
##!
##! $entry = compacted and clean display text for sorting.
##! $subentry = compacted and clean display text for sorting.
##! $idx_struct{$entry}{display} = display text
##! $idx_struct{$entry}{url}[] = array of URL's for the $entry.
##! $idx_struct{$entry}{sub}{$subentry}{display} = display text
##!              for the $entry's $subentry.
##! $idx_struct{$entry}{sub}{$subentry}{url}[] = array of URL's 
##!              for the $entry's $subentry.
##!  
##! @bug None.
##! @ingroup tp_idx
##############################################################################
sub word_chunking {
   local ($unproc_title) = $_[0];
   local ($assoc_t_data) = $_[1];
   local ($capital) = 1;
   local ($w_cnt) = 0;
   undef (@w_chunks);
   undef (@word);
   local ($c_word);
   local ($term);

   local ($proc_title) = &trash_special_characters($unproc_title);
   $proc_title =~ s/ //g;
   $proc_title =~ tr/A-Z/a-z/;
   
   if (0) {
      print "$unproc_title $assoc_t_data\n";
   }
   
   #####
   # Return quickly if it is an additional "used by" cross-reference.
   # It'll already have too many word-chunk references. No problem.
   #####
   if ($unproc_title =~ $globe::ack_used_by){
      return (1);
   }
   
   #####
   # Use for word sorting: spaces, underscores, dashes, em-dashes,
   # slashes
   #####
   @w_chunks = split ( /\s|_|-|\&\#8212\;|\&\#151\;|\/|\\|\./, $unproc_title);

   #####
   # Additional word chunking based on capital letters
   # Create even more word chunks based on whether an individual word
   # has smaller sections in it based on capitalization.
   # For example, "fileIsDirectory" has "Is" and
   # "Directory" in it.
   #####
   if ($capital) {   # on/off for capital word-chunking
      NEXT_CC_WORD: foreach $c_word (@w_chunks){
         #####
         # Get rid of elements in word chunk that we don't want hanging around
         # More special characters introduced for operator++
         # Don't bother word chunking them
         #####
         $c_word = &trash_special_characters($c_word);
         #####
         
         #####
         # Check for one or more capital letters within existing word chunk
         #####
         if ($c_word =~ /[A-Z]+/) {
            # print "==== $c_word  ===\n";
            #####
            # Split the word chunk into capital letter chunks if it is in word chunk
            #####
            @cap_chunk = split (/([A-Z]+)/, $c_word);
            if (@cap_chunk > 1) {
               #####
               # Go through all cap chunks; ignore the first element
               # because this will already be handled as part of a word.
               #####
               for ($k=1; $k < @cap_chunk; $k++) {  # ignore 0 element
                  # print "$k sub_cap = $cap_chunk[$k]\n";
                  if (($cap_chunk[$k] =~ /[A-Z]/) && ($k+1 < @cap_chunk)) {
                     #####
                     # Handles when a cap chunk is followed by a small chunk.
                     # print "Create chunk = $cap_chunk[$k] with $cap_chunk[$k+1]\n";
                     #####
                     $temp = "$cap_chunk[$k]$cap_chunk[$k+1]";
                     $temp =~ tr/A-Z/a-z/;
                     push (@word, $temp);
                     #####
                     # jump past next k
                     #####
                     $k++;
                  } elsif (($cap_chunk[$k] =~ /[A-Z]/) && ($k+1 == @cap_chunk)) {
                     #####
                     # Handles situation where last chunk is all caps.
                     # print "Create chunk = $cap_chunk[$k]\n";
                     #####
                     $temp = "$cap_chunk[$k]";
                     $temp =~ tr/A-Z/a-z/;
                     push (@word, $temp);
                     #####
                     # jump past next k cap chunk
                     #####
                     $k++;
                  } else {
                     #####
                     # should never really get to this point
                     # chunk is in small caps somewhere probably at beginning
                     #####
                     print "$k of $#cap_chunk; \"$cap_chunk[$k]\" of \"$c_word\" from \"$unproc_title\"; probably k=0 or 1\n";
                  } 
               } # for the number of cap chunks
            } # if more than one cap chunk
         } # if there are capital letters in word chunk
      } # for each word chunk
      
      #####
      # Need to add the words to the list.
      #####
      foreach $sw_w (@word) {
         push (@w_chunks, $sw_w);
         if (0) {
            print "fractions: $sw_w\n";
         }
      }

      if ((0) && ($unproc_title =~ /BapiVersion/i)) { #debug
         $_ccc= 0;
         print "==== more chunk \"$unproc_title\"\n";
         foreach $c_word (@w_chunks){
            $_ccc++;
            print "1 $_ccc $c_word\n";      
         }
         $_ccc= 0;
         foreach $c_word (@word){
            $_ccc++;
            print "2 $_ccc $c_word\n";      
         }
         exit(1);
      } # debug
   } # # on/off for capital word-chunking


   #####
   # Do something with all of the word chunks
   #####
   { # bracket level
      # print "======= $unproc_title \n";
      NEXT_C_WORD: foreach $c_word (@w_chunks){
         if ($unproc_title =~ /^$c_word/i) {
            #####
            # Don't do the a word chunk that comes
            # close to matching the original name 
            # (as in "log" for "logFuncDebugGet". Hence,
            # We'll purposely skip over it.
            #####
            next NEXT_C_WORD;
         } 
         $c_word =~ tr/A-Z/a-z/;
         $c_word = &trash_special_characters($c_word);
         
         if ((0) && (&ignore_item($c_word))) {
            #####
            # This means that the entry is not one of interest
            #####
            next NEXT_C_WORD;
         }
         #####
         # We don't want extra hyperlinks to word junk fragment.
         # Calling it specifically with $globe::word_c_boundary as url
         #####
         &add_to_index_struct ($c_word, $c_word, $globe::word_c_boundary);

         if (!(&add_to_lev2_index_struct ($c_word, $proc_title, $unproc_title, $assoc_t_data))) {
            print "Tried adding a second level without the primary level.\n";
         }
         
      } # foreach $c_word
   } # bracket level
} # word_chunking


##############################################################################
##! @fn trash_special_characters
##! @brief Removes all special characters that we don't want in index as
##! word chunks.
##!
##! @param $in_word The word that might have special characters.
##! @return The word without special characters.
##! 
##! @bug Debug statements are left in.
##! @ingroup tp_idx
##############################################################################
sub trash_special_characters {
   $in_word = $_[0];
   
   #####
   $in_word =~ s/\&\#151\;//;
   $in_word =~ s/\&\#8212\;//;
   $in_word =~ s/\s//g;
   $in_word =~ s/\:\://g;
   $in_word =~ s/\(//g;
   $in_word =~ s/\)//g;
   $in_word =~ s/\,//g;
   # $in_word =~ s/\.//g; # 07/08/2002 removed
   $in_word =~ s/\-//g;
   $in_word =~ s/\_//g;
   $in_word =~ s/\://g;
   $in_word =~ s/\[//g;
   $in_word =~ s/\]//g;
   $in_word =~ s/\+//g;
   $in_word =~ s/\=//g;
   $in_word =~ s/\*//g;
   $in_word =~ s/\?//g;
   $in_word =~ s/\&//g;
   $in_word =~ s/\%//g;
   $in_word =~ s/\$//g;
   $in_word =~ s/\#//g;
   $in_word =~ s/\@//g;
   $in_word =~ s/\!//g;
   $in_word =~ s/\|//g;
   $in_word =~ s/\\//g;
   $in_word =~ s/\///g;
   $in_word =~ s/\<//g;
   $in_word =~ s/\>//g;
   $in_word =~ s/\"//g;
   $in_word =~ s/\'//g;
   $in_word =~ s/\~//g;
   $in_word =~ s/\`//g;
   #####
   return ($in_word);
} # trash_special_characters



##############################################################################
##! @fn create_index_files
##! @brief Outputs m_idx_ HTML files from the contents of $idx_struct.
##!
##! @param $idx_struct is the sorted list.
##! @return A series of HTML files that make up the index. The files begin
##! with "m_idx_", are followed by a character, and end with ".html".
##! 
##! This uses the $globe::master_index_html and swaps out the information
##! delineated by $globe::m_define{structure}[0] and $globe::m_define{structure}[1].
##!
##! The $idx_struct is sorted. Every time the first letter of
##! the key changes, a new m_idx_ file is created.
##!
##! When creating the output, it formats it using valid HTML with an anchor href
##! containing the valid URL and the text. It eliminates duplicates (e.g., URL
##! is identical). 
##! 
##! It handles the levels, as in second-level index entries.
##!
##! $entry = compacted and clean display text for sorting.
##! $subentry = compacted and clean display text for sorting.
##! $idx_struct{$entry}{display} = display text
##! $idx_struct{$entry}{url}[] = array of URL's for the $entry.
##! $idx_struct{$entry}{sub}{$subentry}{display} = display text
##!              for the $entry's $subentry.
##! $idx_struct{$entry}{sub}{$subentry}{url}[] = array of URL's 
##!              for the $entry's $subentry.
##! 
##! @bug Debug statements are left in.
##! @ingroup tp_idx
##############################################################################
sub create_index_files {
   local ($inner_index) = "This Letter has no entries.";
   local ($remember_letter) = "0";
   local ($remember_level) = "";
   local ($out_file) = $globe::path . "m_idx_"; # will have letter and .html appended
   
   print "Entering create_index_files\n";
   $_cnt=0;
   if (0){ # debug loop
      foreach $entry (sort keys %idx_struct){
         $_cnt++;
         print "$_cnt \"$entry\" = \"$idx_struct{$entry}{display}\"\n";
         if (($_cnt < 25) && (0)){
            foreach $subentry (sort keys %{$idx_struct{$entry}{sub}}) {
               print "== $_cnt \"$subentry\" = \"$idx_struct{$entry}{sub}{$subentry}{display}\"\n";
            }
         }
         if ((1)){
            foreach $subentry (sort keys %{$idx_struct{$entry}{sub}}) {
               print "== $_cnt \"$subentry\" = \"$idx_struct{$entry}{sub}{$subentry}{display}\"\n";
            }
         }
         if (($_cnt > 12) && (0)){
            exit(1);
         }
      }
      exit(1);
   } # debug loop
   
   unless (open ( OUT_INDEX, ">$out_file$remember_letter.html")) {
      push (@file_errors, "Cannot open file \"$out_file$remember_letter.html\"\n");
      print "Cannot open file \"$out_file$remember_letter.html\"\n";
   }

   $_cnt=0;
   PURGE_ENTRY: foreach $entry (sort keys %idx_struct){
      if ((1) && (&ignore_item ($idx_struct{$entry}{display}))) {
         next PURGE_ENTRY;
      }
      
      #####
      # Take care of writing out stored index info for a given letter
      #####
      $first_letter = substr($entry, 0, 1);
      if ($first_letter !~ /[a-zA-Z0-9]/) {
         $first_letter = "\-";
      }
      if (0) {
         print "===letter \"$first_letter\" ===\n$inner_index\n";
      }
      if ($first_letter =~ /$remember_letter/i){
         if (0) {
            print "Current letter ($remember_letter)\n";
         }
      } else {
         #####
         # write to the previous index file
         #####
         @chunks = split ( /$globe::m_define{structure}[0]|$globe::m_define{structure}[1]/, $globe::master_index_html, 3);
         $chunks[1] = $inner_index;
         $globe::master_index_html = join ("", $chunks[0], $globe::m_define{structure}[0], $chunks[1], $globe::m_define{structure}[1], $chunks[2]);
         print (OUT_INDEX $globe::master_index_html);
         if (0) {
            print "===index entries \"$first_letter\" ===\n$inner_index\n";
         }
         # close the previous index file
         close (OUT_INDEX); 
         # Change the letter and open the next index file for writing
         $remember_letter = $first_letter;
         $temp_letter = $remember_letter;
         $temp_letter =~ tr/a-z/A-Z/; 
         unless (open ( OUT_INDEX, ">$out_file$remember_letter.html")) {
            push (@file_errors, "Cannot open file \"$out_file$remember_letter.html\"\n");
            print "Cannot open file \"$out_file$remember_letter.html\"\n";
         }
         # reset the index information
         $inner_index = "<p class=\"GroupTitlesIX\"><center><b>-$temp_letter-</b></center></p>\n";
      }
      #####
      
      #####
      # Take care of writing out stored index info for a given letter
      # Ignore duplicates.
      #####
      
      $inner_index .= "<pre class=\"Level1IX\">";
      if ($#{$idx_struct{$entry}{url}} > 1) {
         #####
         # Has multiple destinations to worry about
         #####
         $inner_index .= "$idx_struct{$entry}{display}";
         $inner_index .= "<br>";
         for ($i=0; $i <= $#{$idx_struct{$entry}{url}}; $i++) {
            $inner_index .= "$idx_struct{$entry}{url}[$i]";
            $inner_index .= "<img src=\"nav_doc.gif\" border=\"0\"></a>";
            $_cnt++;
         }
         $inner_index .= "</pre>\n";
      } else {
         #####
         # Only has one destination to worry about
         #####
         $inner_index .= "$idx_struct{$entry}{url}[0]";
         $inner_index .= "$idx_struct{$entry}{display}";
         $inner_index .= "</a></pre>\n";
         $_cnt++;
      }
      
      PURGE_SUBENTRY: foreach $subentry (sort keys %{$idx_struct{$entry}{sub}}){
         if ((0) && (&ignore_item ($idx_struct{$entry}{sub}{$subentry}{display}))) {
            next PURGE_SUBENTRY;
         }
         $inner_index .= "<pre class=\"Level2IX\">";
         if ($#{$idx_struct{$entry}{sub}{$subentry}{url}} > 1) {
            #####
            # Has multiple destinations to worry about
            #####
            $inner_index .= "$idx_struct{$entry}{sub}{$subentry}{display}";
            $inner_index .= "<br>";
            for ($i=0; $i <= $#{$idx_struct{$entry}{sub}{$subentry}{url}}; $i++) {
               $inner_index .= "$idx_struct{$entry}{sub}{$subentry}{url}[$i]";
               $inner_index .= "<img src=\"nav_doc.gif\" border=\"0\"></a>";
               $_cnt++;
            }
            $inner_index .= "</pre>\n";
         } else {
            #####
            # Only has one destination to worry about
            #####
            $inner_index .= "$idx_struct{$entry}{sub}{$subentry}{url}[0]";
            $inner_index .= "$idx_struct{$entry}{sub}{$subentry}{display}";
            $inner_index .= "</a></pre>\n";
            $_cnt++;
         }
      } # PURGE_SUBENTRY
   } # PURGE_ENTRY for each entry
      
   
   print "Duplicates and ignore terms removed.\n";
   print "Total Individual Hyperlinks: $_cnt\n";

   
   #####
   # Clean-up for last letter/file.
   #####
   # write to the previous index file
   @chunks = split ( /$globe::m_define{structure}[0]|$globe::m_define{structure}[1]/, $globe::master_index_html, 3);
   $chunks[1] = $inner_index;
   $globe::master_index_html = join ("", $chunks[0], $globe::m_define{structure}[0], $chunks[1], $globe::m_define{structure}[1], $chunks[2]);
   print (OUT_INDEX $globe::master_index_html);
   if (0) {
      print "===index entries \"$first_letter\" ===\n$inner_index\n";
   }
   # close the previous index file
   close (OUT_INDEX); 
   #####

} # create_index_files

##############################################################################
##! @fn create_index_script
##! @brief Outputs index script file from the contents of $idx_struct.
##!
##! @param $idx_struct is the sorted list.
##! @return A single script files that make up the index. 
##! 
##! When creating the output, it formats it using valid script with an anchor href
##! containing the valid URL and the text. It eliminates duplicates (e.g., URL
##! is identical). 
##! 
##! It handles the levels, as in second-level index entries.
##!
##! $entry = compacted and clean display text for sorting.
##! $subentry = compacted and clean display text for sorting.
##! $idx_struct{$entry}{display} = display text
##! $idx_struct{$entry}{url}[] = array of URL's for the $entry.
##! $idx_struct{$entry}{sub}{$subentry}{display} = display text
##!              for the $entry's $subentry.
##! $idx_struct{$entry}{sub}{$subentry}{url}[] = array of URL's 
##!              for the $entry's $subentry.
##! 
##! @bug Debug statements are left in.
##! @ingroup tp_idx
##############################################################################
sub create_index_script {
   local ($remember_letter) = "0";
   local ($remember_level) = "";
   local ($out_file) = $globe::path . "m_idx"; # will have letter and .script appended
   local ($very_critical) = 0;
   
   print "Entering create_index_script\n";
   $_cnt=0;
   if (0){ # debug loop
      foreach $entry (sort keys %idx_struct){
         $_cnt++;
         print "$_cnt \"$entry\" = \"$idx_struct{$entry}{display}\"\n";
         if (($_cnt < 25) && (0)){
            foreach $subentry (sort keys %{$idx_struct{$entry}{sub}}) {
               print "== $_cnt \"$subentry\" = \"$idx_struct{$entry}{sub}{$subentry}{display}\"\n";
            }
         }
         if ((1)){
            foreach $subentry (sort keys %{$idx_struct{$entry}{sub}}) {
               print "== $_cnt \"$subentry\" = \"$idx_struct{$entry}{sub}{$subentry}{display}\"\n";
            }
         }
         if (($_cnt > 12) && (0)){
            exit(1);
         }
      }
      exit(1);
   } # debug loop
   


   #####
   # Handle the script file implementation
   #####
      
   unless (open ( OUT_SCRIPT, ">$out_file.script")) {
      push (@file_errors, "Cannot open file \"$out_file.script\"\n");
      print "Cannot open file \"$out_file.script\"\n";
   }
   print "Preparing to output \"$out_file.script\"\n";
   # $inner_script = "Item level=1 image=nav_folderclosed.gif text=Master Index\r\n";
   $inner_script = "";
   
   PURGE_ENTRY_SCRIPT: foreach $entry (sort keys %idx_struct){
      if ((0) && (&ignore_item ($idx_struct{$entry}{display}))) {
         next PURGE_ENTRY_SCRIPT;
      }
      
      #####
      # Take care of index info for a given letter
      #####
      $first_letter = substr($entry, 0, 1);
      if ($first_letter !~ /[a-zA-Z]/) {
         $first_letter = "\-";
         next PURGE_ENTRY_SCRIPT;
      }
      if (0) {
         print "===letter \"$first_letter\" \n";
      }
      if ($first_letter =~ /$remember_letter/i){
         if (0) {
            print "Current letter ($remember_letter)\n";
         }
      } else {
         #####
         # write to the previous index file
         #####
         # Change the letter and open the next index file for writing
         $remember_letter = $first_letter;
         $temp_letter = $remember_letter;
         $temp_letter =~ tr/a-z/A-Z/; 
         # $inner_script .= "Item level=2 image=nav_folderclosed.gif text=-$temp_letter-\r\n";
         $inner_script .= "Item level=1 image=nav_folderclosed.gif text=-$temp_letter-\r\n";
      }
      #####
      
      if ((0) && ($#{$idx_struct{$entry}{url}} > 1)) {
         #####
         # Has multiple destinations to worry about
         #####
         # $inner_script .= "Item level=3 image=nav_folderclosed.gif text=";
         $inner_script .= "Item level=2 image=nav_folderclosed.gif text=";
         $inner_script .= "$idx_struct{$entry}{display}";
         $inner_script .= "\r\n";
         for ($i=0; $i <= $#{$idx_struct{$entry}{url}}; $i++) {
            # $inner_script .= "Item level=4 image=nav_doc.gif url=";
            $inner_script .= "Item level=3 image=nav_doc.gif url=";
            # $inner_script .= "$idx_struct{$entry}{url}[$i]";
            ($before, $piece, $after) = &globe::get_tag_chunk( $idx_struct{$entry}{url}[$i],
                                 "href[\s]*\=[\s]*\"", 
                                 "\"", 
                                 $very_critical);

            if ($piece) {
               $inner_script .= "$piece";
               $inner_script .= ",basefrm ";
            }
            $inner_script .= "\r\n";
         }
      } else {
         #####
         # Only has one destination to worry about
         #####
         # $inner_script .= "Item level=3 image=nav_doc.gif url=";
         $inner_script .= "Item level=2 image=nav_doc.gif url=";
         # $inner_script .= "$idx_struct{$entry}{url}[0]";
         ($before, $piece, $after) = &globe::get_tag_chunk( $idx_struct{$entry}{url}[0],
                                 "href[\s]*\=[\s]*\"", 
                                 "\"", 
                                 $very_critical);
         if ($piece) {
             $inner_script .= "$piece";
             $inner_script .= ",basefrm ";
         }
         $inner_script .= " text=";
         $inner_script .= "$idx_struct{$entry}{display}";
         $inner_script .= "\r\n";
      }
      PURGE_SUBENTRY_SCRIPT: foreach $subentry (sort keys %{$idx_struct{$entry}{sub}}){
         if ((0) && (&ignore_item ($idx_struct{$entry}{sub}{$subentry}{display}))) {
            next PURGE_SUBENTRY_SCRIPT;
         }
         if ((0) && ($#{$idx_struct{$entry}{sub}{$subentry}{url}} > 1)) {
            #####
            # Has multiple destinations to worry about
            #####
            # $inner_script .= "Item level=4 image=nav_folderclosed.gif text=";
            $inner_script .= "Item level=3 image=nav_folderclosed.gif text=";
            $inner_script .= "$idx_struct{$entry}{sub}{$subentry}{display}";
            $inner_script .= "\r\n";
            for ($i=0; $i <= $#{$idx_struct{$entry}{sub}{$subentry}{url}}; $i++) {
               # $inner_script .= "Item level=5 image=nav_doc.gif url=";
               $inner_script .= "Item level=4 image=nav_doc.gif url=";
               # $inner_script .= "$idx_struct{$entry}{sub}{$subentry}{url}[$i]";
               ($before, $piece, $after) = &globe::get_tag_chunk( $idx_struct{$entry}{sub}{$subentry}{url}[$i],
                                 "href[\s]*\=[\s]*\"", 
                                 "\"", 
                                 $very_critical);

               if ($piece) {
                  $inner_script .= "$piece";
                  $inner_script .= ",basefrm ";
               }
               $inner_script .= "\r\n";
            }
         } else {
            #####
            # Only has one destination to worry about
            #####
            # $inner_script .= "Item level=4 image=nav_doc.gif url=";
            $inner_script .= "Item level=3 image=nav_doc.gif url=";
            # $inner_script .= "$idx_struct{$entry}{sub}{$subentry}{url}[0]";
            ($before, $piece, $after) = &globe::get_tag_chunk( $idx_struct{$entry}{sub}{$subentry}{url}[0],
                                 "href[\s]*\=[\s]*\"", 
                                 "\"", 
                                 $very_critical);

            if ($piece) {
               $inner_script .= "$piece";
               $inner_script .= ",basefrm ";
            }
             
            $inner_script .= " text=";
            $inner_script .= "$idx_struct{$entry}{sub}{$subentry}{display}";
            $inner_script .= "\r\n";
         }
      } # PURGE_SUBENTRY_SCRIPT:
   } # PURGE_ENTRY_SCRIPT for each entry

   print (OUT_SCRIPT $inner_script);
   close (OUT_SCRIPT); 

} # create_index_script


##############################################################################
##! @fn using_indexer
##! @brief What to do when no arguments are given.
##! @param None
##! @return None
##! 
##! @bug None
##! @ingroup tp_idx
##############################################################################
sub using_indexer {
   print "\nvoyant_indexer.pl creates index information to be displayed in the navigation\n";
   print "pane. It assumes all input index files have been copied into the\n";
   print "input directory are are named uniquely. The index files\n";
   print "were generated by voyant_nav.pl when processing Doxygen or Mif2Go output.\n";
   print "The index files are temporary files.\n";
   print "\n-- All index files must begin with \"index_\".\n";
   print "-- All files must reside in the input directory.\n\n";
   print "This takes three arguments:\n";
   print "[1] The directory (with slash \) of where to find the raw index files.\n";
   print "[2] The path & name of the HTML file to use as the master template for the individual index files.\n";
   print "   It must have sections for voyant_header, voyant_structure,\n";
   print "   and voyant_footer.\n";
   print "   A section has a begin and end, such as:\n";
   print "     <!-- begin voy_structure -->\n";
   print "     <!-- end voy_structure -->\n";
   print "[3] The path & name of the file containing words to ignore in word-chunking.\n";
   print "\nThe output are files begin with \"m_idx_\".\n";
   print "[4] If you don't want word-chunking, enter \"no_chunk\".\n";
   print "The m_idx output is displayed in a treefrm while the content it controls\ndisplays in the basefrm.\n";
   print "\nTerminating voyant_indexer.pl without doing anything.\n";
   return;
} # using_indexer




##############################################################################
##! @fn int END
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub END {
   undef ($_index_file_list);     # "_index_file_list";
   undef ($in_file);     # "";
   undef ($f_type);     # "index_";


   
   ##############################################################################
   ## Memory clean-up.
   ##############################################################################
   &globe::memory_clean_up();

   print "\n============  Finished voyant_indexer.pl ==================================\n";
} # END





   