#!/usr/local/bin/perl
##############################################################################
##! @file
##! @brief Changes two Latex files in each zlx_ directory of the system
##! so that they contain the appropriate information for
##! the project.
##! 
##! @param $globe::path default location to find the Latex temporary directories. 
##! The name should be terminated with a slash (\). 
##! This should be the top directory where various master-doxygen.sty
##! and master-header.tex can be found and more importantly, the zlx_ directories 
##! where generated LaTex files from Doxygen reside.
##!
##! @param $globe::master_tree_file [optional path and] filename for the file
##! that stores the information needed for the project and the latex files.
##! It does need additional comment tags containing the
##! voy_latex variables and the latex_header variables. 
##!
##! @return Generates appropriate header.tex and doxygen.sty
##! files for each zlx_ directory in the system. These are then used
##! in Latex's PDF generation for each subproject. 
##!
##! This tool reads in the voyant_master_nav.html file (or equivalent).
##! This file must specify:
##! <ul><li>The voy_order for the system that lists all directories, associated
##! names, and associated PDF files.</li>
##! <li>The voy_latex which contains variables used in LaTex build such as
##! master file names, company name, etc.</li>
##! <li>The latex_head which contains additional information for the refman.tex
##! file.</li>
##! </ul>
##!
##! This tool steps through all cref_ directory names in the voy_order and generates
##! an equivalent zlx_ directory name which should be found under $globe::path.
##! The voy_latex variables declare where additional master.tex and master.sty
##! files reside. These contain call-outs for variables that voy_latex resolves
##! for the given project. All variables in the master files are resolved
##! with information appropriate for the project.
##!
##! Upon conclusion, appropriate refman.tex and doxygen.sty files are written
##! to each zlx_ directory.
##!
##! @ingroup tp_tools tp_latex
##!
##! @author Glenn C. Maxey
##
##    $Id: voyant_latex.pl,v 1.12 2002/11/20 15:23:07 gmaxe Exp $
##
## 2002 Created by Voyant Technologies, Inc., Westminster, Colorado, USA.
##
## Permission to use, copy, modify, and distribute this software and its 
## documentation under the terms of the GNU General Public License is hereby 
## granted. No representations are made about the suitability of this software 
## for any purpose. It is provided "as is" without express or implied warranty. 
## See the GNU General Public License (http://www.gnu.org/copyleft/gpl.html) 
## for more details.
## 
## Documents produced by this script are derivative works derived from the 
## input used in their production; they are not affected by this license.
##
## Revision Information:
##
## $Log: voyant_latex.pl,v $
## Revision 1.12  2002/11/20 15:23:07  gmaxe
## Added exit codes so that wrapper scripts can catch errors properly.
##
## Revision 1.11  2002/07/26 18:56:12  gmaxe
## Got rid of old definitions and migrated new structures into all;
## enhanced xhelp output file names and format; now everything is
## alphebetized.
##
## Revision 1.10  2002/07/22 21:51:32  gmaxe
## Created separate configuration file; allow for two inputs to
## the voyant_latex.pl. commented out old variables in globe.pm.
##
## Revision 1.9  2002/04/06 01:55:16  gmaxe
## New files and new general constructs for handling tags. Uses globals better.
## The html_look* files are designed to handle spider tracing of html systems.
##
## Revision 1.8  2002/01/16 19:49:09  gmaxe
## Removed much of revision history in comments.
##
## Revision 1.7  2002/01/16 19:41:46  gmaxe
## Updated description.
##
##############################################################################

##############################################################################
##! @fn int BEGIN
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub BEGIN {
   print "\n============  Starting voyant_latex.pl ==================================\n";
   $_file_list = "_file_list";
   $in_file = "";
   $f_type = "htm";
   $loc_path = "./";
   $_arg_inc = 0;

   $no_scope_file = 0;
   $scope_pm = "globe.pm";  # first time through; other scope stuff passed in.

   push (@INC, `pwd`);
   push (@INC, '../perl');
   if (0){
      print (@INC, "\n");
   }
   #####
   # All global variables are defined in the following file
   #####
   unless (open ( IN_LIST, $scope_pm)) {
      $scope_pm = "../perl/$scope_pm";
      unless (open ( IN_LIST, $scope_pm)) {
         push (@file_errors, "Cannot open file \"$scope_pm\"");
         $no_scope_file++;
      }
   }
   close (IN_LIST);
   push (@INC, $scope_pm);
   if (!@file_errors) {
      #####
      # All global variables are defined in the following file
      #####
      require $scope_pm;

      if (&globe::declare_variables()) {
         print "Variables initialized from $scope_pm.\n";
      } else {
         push (@file_errors, "Could not initialize variables from $scope_pm.");
      }
   } # if not @file_errors
   
   # Get path if there is one.
   if (@ARGV > $_arg_inc) {
      $globe::path = @ARGV[$_arg_inc];
      if ($globe::path =~ /\//) {
         print "The path specified is $globe::path\n";
   
         @path_chunks = split ( /\//, $globe::path);
         if (@path_chunks < 2) {
            $globe::rel_path_to_start_point = "./";
         } else {
            $globe::rel_path_to_start_point = "../";
         }
      } else {
         push (@file_errors, "The input argument \"$globe::path\" requires a forward slash (\/) at the end.\n");
      }
   } # if 1 or more arguments
   $_arg_inc++;

   if (@ARGV > $_arg_inc) {
      $globe::master_nav_file = @ARGV[$_arg_inc];
      print "The master file is $globe::master_nav_file\n";
      unless (open ( IN_MASTER, $globe::master_nav_file)) {
         push (@file_errors, "Cannot open file \"$globe::master_nav_file\"");
      }
      close (IN_MASTER);
   } # if 2 or more arguments
   $_arg_inc++;

   undef ($globe::master_nav_file2);
   if (@ARGV > $_arg_inc) {
      $globe::master_nav_file2 = @ARGV[$_arg_inc];
      print "The master file is $globe::master_nav_file2\n";
      unless (open ( IN_MASTER, $globe::master_nav_file2)) {
         push (@file_errors, "Cannot open file \"$globe::master_nav_file2\"");
      }
      close (IN_MASTER);
   } # if 2 or more arguments
   $_arg_inc++;


}

##############################################################################
##! @fn int main
##! @brief The main program.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
# sub main {
{
   ##############################################################################
   ## Start of Main Program
   ##############################################################################
   if (0){
      print "=== Definitions 3 \n";
      exit(1);
   }

   if (@file_errors) {
      # Makes no sense to go on if input parameters are off.
      print "\n============  Summary of errors =================================a\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
      &using_voy_latex();
      exit(1);
   }

   unless (open ( IN_MASTER, $globe::master_nav_file)) {
      push (@file_errors, "Cannot open file \"$globe::master_nav_file\"");
   }

   # Get the master definitions
   while (<IN_MASTER>){    # entire master file into memory.
      $globe::master_nav .= $_;
   }
   close (IN_MASTER);

   if ($globe::master_nav_file2) {
      unless (open ( IN_MASTER, $globe::master_nav_file2)) {
         push (@file_errors, "Cannot open file \"$globe::master_nav_file2\"");
      }
   }

   # Get the master definitions
   while (<IN_MASTER>){    # entire master file into memory.
      $globe::master_nav .= $_;
   }
   close (IN_MASTER);

   
   if (&fill_in_master_sections() =~ 1){
      print "Master Navigation file \"$globe::master_nav_file\" has definition errors.\n";
      push (@file_errors, "Master Navigation file \"$globe::master_nav_file\" has definition errors.\n");
   }

   # Write the TEX and STY files
   &new_doxygen_sty();
   # Update any existing refman.tex files
   &new_refman_tex();


   ##############################################################################
   ## End of most of work
   ##############################################################################
   exit(0);
}

##############################################################################
##! @fn int new_refman_tex
##! @brief Creates a new refman.tex file to use as a master.
##!
##! @param global variables are used for the master definitions.
##!
##! @return Writes out a brand-new refman_tex file.
##!
##! @bug This makes heavy use of global variables.
##!
##! @ingroup tp_latex
##############################################################################
sub new_refman_tex {


   ######
   # Extract latex title page from the master_header_tex
   ######
   # escape out important characters
   $start = $globe::latex_title[1];
   $stop = $globe::latex_title[2];
   $start =~ s/\\/\\\\/g;
   $start =~ s/\{/\\\{/g;
   $start =~ s/\}/\\\}/g;
   $stop =~ s/\\/\\\\/g;
   $stop =~ s/\{/\\\{/g;
   $stop =~ s/\}/\\\}/g;
   # end of escape characters
   
   # print "start=$globe::latex_title[1]\nstop=$globe::latex_title[2]\n";

   use POSIX 'strftime';
   $dddate = strftime "%h %d, %Y", localtime;


   ALL_DIR: for ($i=0; $i < @globe::master_order; $i++){
      if ($globe::master_order[$i] !~ /cref_/i){
         next ALL_DIR;
      }
      # print "$i = dir $globe::master_order[$i] title \"$globe::master_order_title[$i]\"\n";
      # print "$i = pdf $globe::master_order_pdf[$i] and doc_num \# $globe::master_order_doc_num[$i]\n";
   
      $loc_path = "$globe::path$globe::master_order[$i]\/";
      $loc_path =~ s/cref_/zlx_/i;
      unless (open ( IN_OUT_TEX, "$loc_path$globe::latex_title[0]")) {
        print "Cannot open file \"$loc_path$globe::latex_title[0]\" for reading: $globe::master_order[$i].\n";
        push (@file_errors, "Cannot open file \"$globe::latex_title[0]\" for reading: $globe::master_order[$i].");
        next ALL_DIR;
      }
      local ($refman) = "";
      while (<IN_OUT_TEX>){
         $refman .= $_;
      }
      close ( IN_OUT_TEX);
      
      #######
      # Update contents with new title page
      #######
      
      @t_chunk = split (/$start|$stop/, $refman, 3);
      if (@t_chunk =~ /3/){
         $t_chunk[1] = "$globe::latex_title[1]$globe::latex_title_page$globe::latex_title[2]";
         # print "Latex title page =====\n$globe::latex_title_page\n";
         $refman = join ("", @t_chunk);
      } else {
         print "Latex file doesn't have title page defined.\n";
         push (@file_errors, "Latex file doesn't have title page defined.\n");
         push (@file_errors, "Needs definitions for start ($globe::latex_title[1]) and stop ($globe::latex_title[2])\n");
      }

      #######
      
      
      unless (open ( IN_OUT_TEX, ">$loc_path$globe::latex_title[0]")) {
        print "Cannot open file \"$loc_path$globe::latex_title[0]\" for writing.\n";
        push (@file_errors, "Cannot open file \"$loc_path$globe::latex_title[0]\" for writing.");
        next ALL_DIR;
      }
      # print "output file = $loc_path$globe::latex_title[0]\n";
      # OVERWRITE FILENAME IN THESE TWO WITH COPY OF MASTER.
      $local'output_header_tex = $refman;
      # Swap out flags with proper text;
      foreach $key (keys %globe::latex_flags) {  
         # print "key = $key = $globe::latex_flags{$key}\n";
         if ($local'output_header_tex =~ /$key/){
            $local'output_header_tex =~ s/$key/$globe::latex_flags{$key}/g;
         }
      }
      # Take care of the title
      if ($local'output_header_tex =~ /$globe::flag_title/){
         $local'output_header_tex =~ s/$globe::flag_title/$globe::master_order_title[$i]/g;
      }
      # Take care of the document number
      if ($local'output_header_tex =~ /$globe::flag_doc_num/){
         $temp = $globe::master_order_pdf[$i];
         $temp =~ s/.pdf//gi;
         $local'output_header_tex =~ s/$globe::flag_doc_num/$temp/g;
      }
      # Take care of the date
      if ($local'output_header_tex =~ /$globe::flag_date/){
         $local'output_header_tex =~ s/$globe::flag_date/$dddate/g;
      }
      if ($local'output_doxygen_sty =~ /$globe::flag_date/){
         $local'output_doxygen_sty =~ s/$globe::flag_date/$dddate/g;
      }
      # Output the stuff to where it belongs...
      print (IN_OUT_TEX  $local'output_header_tex);
      close (IN_OUT_TEX);
      # print "\n\n\n=====\noutput file = $loc_path$globe::latex_title[0]======\n";
      # print ($local'output_header_tex);
   }  # for all indices into globe master order

} # new_refman_tex

##############################################################################
##! @fn int new_doxygen_sty
##! @brief Retrieves the master TEX/STY definitions and swaps in information
##! from the mast files about document names, number, etc.
##!
##! @param global variables are used for the master definitions.
##!
##! @return Sets several global variables and flags. Outputs to both a TEX
##! and STY file.
##!
##! @bug This makes heavy use of global variables.
##!
##! @ingroup tp_latex
##############################################################################
sub new_doxygen_sty {
   foreach $key (keys %globe::latex_flags) {  
      if ($key =~ /$globe::master_header_tex/){
         $globe::master_header_tex = $globe::latex_flags{$key};  # now has filename
         delete( $globe::latex_flags{$key}); 
      } elsif ($key =~ /$globe::master_doxygen_sty/){
         $globe::master_doxygen_sty = $globe::latex_flags{$key};  # now has filename
         delete( $globe::latex_flags{$key}); 
      } elsif ($key =~ /$globe::output_header_tex/){
         $globe::output_header_tex = $globe::latex_flags{$key};  # now has filename
         delete( $globe::latex_flags{$key}); 
      } elsif ($key =~ /$globe::output_doxygen_sty/){
         $globe::output_doxygen_sty = $globe::latex_flags{$key};  # now has filename
         delete( $globe::latex_flags{$key}); 
      }
      # print "key = $key = $globe::latex_flags{$key}\n";
   }

   unless (open ( IN_MASTER, $globe::master_header_tex)) {
     print "Cannot open file \"$globe::master_header_tex\"\n";
     push (@file_errors, "Cannot open file \"$globe::master_header_tex\"");
     exit(1);
   }
   $globe::master_header_tex = "";
   # Get the master definitions
   while (<IN_MASTER>){    # entire master file into memory.
      $globe::master_header_tex .= $_; # Now has contents
   }
   close (IN_MASTER);

   # print "Master tex\n=========\n$globe::master_header_tex=========\n";
   
   ######
   # Extract latex title page from the master_header_tex
   ######
   # escape out important characters
   $start = $globe::latex_title[1];
   $stop = $globe::latex_title[2];
   $start =~ s/\\/\\\\/g;
   $start =~ s/\{/\\\{/g;
   $start =~ s/\}/\\\}/g;
   $stop =~ s/\\/\\\\/g;
   $stop =~ s/\{/\\\{/g;
   $stop =~ s/\}/\\\}/g;
   # end of escape characters
   
   # print "start=$globe::latex_title[1]\nstop=$globe::latex_title[2]\n";
   @t_chunk = split (/$start|$stop/, $globe::master_header_tex, 3);
   # for ($j=0; $j < @t_chunk; $j++){
   #   print "$j = $t_chunk[$j]\n";
   # }
   if (@t_chunk =~ /3/){
      $globe::latex_title_page = $t_chunk[1];
      # print "Latex title page =====\n$globe::latex_title_page\n";
   } else {
      print "Master Latex file doesn't have title page defined.\n";
      push (@file_errors, "Master Latex file doesn't have title page defined.\n");
      push (@file_errors, "Needs definitions for start ($globe::latex_title[1]) and stop ($globe::latex_title[2])\n");
   }
   ######


   unless (open ( IN_MASTER, $globe::master_doxygen_sty)) {
     print "Cannot open file \"$globe::master_doxygen_sty\"\n";
     push (@file_errors, "Cannot open file \"$globe::master_doxygen_sty\"");
     exit(1);
   }
   $globe::master_doxygen_sty = "";
   # Get the master definitions
   while (<IN_MASTER>){    # entire master file into memory.
      $globe::master_doxygen_sty .= $_; # Now has contents
   }
   close (IN_MASTER);
   # print "Master sty\n=========\n$globe::master_doxygen_sty=========\n";

   use POSIX 'strftime';
   $dddate = strftime "%h %d, %Y", localtime;

   ALL_DIR: for ($i=0; $i < @globe::master_order; $i++){
      if ($globe::master_order[$i] !~ /cref_/i){
         next ALL_DIR;
      }
      # print "$i = dir $globe::master_order[$i] title \"$globe::master_order_title[$i]\"\n";
      # print "$i = pdf $globe::master_order_pdf[$i] and doc_num \# $globe::master_order_doc_num[$i]\n";
   
      $loc_path = "$globe::path$globe::master_order[$i]\/";
      $loc_path =~ s/cref_/zlx_/i;
      unless (open ( OUT_TEX, ">$loc_path$globe::output_header_tex")) {
        print "Cannot open file \"$loc_path$globe::output_header_tex\" for writing.\n";
        push (@file_errors, "Cannot open file \"$loc_path$globe::output_header_tex\" for writing.");
        next ALL_DIR;
      }
      unless (open ( OUT_STY, ">$loc_path$globe::output_doxygen_sty")) {
        print "Cannot open file \"$loc_path$globe::output_doxygen_sty\" for writing.\n";
        push (@file_errors, "Cannot open file \"$loc_path$globe::output_doxygen_sty\" for writing.");
        next ALL_DIR;
      }
      # OVERWRITE FILENAME IN THESE TWO WITH COPY OF MASTER.
      $local'output_header_tex = $globe::master_header_tex;
      $local'output_doxygen_sty = $globe::master_doxygen_sty;
      # Swap out flags with proper text;
      foreach $key (keys %globe::latex_flags) {  
         # print "key = $key = $globe::latex_flags{$key}\n";
         if ($local'output_header_tex =~ /$key/){
            $local'output_header_tex =~ s/$key/$globe::latex_flags{$key}/g;
         }
         if ($local'output_doxygen_sty =~ /$key/){
            $local'output_doxygen_sty =~ s/$key/$globe::latex_flags{$key}/g;
         }
      }
      # Take care of the title
      if ($local'output_header_tex =~ /$globe::flag_title/){
         $local'output_header_tex =~ s/$globe::flag_title/$globe::master_order_title[$i]/g;
      }
      if ($local'output_doxygen_sty =~ /$globe::flag_title/){
         $local'output_doxygen_sty =~ s/$globe::flag_title/$globe::master_order_title[$i]/g;
      }
      # Take care of the document number
      if ($local'output_header_tex =~ /$globe::flag_doc_num/){
         $temp = $globe::master_order_pdf[$i];
         $temp =~ s/.pdf//gi;
         $local'output_header_tex =~ s/$globe::flag_doc_num/$temp/g;
      }
      if ($local'output_doxygen_sty =~ /$globe::flag_doc_num/){
         $temp = $globe::master_order_pdf[$i];
         $temp =~ s/.pdf//gi;
         $local'output_doxygen_sty =~ s/$globe::flag_doc_num/$temp/g;
      }
      # Take care of the date
      if ($local'output_header_tex =~ /$globe::flag_date/){
         $local'output_header_tex =~ s/$globe::flag_date/$dddate/g;
      }
      if ($local'output_doxygen_sty =~ /$globe::flag_date/){
         $local'output_doxygen_sty =~ s/$globe::flag_date/$dddate/g;
      }
      # Output the stuff to where it belongs...
      print (OUT_TEX  $local'output_header_tex);
      print (OUT_STY  $local'output_doxygen_sty);
      close (OUT_TEX);
      close (OUT_STY);
   }  # for all indices into globe master order

} # new_doxygen_sty



##############################################################################
##! @fn int using_voy_latex
##! @brief What to do when no arguments are given.
##! @param None
##! @return None
##! 
##! @bug None
##! @ingroup tp_latex
##############################################################################
sub using_voy_latex {
   print "\nvoyant_latex.pl creates TEX and STY files output files for \n";
   print "Latex to use when creating PDF files from Doxygen output. \n";
   print "-- The first argument is the path where to start.\n";
   print "   It is assumed that the latex subdirectories will be under here.\n";
   print "-- The second argument is file to use to find the master order.\n";
   print "-- The optional third argument is file to use to find the latex definitions.\n";
   print "NOTE: The lates directories have to be named similar to\n";
   print "      their code reference directories \"zlx_\" instead of \"cref_\".\n";
   print "\nTerminating voyant_latex.pl without doing anything.\n";
   return;
}



##############################################################################
##! @fn int fill_in_master_sections
##! @brief Gets the ordering out of the master file.
##! @param None.
##! @return Stores information into global variables for potential use.
##! 
##! Parses the master file for necessary information. This includes:<ul>
##! <li>directory.</li>
##! <li>document name.</li>
##! <li>PDF file name.</li>
##! <li>document number.</li></ul>
##! The ordering is the directory under doc_publish that the material is supposed
##! to appear in.
##! @bug None.
##! @ingroup tp_latex
##############################################################################
sub fill_in_master_sections {

   &globe::get_master_nav_info();
   #####
   # Get the latex file information
   #####
   $def_type = "latex";
   if (exists ($globe::m_info{$def_type})){
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::m_info{$def_type},
                               $globe::m_define{$def_type}[0], 
                               $globe::m_define{$def_type}[1], 
                               $not_critical);
      if ($piece){
         $globe::m_info{$def_type} = $piece;
         ##################
         # Latex flag definitions
         ##################
         @pre_chunks = split ( /\n/, $globe::m_info{$def_type});
         if (0){
            for ($i=0; $i < @pre_chunks; $i++){  
               print "$i == $pre_chunks[$i]\n";
            }
         }
         $c_cnt=0;
         for ($i=0; $i <= $#pre_chunks; $i++){
            if ($pre_chunks[$i] =~ /[\w+]/) {
               @lx_chunk = split ( /\,/, $pre_chunks[$i], 2);
               for ($j=0; $j < @lx_chunk; $j++){
                  # strip out quotation marks
                  $lx_chunk[$j] =~ s/[\s]*\"[\s]*//g;
               }
               $lx_flag[$c_cnt] = $lx_chunk[0];
               $globe::latex_flags{$lx_flag[$c_cnt]} = $lx_chunk[1];
               # we use the c_cnt, because they're might be blank lines
               $c_cnt++;
            }
         }  # for $i
      }
   } else {
      push (@file_errors, "ERROR: Master file does not have proper latex definition.");
   }
   
   #####
   # Get the latex header information
   # Latex title page definitions
   #####
   $def_type = "latex_header";
   if (exists ($globe::m_info{$def_type})){
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::m_info{$def_type},
                               $globe::m_define{$def_type}[0], 
                               $globe::m_define{$def_type}[1], 
                               $not_critical);
      if ($piece){
         @globe::latex_title = split ( /\,/, $piece, 3);
         for ($i=0; $i <= $#globe::latex_title; $i++){  
             # strip out quotation marks
             $globe::latex_title[$i] =~  s/[\s]*\"[\s]*//g;
             if (0){
                print "$i == $globe::latex_title[$i]\n";
             }
         }
      }
   } else {
      push (@file_errors, "ERROR: Master file does not have proper latex header definition.");
   }
   
   if (0) {
      foreach $k (keys %globe::m_info) {
         if ($k =~ /^latex/i){
            print "$k $globe::m_info{$k}\n";
         }
      }
      foreach $key (keys %globe::latex_flags) {  
         print "key = $key = $globe::latex_flags{$key}\n";
      }
      for ($i=0; $i <= $#globe::latex_title; $i++){  
         print "$i == $globe::latex_title[$i]\n";
      }

      print "glenn";
      exit(1);
   }

   undef (@pre_chunks);
   undef (@chunks);
   undef (@lx_chunks);
   undef ($start);
   undef ($stop);
   
} # fill_in_master_sections


##############################################################################
##! @fn int END
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub END {
   undef ($_file_list); # = "_file_list";
   undef ($in_file); # = "";
   undef ($f_type); # = "htm";
   undef ($loc_path); # = "./";

   if (@file_errors) {
      print "\n============  Summary of errors =================================b\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
   }
   
   ##############################################################################
   ## Memory clean-up.
   ##############################################################################
   &globe::memory_clean_up();

   print "\n============  Finished voyant_latex.pl ==================================\n";
} # END
