#!/usr/local/bin/perl
##############################################################################
##! @file
##! @brief Creates table of contents navigation to be displayed in the 
##! navigation pane.
##! 
##! @param $globe::path is the directory_name or location to find the 
##! tree script files. The name should be terminated with a slash (\). 
##! Although the directory_name is the first command line parameter,
##! the true input are the tree script files contained within that directory.
##! Tree script files must begin with "tree_" and end with ".script". 
##!
##! @param $globe::master_filename [optional path and] filename for the project file
##! that contains the directory ordering and official titles.
##! 
##! @return m_tree_ files for each directory of the system. The files are
##! named such that they are associated with their owning directory. 
##!
##! Tree script files are another way of saying "Table of Contents" files.
##! Tree files themselves are mini-TOC files that were generated by
##! voyant_nav.pl or tree_js_2_script.pl.
##!
##! This program reads in the project file (master_filename).
##! Based on the directory names, it makes some assumptions about the names
##! of the tree script files that it expects to see. It steps through all
##! tree script files, reads them in, creates associated master tree script files,
##! and then creates a master tree script file that nests the other tree
##! script files in order.
##!
##! If an expected tree script file is missing, an error message is output but
##! a working master tree is still generated that ignores the missing mini-TOC.
##! 
##! @note When the original tree script files were created, they had no
##! path information in the hyperlinks. This was to make them more general
##! and more easy to use in other directory structures. The tree script
##! files needed to stand on their own and be independent of any resulting
##! directory structure where they might be shared.
##! 
##! Hence, this program must create master script files that include the expect 
##! relative paths to the destination topics. The relative path information is
##! again the directory names from the project file.
##!
##! @bug The tree files (mini-TOC) that are used as input have unique
##! naming requirements.<ul>
##! <li>The mini-TOC files must begin with "tree_"</li>
##! <li>The mini-TOC files must end with ".script".</li>
##! <li>The mini-TOC files must contain the name of the directory they are associated
##! with.</li>
##! </ul>
##!
##! Rather than passing in variables which can create copies in memory, many items
##! use global variables that are defined in globe.pm. When a
##! variable is known to be global, its name begins with "$globe::". The intent is
##! to facilitate maintenance by having all user-defined tags in one place outside 
##! of the program.
##!
##! @ingroup tp_tools tp_toc
##!
##! @author Glenn C. Maxey
##
##    $Id: voyant_mt_app.pl,v 1.6 2003/01/03 19:40:12 gmaxe Exp $
##
## 2002 Created by Voyant Technologies, Inc., Westminster, Colorado, USA.
##
## Permission to use, copy, modify, and distribute this software and its 
## documentation under the terms of the GNU General Public License is hereby 
## granted. No representations are made about the suitability of this software 
## for any purpose. It is provided "as is" without express or implied warranty. 
## See the GNU General Public License (http://www.gnu.org/copyleft/gpl.html) 
## for more details.
## 
## Documents produced by this script are derivative works derived from the 
## input used in their production; they are not affected by this license.
##
##    Revision Information:
##
##    $Log: voyant_mt_app.pl,v $
##    Revision 1.6  2003/01/03 19:40:12  gmaxe
##    More accurate comments.
##
##    Revision 1.5  2002/12/24 15:05:55  gmaxe
##    New name and tweaks to add full paths.
##
##    Revision 1.4  2002/11/20 15:23:07  gmaxe
##    Added exit codes so that wrapper scripts can catch errors properly.
##
##    Revision 1.3  2002/04/06 01:55:16  gmaxe
##    New files and new general constructs for handling tags. Uses globals better.
##    The html_look* files are designed to handle spider tracing of html systems.
##
##    Revision 1.2  2002/03/04 22:21:55  gmaxe
##    Enhanced to handle nested script files.
##
##    Revision 1.1  2002/02/16 00:59:44  gmaxe
##    Added support for script files needed for TOC applet.
##
##    Revision 1.14  2002/02/12 01:50:29  gmaxe
##    Added support for more level chunking;
##    added files to support xhelp;
##    added file to support generating TOC, index from lists.
##    PTR#:
##
##    Revision 1.13  2002/01/16 19:49:10  gmaxe
##    Removed much of revision history in comments.
##
##    Revision 1.12  2002/01/11 21:53:54  gmaxe
##    Updated comments.
##
##############################################################################


##############################################################################
##! @fn int BEGIN
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub BEGIN {
   print "\n============  Starting voyant_mt_app.pl ==================================\n";
   $_file_list = "_file_list";
   $_arg_inc = 0;

   $no_scope_file = 0;
   $scope_pm = "globe.pm";  # first time through; other scope stuff passed in.
   $in_file = "";

   push (@INC, `pwd`);
   push (@INC, '/rtfm/techpubs/perl');
   if (0){
      print (@INC, "\n");
   }
   #####
   # All global variables are defined in the following file
   #####
   unless (open ( IN_LIST, $scope_pm)) {
      unless (open ( IN_LIST, "/rtfm/techpubs/perl/$scope_pm")) {
         push (@file_errors, "Cannot open file \"$scope_pm\" or \"/rtfm/techpubs/perl/$scope_pm\"\n");
         $no_scope_file++;
      }
   }
   close (IN_LIST);
   push (@INC, $scope_pm);
   push (@INC, "/rtfm/techpubs/perl/$scope_pm");

   if (!@file_errors) {
      #####
      # All global variables are defined in the following file
      #####
      require $scope_pm;
   
      if (&globe::declare_variables()) {
         print "Variables initialized from $scope_pm.\n";
      } else {
         push (@file_errors, "Could not initialize variables from $scope_pm.\n");
      }
   } # if not @file_errors


   if (@ARGV < 1) {
      push (@file_errors, "No path specified as first argument.\n");
   }
   
   # Get path if there is one.
   if (@ARGV > $_arg_inc) {
      $globe::path = @ARGV[$_arg_inc];
      if ($globe::path =~ /\//) {
         print "The path specified is $globe::path\n";
      
         @path_chunks = split ( /\//, $globe::path);
         if (@path_chunks < 2) {
            $globe::rel_path_to_start_point = "./";
         } else {
            $globe::rel_path_to_start_point = "../";
         }
      } else {
         push (@file_errors, "The input argument \"$globe::path\" requires a forward slash (\/).\n");
      }
   } # if 1 or more arguments
   $_arg_inc++;
   
   if (@ARGV > $_arg_inc) {
      $globe::master_filename = @ARGV[$_arg_inc];
      print "The master file is $globe::master_filename\n";
      unless (open ( IN_MASTER, $globe::master_filename)) {
         push (@file_errors, "Cannot open file \"$globe::master_filename\"\n");
      }
   } # if 2 or more arguments
   $_arg_inc++;

   if (@ARGV > $_arg_inc) {
      # This is optional; a default is set in globe.pm.
      $globe::script_prefix = @ARGV[$_arg_inc];
   } # if 3 or more arguments
   $_arg_inc++;

   if (@ARGV > $_arg_inc) {
      # This is optional; a default is set in globe.pm.
      $globe::script_master = @ARGV[$_arg_inc];
   } # if 3 or more arguments
   $_arg_inc++;
   
}

##############################################################################
##! @fn int main
##! @brief The main program.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
# sub main {
{
   ##############################################################################
   ## Program start
   ##############################################################################

   if (@file_errors) {
      # Makes no sense to go on if input parameters are off.
      print "\n============  Summary of errors ==================================\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
      &using_voy_mt_app();
      exit(1);
   }


   # Get the master definitions
   while (<IN_MASTER>){    # entire master file into memory.
      $globe::master_nav .= $_;
   }
   close (IN_MASTER);
   
   &globe::get_master_nav_info();
   if (!exists ($globe::m_info{order})) {
      if (1) {
         push (@file_errors, "ERROR: The master file \"$globe::master_tree_file\" does not have the ordering.");
         print "Ordering begins with \"$globe::m_define{order}[0]\" \nand ends with \"$globe::m_define{order}[1]\"\n";
         print "Inbetween on a line by themselves is the directory doc_publish subdirectory name.\n";
      }
      exit(1);
   } else {
      if (0) {
         print "Got to here...\n$globe::m_info{order}\n======\n";
         for ($i = 0; $i <= $#{$globe::m_info{_array_order}}; $i++) {
            foreach $k (keys %{$globe::m_info{_array_order}[$i]}) {
               print "$i $k $globe::m_info{_array_order}[$i]{$k}\n";
            }
         }
         exit(1);
      }
   }


   $nested_scripts = 1;
   if ($nested_scripts) {
      # new way; preferred way.
      if (&create_nested_tree_scripts() ) {
         push (@file_errors, "Did not create nessecary script files.\n");
      } else {
         print "Created nessecary script files.\n";
      }
   } else {
      # old way
      &create_master_tree_script();
   }
   

   ##############################################################################
   ## End of most of work
   ##############################################################################
   exit(0);
} # end of main


##############################################################################
##! @fn int create_nested_tree_scripts
##! @brief Creates multiple script output files that represent the
##! expanding/collapsing elements of the table of contents.
##!
##! @param global_variables are used throughout.
##!
##! @return A series of master tree (m_tree_...script) files that 
##! make up the navigation.
##!
##! This is the preferred way.
##!
##! @bug None.
##!
##! @ingroup tp_toc
##############################################################################
sub create_nested_tree_scripts {
   local ($key) = "";
   local ($out_file) = "";

   local ($section);
   local ($holder);
   print "Entering create_master_tree_script...\n";
   
   local ($master_script) = "";
   


   #####
   # Go through all items in the master order.
   # We are only worrying about the things I know about
   #####
   TOORDER: for ($i=0; $i <= $#{$globe::m_info{_array_order}}; $i++){
      next TOORDER if ($globe::m_info{_array_order}[$i]{dir} !~ /^[\s]*[\w]/); # if empty don't do.

      $book_or_cref = "$globe::m_info{_array_order}[$i]{dir}";

      if (0) {
         print "ggg $book_or_cref\n";
      }
      $book_or_cref = "$globe::script_prefix$book_or_cref.script";
      
      if (open ( OUT_MNAV, "$globe::path$book_or_cref")) {
         close (OUT_MNAV);
         $holder = &creating_master_book_toc("$globe::path$book_or_cref", $globe::m_info{_array_order}[$i]{dir});
         $book_or_cref = "$globe::script_master_prefix$book_or_cref";
         if (open ( OUT_MNAV, ">$globe::path$book_or_cref")) {
            print (OUT_MNAV "$holder");
            close (OUT_MNAV);
            $section .= "Item level=1 ";
            $section .= "image=$globe::nav_folderclosed ";
            $section .= "selected=YES ";
            $section .= "subscript=$book_or_cref ";
            $section .= "text=$globe::m_info{_array_order}[$i]{title}\r\n";
         } else {
            push (@file_errors, "Cannot open file \"$globe::path$book_or_cref\"");
         }
      } else {
         push (@file_errors, "Cannot open file \"$globe::path$book_or_cref\"");
      }
      
      if (0) {
         print "ggg $book_or_cref\n";
      }
   } # for each file in the order

   $out_file = "m_tree.script";
   
   if (open ( OUT_MNAV, ">$globe::path$out_file")) {
      print (OUT_MNAV "$section");
      close (OUT_MNAV);
      return (0);
   } else {
      push (@file_errors, "Cannot open file \"$globe::path$out_file\"");
      return (1);
   }

} # create_nested_tree_scripts


##############################################################################
##! @fn int creating_master_book_toc
##! @brief Reads in the tree file associated with the master order of a book 
##! and updates path information appropriate for plug-in to a big system.
##!
##! @param temp defines the file name of the original tree file.
##! @param up_path is the updated path to place on URL references.
##!
##! @retval temp_file_buf is a fully formatted and tagged section that can be
##! placed directly into a master script file.
##!
##! @bug None
##!
##! @ingroup tp_toc
##############################################################################
sub creating_master_book_toc {
   $temp = $_[0];
   $up_path = $_[1];
   
   # $temp = sprintf ("%s%s", $globe::script_master_prefix, $temp);
   unless (open ( IN_HTREE, "$temp")) {
      push (@file_errors, "Cannot open file \"$temp\"");
      return ("ERROR");
   }

   local ($temp_file_buf) = "";
   while (<IN_HTREE>){
      $temp_file_buf .= $_;
      # print "incoming = $_";
   }
   close (IN_HTREE);

   # Fix up the url paths.
   $temp_file_buf =~ s/url\=/url\=$up_path\//g;
   
   return ($temp_file_buf);
} # creating_master_book_toc




##############################################################################
##! @fn int create_master_tree_script
##! @brief Creates a single script output files that represents the
##! expanding/collapsing elements of the table of contents.
##!
##! @param global_variables are used throughout.
##!
##! @return A single master tree (m_tree.script) file that 
##! makes up the navigation.
##!
##! This is the old method that is no longer preferred, because both
##! book and cref directories will now have unique tree files. It is
##! better to do a nested design.
##! 
##! This creates two types of mini-TOC's. <ol>
##! <li>When it is known that one of them
##! is a book (e.g., from FrameMaker) and had a tree.html file generated
##! by voyant_nav.pl, then it uses that mini-TOC here.</li>
##! <li>When it is something else like a code reference item, it generates
##! a rigid 8 item mini-TOC. It assumes that this piece is for a code
##! reference item and that it is plugging into Doxygen output where
##! certain files are given from the start.</li></ol>
##!
##! @bug This assumes either that a mini-TOC is available that can be
##! massaged and plugged in, or that the desired piece is a code reference
##! section which plugs into Doxygen.
##!
##! The main cue is having "book" in the directory name.
##!
##! @ingroup tp_toc
##############################################################################
sub create_master_tree_script {
   local ($key) = "";
   local ($out_file) = "";
   local ($section);
   print "Entering create_master_tree_script...\n";
   
   local ($master_script) = "";
   


   #####
   # Go through all items in the master order.
   # We are only worrying about the things I know about
   #####
   TOORDER: for ($i=0; $i <= $#{$globe::m_info{_array_order}}; $i++){
      next TOORDER if ($globe::m_info{_array_order}[$i]{dir} !~ /^[\s]*[\w]/); # if empty don't do.

      $book_or_cref = "$globe::m_info{_array_order}[$i]{dir}";
      # print "Input to use is $book_or_cref\n";

      $section = "";
      
     TIORDER: for ($j=0; $j <= $#{$globe::m_info{_array_order}}; $j++){
         # if it is empty, don't put an entry in.
         next TIORDER if ($globe::m_info{_array_order}[$j]{dir} !~ /^[\s]*[\w]/);
         if ($i == $j) {
            if ($book_or_cref =~ /book/i) {
                &creating_book_toc();
            } else {    # if ($book_or_cref !~ /book/i)
               # Create a real stupid second pane navigation
               $section .= "Item level=1 image=nav_folderclosed.gif selected=YES text=$globe::m_info{_array_order}[$j]{title}\r\n";
               $section .= "Item level=2 image=nav_doc.gif url=$globe::m_info{_array_order}[$j]{dir}/main.html,basefrm selected=YES text=Main Page\r\n";
               $section .= "Item level=2 image=nav_doc.gif url=$globe::m_info{_array_order}[$j]{dir}/modules.html,basefrm selected=YES text=Groups\r\n";
               $section .= "Item level=2 image=nav_doc.gif url=$globe::m_info{_array_order}[$j]{dir}/globals.html,basefrm selected=YES text=Globals [C++, IVE] (Functions, Enum, Struct, Class, Define, etc.)\r\n";
               $section .= "Item level=2 image=nav_doc.gif url=$globe::m_info{_array_order}[$j]{dir}/annotated.html,basefrm selected=YES text=Classes (annotated)\r\n";
               $section .= "Item level=2 image=nav_doc.gif url=$globe::m_info{_array_order}[$j]{dir}/classes.html,basefrm selected=YES text=Classes (listed)\r\n";
               $section .= "Item level=2 image=nav_doc.gif url=$globe::m_info{_array_order}[$j]{dir}/hierarchy.html,basefrm selected=YES text=Class Hierarchy\r\n";
               $section .= "Item level=2 image=nav_doc.gif url=$globe::m_info{_array_order}[$j]{dir}/functions.html,basefrm selected=YES text=Class Member Functions\r\n";
               $section .= "Item level=2 image=nav_doc.gif url=$globe::m_info{_array_order}[$j]{dir}/files.html,basefrm selected=YES text=Files in Project\r\n";

            } # if ($book_or_cref =~ /book/i)
         } else {  # if ($i != $j)
            # Gets directory without path
            @chunk = split ( /\//, $globe::m_info{_array_order}[$j]{dir});
            $m_file = $chunk[@chunk-1];
            undef (@chunk);
            $m_file =~ s/\.js/\.html/;
            # Write out link to m_ html file
            if (0) {
               $section .= "Item level=1 image=nav_doc.gif url=m_tree_$m_file.html\#$j,basefrm selected=YES text=$globe::m_info{_array_order}[$j]{title}\r\n";
            }
         } # if ($i == $j)
      } # for $j
      if (($i =~ 3) && (0)) {
          print "Section $i:\n$section\n";
          exit(1);
      }
      # swap out with the actual stuff
      $globe::master_tree_html = $section;
      $master_script .= $section;
      if (0) {
         print $globe::master_tree_html;
      }

      &script_output();
         
   } # for each file in the order


   $out_file = "m_tree.script";

   unless (open ( OUT_MNAV, ">$globe::path$out_file")) {
      print "Cannot open file \"$globe::path$out_file\"\n";
      exit(1);
   }

   print (OUT_MNAV "$master_script");
   close (OUT_MNAV);


} # create_master_tree_script


##############################################################################
##! @fn int creating_book_toc
##! @brief Reads in the tree file associated with the master order of a book and
##! creates the table of contents section for it.
##!
##! @param Varables defines aboved.
##! @param IN_HTREE files that were generated from the voyant_nav.pl
##! program. Moreover, these are expected to begin with "tree_" and then
##! contain the path of where they came from.
##!
##! @retval $section is a fully formatted and tagged section that can be
##! placed directly into a template.
##!
##! This is the old way called by create_master_tree_script, which itself
##! is no longer called.
##!
##! This was originally part of the TIORDER loop above and still uses
##! variables as if still above. It was broken out to provide more clarity
##! particularly when adding collapsing/expanding chapters for a book.
##!
##! @bug This uses not only globe variables, but also local variables
##! defined in create_master_tree_script.
##! This expects the tree_...html files contain the path as part of the
##! name.
##!
##! @ingroup tp_toc
##############################################################################
sub creating_book_toc {
   $temp = sprintf ("%stree_%s.script", $globe::path, $globe::m_info{_array_order}[$i]{dir});
   unless (open ( IN_HTREE, "$temp")) {
      print "Cannot open file \"$temp\"\n";
      next TIORDER;
   }
   # print "Can open file \"$temp\"\n";

   local ($temp_file_buf) = "";
   while (<IN_HTREE>){
      $temp_file_buf .= $_;
      # print "incoming = $_";
   }
   close (IN_HTREE);

   # Fix up the url paths.
   $temp_file_buf =~ s/url\=/url\=$globe::m_info{_array_order}[$i]{dir}\//g;
   
   # @f_sect = split ( /url\=/i, $temp_file_buf);
   # foreach $t_sect (@f_sect){
   #    $t_sect = "$globe::m_info{_array_order}[$i]{dir}/$t_sect";
   # }
   # $temp_file_buf = join ("url=", @f_sect);
   $section .= $temp_file_buf;

   undef ($temp_file_buf);
}


##############################################################################
##! @fn int script_output
##! @brief Breaks up the book tree files into even smaller chapter
##! tree files.
##!
##! @param global_variables
##!
##! @return Returns a series of HTML files that represent the
##! expanding/collapsing of chapters within a book.
##!
##! This is the old way called by create_master_tree_script, which itself
##! is no longer called.
##!
##! It handles the naming and the writing of all tree 
##! navigation files. It changes icons where appropriate.
##!
##! The entire mini-TOC for the book was already available. This accomplishes
##! the task of creating chapter chunks that expand/collapse by stripping
##! all the topics from other chapters except the top level. The top level is
##! left, but changes its icon to be a closed book and changes the hyperlink
##! to point to the appropriate master tree file associated with that chapter.
##!
##! @bug Upon entering, it is known that this is a book file. In fact,
##! the entire mini-TOC for that book is already available.
##!
##! @ingroup tp_toc
##############################################################################
sub script_output {
   # Test to see if it is a book file
   $out_file = "m_tree_$globe::m_info{_array_order}[$i]{dir}.script";

   unless (open ( OUT_MNAV, ">$globe::path$out_file")) {
      print "Cannot open file \"$globe::path$out_file\"\n";
      exit(1);
   }

   print (OUT_MNAV "$globe::master_tree_html");
   close (OUT_MNAV);
} # script_output



##############################################################################
##! @fn int using_voy_mt_app
##! @brief What to do when no arguments are given.
##! @param None
##! @return None
##! 
##! @bug None
##! @ingroup tp_toc
##############################################################################
sub using_voy_mt_app {
   print "\nvoyant_mt_app.pl creates content navigation to be displayed in the navigation \n";
   print "pane. It assumes that all tree files have been copied into the \n";
   print "input directory and are named uniquely. The tree files\n";
   print "were generated by voyant_nav.pl or tree_js_2_script.pl.\n";
   print "The tree files are mini-TOC files.\n";
   print "\n-- All tree files must begin with \"tree_\" and have .script extensions.\n";
   print "-- All files must reside in the input directory.\n\n";
   print "-- The first argument is the directory for tree files including the forward slash (\/).\n";
   print "   If no path is specified as the first argument, \".\/\" is used.\n";
   print "-- The second argument is the master file to use for the navigation.\n";
   print "   It must have sections for voyant_order.\n";
   print "\nThe output are files begin with \"m_tree_\".\n";
   print "The m_tree.script output is displayed in a toc frame while the content it controls\ndisplays in the basefrm.\n";
   print "\nTerminating voyant_mt_app.pl without doing anything.\n";
   return;
}



##############################################################################
##! @fn int END
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub END {
   undef ($_file_list);     # "_file_list";
   undef ($_index_file); # = "_index_list";
   undef ($in_file);     # "";

   
   ##############################################################################
   ## Memory clean-up.
   ##############################################################################
   if ($no_scope_file > 0){
      &globe::memory_clean_up();
   }

   if (@file_errors) {
      print "\n============  Summary of errors ==================================\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
   }

   print "\n============  Finished voyant_mt_app.pl ==================================\n";
} # END




