#!/usr/local/bin/perl
##############################################################################
##! @file
##! @brief Swaps out tagged areas in all HTML files in a given directory.
##! 
##! @param $globe::path location to find the HTML files. 
##! The name should be terminated with a slash (\). 
##! Although the directory_name is the first command line parameter,
##! the true input are the HTML files contained within that directory.
##!
##! @param $globe::master_nav_file [optional path and] filename for the 
##! HTML file to use as the master for information.
##! This file has several specially flagged
##! HTML comment sections that are required. Information from
##! the tagged sections are lifted and placed into the tagged sections
##! of the input HTML files.
##!
##! @retval input_html_file is modified with new information in the tagged
##! areas. This includes the information in the head, navigation at the top
##! of the body, and copyright information at the bottom of the body.
##!
##! @retval _file_list is a temporary file with a list of all HTML files 
##! in the given directory.
##!
##! @retval _index_list is a temporary file with extracted index tokens.
##!
##! @retval tree.html is a temporary file with a mini-TOC (table of contents)
##! for the given book.
##!
##! This program does most of the work in creating a comprehensive HTML
##! system that spans the mini-HTML systems generated by Doxygen and Mif2Go
##! (from FrameMaker source). The reason it does most of the work is that
##! it has to read the files anyway. 
##! 
##! It started out as a tool just to
##! swap out the header, navigation, and copyright areas. 
##! However, as long as it was visiting each and every HTML file in the
##! input directory and "knew" what was in those files and how the files
##! related to one another, this tool was expanded to handle generating
##! a mini-TOC file () and an index file. It: <ul>
##!
##! <li>Reads in the master file to acquire information from tagged areas.</li>
##! 
##! <li>Creates a list of HTML files in the input directory.</li>
##! 
##! <li>Opens each HTML file and swaps out common navigation with that coming 
##! from the master file.</li></ul>
##! 
##! Before writing a _temp file (and then copying this back over the 
##! original input file) and as it processes the file: <ul>
##! 
##! <li>It creates several hash tables that ultimately determine topic
##! order, topic level in tree, etc.</li>
##! 
##! <li>It generates a mini-TOC file (tree.html) before finishing.</li>
##! 
##! <li>It processes object tokens coming from the MS-HTML output of
##! Mif2Go to turn them into index token comments. It also writes these
##! tokens to a temporary file _index_file.</li>
##! 
##! <li>From the topic hash tables, it generates more index tokens for 
##! the _index_file.</li>
##! </ul>
##!
##! @bug In order for this to work correctly, the file names for the chapters
##! have to be named correctly so that they sort into a natural order. This
##! is achieved by having the generation place the same prefix on all HTML
##! files for a given chapter. I chose a two-digit number followed by an
##! underscore. In this manner, all files for a given chapter are grouped
##! together by the prefix. Earlier processing of the files determined which
##! of those topics was the first and last for the chapter.
##!
##! Rather than passing in variables which can create copies in memory, many items
##! use global variables that are defined in globe.pm. When a
##! variable is known to be global, its name begins with "$globe::". The intent is
##! to facilitate maintenance by having all user-defined tags in one place outside 
##! of the program.
##!
##! This used information that came specifically from our implementation
##! of styles in FrameMaker. In particular, we know what the format names are
##! that are used for chapter headings. The number is picked off of that.
##!
##! @ingroup tp_tools tp_nav
##!
##! @author Glenn C. Maxey
##!
##    $Id: voyant_nav.pl,v 1.45 2002/11/20 15:23:08 gmaxe Exp $
##
## 2002 Created by Voyant Technologies, Inc., Westminster, Colorado, USA.
##
## Permission to use, copy, modify, and distribute this software and its 
## documentation under the terms of the GNU General Public License is hereby 
## granted. No representations are made about the suitability of this software 
## for any purpose. It is provided "as is" without express or implied warranty. 
## See the GNU General Public License (http://www.gnu.org/copyleft/gpl.html) 
## for more details.
## 
## Documents produced by this script are derivative works derived from the 
## input used in their production; they are not affected by this license.
##
##    Revision Information:
##
##    $Log: voyant_nav.pl,v $
##    Revision 1.45  2002/11/20 15:23:08  gmaxe
##    Added exit codes so that wrapper scripts can catch errors properly.
##
##    Revision 1.44  2002/11/20 01:49:14  gmaxe
##    Minor changes to exit codes.
##
##    Revision 1.43  2002/07/26 18:56:13  gmaxe
##    Got rid of old definitions and migrated new structures into all;
##    enhanced xhelp output file names and format; now everything is
##    alphebetized.
##
##    Revision 1.42  2002/07/22 23:05:43  gmaxe
##    Added additional input parameter for project information.
##    Needed for proper ordering.
##
##    Revision 1.41  2002/07/03 23:07:23  gmaxe
##    Added support for mif2go index tokens. Later, remove the old
##    style index tokens and objects.
##
##    Revision 1.40  2002/07/02 23:10:54  gmaxe
##    Tweaked to get rid of extra expanding line feeds in voyant_nav.pl.
##    Added topper.html and other files to skip list in globe.pm.
##    Enhanced log_html_gen.gl to handle new format of message.txt;
##    it now has a flatter data structure and no enums numbers.
##
##    Revision 1.39  2002/06/24 19:14:09  gmaxe
##    Added support for Decision Tree: involved new FM format as well as
##    new handling of HTML pages with multiple H1 targets. Had to create
##    the appropriate entries in the first/last arrays as well as
##    get the tracing of the output arrays proper.
##    PTR#:
##
##    Revision 1.38  2002/05/08 15:26:57  gmaxe
##    Minor scoping error on Order key.
##
##    Revision 1.37  2002/04/24 00:08:56  gmaxe
##    Improved how items are excluded from xhelp_gen.pl, because
##    they were creeping back in when file got processed.
##
##    Revision 1.36  2002/04/12 18:01:15  gmaxe
##    Various tweaks; work on tag extraction and change_nav;
##    more exclused special perl characters from indexer;
##    improved support for sapi group pm files in xhelp.
##    spider tool now works regularly and as expected.
##
##    Revision 1.35  2002/04/06 01:55:18  gmaxe
##    New files and new general constructs for handling tags. Uses globals better.
##    The html_look* files are designed to handle spider tracing of html systems.
##
##    Revision 1.34  2002/03/13 18:54:29  gmaxe
##    Minor improvement to skip proper files and definitions.
##
##    Revision 1.33  2002/03/05 01:47:42  gmaxe
##    Supports previous/next on Doxygen files;
##    supports better script generation.
##
##    Revision 1.32  2002/02/28 00:41:23  gmaxe
##    Added support for nested script files; not yet turned on.
##    Re-arrange file; created more subroutines.
##
##    Revision 1.31  2002/02/16 00:59:46  gmaxe
##    Added support for script files needed for TOC applet.
##
##    Revision 1.30  2002/02/12 01:50:29  gmaxe
##    Added support for more level chunking;
##    added files to support xhelp;
##    added file to support generating TOC, index from lists.
##    PTR#:
##
##    Revision 1.29  2002/02/02 01:48:54  gmaxe
##    Supports index tokens from all dox files;
##    supports used-by index tokens.
##    asn_bapitypes.pl is a new file to handle remerging of comments
##    into what the tools see.
##
##    Revision 1.28  2002/01/16 19:49:10  gmaxe
##    Removed much of revision history in comments.
##
##    Revision 1.27  2002/01/11 21:53:55  gmaxe
##    Updated comments.
##
##############################################################################


##############################################################################
##! @fn int BEGIN
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub BEGIN {
   print "\n============  Starting voyant_nav.pl ==================================\n";
   $_file_list = "_file_list";
   $_index_file = "_index_list";
   $scope_pm = "globe.pm";
   $_arg_inc = 0;
   $in_file = "";
   $f_type = "htm";
   $no_scope_file = 0;

   push (@INC, `pwd`);
   push (@INC, '../perl');
   push (@INC, '/rtfm/techpubs/perl');
   if (0){
      print (@INC, "\n");
   }
   #####
   # All global variables are defined in the following file
   #####
   unless (open ( IN_LIST, $scope_pm)) {
      unless (open ( IN_LIST, "../perl/$scope_pm")) {
         unless (open ( IN_LIST, "/rtfm/techpubs/perl/$scope_pm")) {
            push (@file_errors, "Cannot open file \"$scope_pm\" or \"../perl/$scope_pm\" or \"/rtfm/techpubs/perl/$scope_pm\"\n");
            $no_scope_file++;
         }
      }
   }
   close (IN_LIST);
   push (@INC, $scope_pm);
   push (@INC, "../perl/$scope_pm");
   push (@INC, "/rtfm/techpubs/perl/$scope_pm");

   if (!@file_errors) {
      #####
      # All global variables are defined in the following file
      #####
      require $scope_pm;
   
      if (&globe::declare_variables()) {
         print "Variables initialized from $scope_pm.\n";
      } else {
         push (@file_errors, "Could not initialize variables from $scope_pm.\n");
      }
   } # if not @file_errors

   if (@ARGV < 1) {
      push (@file_errors, "No path specified as first argument.\n");
   }
   
   # Get path if there is one.
   if (@ARGV > $_arg_inc) {
      $globe::path = @ARGV[$_arg_inc];
      if ($globe::path =~ /\//) {
         print "The path specified is $globe::path\n";
      
         @path_chunks = split ( /\//, $globe::path);
         if (@path_chunks < 2) {
            $globe::rel_path_to_start_point = "./";
         } else {
            $globe::rel_path_to_start_point = "../";
         }
      } else {
         push (@file_errors, "The input argument \"$globe::path\" requires a forward slash (\/).\n");
      }
   } # if 1 or more arguments
   $_arg_inc++;
   
   if (@ARGV > $_arg_inc) {
      $globe::master_nav_file = @ARGV[$_arg_inc];
      print "The master file is $globe::master_nav_file\n";
      unless (open ( IN_MASTER, $globe::master_nav_file)) {
         push (@file_errors, "Cannot open file \"$globe::master_nav_file\"\n");
      }
      # Get the master definitions
      while (<IN_MASTER>){    # entire master file into memory.
         $globe::master_nav .= $_;
      }
      close (IN_MASTER);
   } # if 2 or more arguments
   $_arg_inc++;

   if (@ARGV > $_arg_inc) {
      $globe::master_nav_file2 = @ARGV[$_arg_inc];
      print "The master file is $globe::master_nav_file2\n";
      unless (open ( IN_MASTER, $globe::master_nav_file2)) {
         push (@file_errors, "Cannot open file \"$globe::master_nav_file2\"\n");
      }
      # Get the master definitions
      while (<IN_MASTER>){    # entire master file into memory.
         $globe::master_nav .= $_;
      }
      close (IN_MASTER);
   } # if 2 or more arguments
   $_arg_inc++;
}

##############################################################################
##! @fn int main
##! @brief The main program.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
# sub main {
{
   ##############################################################################
   ## Program start
   ##############################################################################

   if (0){
      print "=== Definitions 3 \n";
      print "=== header ==== $globe::m_define{header}[0]\n";
      exit(1);
   }
   if (@file_errors) {
      # Makes no sense to go on if input parameters are off.
      print "\n============  Summary of errors ==================================\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
      &using_voy_nav();
      exit(1);
   }

   if (&globe::get_master_nav_info() =~ 1){
      print "Master Navigation file \"$globe::master_nav_file\" has definition errors.\n";
      if (@file_errors) {
         print "\n============  Summary of errors ==================================\n";
         for ($i=0; $i<@file_errors; $i++){
            print "$i = $file_errors[$i]\n";
         }
      }
      exit(1);
   } else {
     # Create special navigation in the footer that repeats buttons.
     $globe::voy_footer_fm = "$globe::m_define{footer}[0]";
     $globe::voy_footer_dox = "$globe::m_define{footer}[0]";
     $def_type = "nav_book";
     ($before, $piece, $after) = &globe::get_tag_chunk( $globe::m_info{$def_type},
                                 $globe::m_define{$def_type}[0], 
                                 $globe::m_define{$def_type}[1], 
                                 $very_critical);

     if ($piece) {
        $globe::voy_footer_fm .= "\n<br><br>\n$piece\n";
     }
     $def_type = "nav_doxygen";
     ($before, $piece, $after) = &globe::get_tag_chunk( $globe::m_info{$def_type},
                                 $globe::m_define{$def_type}[0], 
                                 $globe::m_define{$def_type}[1], 
                                 $very_critical);
     if ($piece) {
        $globe::voy_footer_dox .= "\n<br><br>\n$piece\n";
     }
     
     $def_type = "footer";
     ($before, $piece, $after) = &globe::get_tag_chunk( $globe::m_info{$def_type},
                                 $globe::m_define{$def_type}[0], 
                                 $globe::m_define{$def_type}[1], 
                                 $very_critical);
     
     if ($piece) {
        $globe::voy_footer_fm .= "$piece";
     }
     $globe::voy_footer_fm .= "$globe::m_define{footer}[1]";
     if ($piece) {
        $globe::voy_footer_dox .= "$piece";
     }
     $globe::voy_footer_dox .= "$globe::m_define{footer}[1]";

     
     # 4/22/2002 Changed because introduced an off-by-one error.
     # $globe::master_order_key = @globe::master_order;  # sets the default value
     $globe::master_order_key = $#globe::master_order;  # sets the default value
     for ($i=0; $i < $#globe::master_order; $i++){
        # does not go to the last one, because that is the default, already set
        if ($globe::path =~ /$globe::master_order[$i]/){
           $globe::master_order_key = $i;
           if (1) {
              print "Order key $i with directory \"$globe::master_order[$globe::master_order_key]\"\n";
              print "Title is \"$globe::master_order_title[$globe::master_order_key]\"\n";
           }
        }
     }
     if (0) {
        print "=== Master Header ==\n$globe::m_info{header}\n";
        print "=== Master Common Top ==\n$globe::m_info{nav_common}\n";
        print "=== Master Doxygen ==\n$globe::m_info{nav_doxygen}\n";
        print "=== Master FM Book ==\n$globe::m_info{nav_book}\n";
        print "=== Master Footer ==>$globe::m_info{footer}<==\n";
        print "=== Master Dox Footer ==>$globe::voy_footer_dox<==\n";
        print "=== Master FM Footer ==>$globe::voy_footer_fm<==\n";
        print "=== Master variables ==\n$globe::m_info{variable}\n";
        exit(1);
     }
   }
   
   $_file_list = "$globe::path$_file_list";
   
   ###################
   # Does most of work and calls change_nav.
   ###################
   if (!&generate_and_step_through_list ($_file_list)) {
      exit(1);
   }
   
   ###################
   # Take care of file sequence... previous next
   ###################
   if ($globe::path =~ /cref/i) {
      if (1){
         print "Starting updating previous\/next designation on all files.\n";
         &do_dox_prev_next();
         print "Finished updating previous\/next designation.\n";
      }
      
   }

   ###################
   # Take care of index tokens
   ###################
   if (@globe::index_info) {
      unless (open ( OUT_INDEX_LIST, ">$globe::path$_index_file")) {
         print "Cannot open file \"$globe::path$_index_file\"\n";
         exit(1);
      }
      for ($i=0; $i<@globe::index_info; $i++){
         # print "$i = $globe::index_info[$i]\n";
         print (OUT_INDEX_LIST "$globe::index_info[$i]\n");
      }
      close (OUT_INDEX_LIST);
   }
   
   &organize_topic_order();

   ##############################################################################
   ## End of Program
   ##############################################################################
   exit(0);
} # the main routine

##############################################################################
##! @fn int using_voy_nav  ( )
##! @brief What to do when no arguments are given.
##! @param None
##! @return None
##! 
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub using_voy_nav {
   print "\nvoyant_nav.pl operates on specific Doxygen\mif2go output files from \n";
   print "various Doxygen runs. \n";
   print "dox_nav's purpose is to give all HTML files the same navigation bar.\n";
   print "\n-- These files must be named with <htm> or <html> as the extension.\n";
   print "-- All files must reside in the input directory.\n\n";
   print "-- The first argument is the path including the forward slash (\/).\n";
   print "   If no path is specified as the first argument, \".\/\" is used.\n\n";
   print "-- The second argument is file to use for the navigation.\n";
   print "   It must have sections for voyant_header, common_top, doxygen_group,\n";
   print "   fm_book, and voyant_footer.\n";
   print "   A section has a begin and end, such as:\n";
   print "     <!-- begin voyant_footer definition -->\m";
   print "     <!-- end voyant_footer definition -->\m";
   print "-- The third argument is file to use for the toc ordering (project_toc.txt).\n";
   print "NOTE: The PDF file placed in \"print_pdf\" has to be named the same as\n";
   print "      the directory containing its associated HTML files.\n";
   print "\nTerminating voyant_nav.pl without doing anything.\n";
   return;
}

##############################################################################
##! @fn int generate_and_step_through_list  ( )
##! @brief Generates list of files to look at.
##! @param None
##! @return None
##! 
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub generate_and_step_through_list {
   local ($_file_list) = $_[0];
   
   if (system ("ls $globe::path*.$f_type* > $_file_list")) {
      print "ls $globe::path*.$f_type* > $_file_list\n";
      push (@file_errors, "There is no $globe::path directory. Nothing done.\n");
      return(0);
   }

   unless (open ( IN_LIST, $_file_list)) {
      push (@file_errors, "Cannot open file \"$_file_list\"\n");
      return(0);
   }
   
   undef (@globe::sequence);
   while (<IN_LIST>){    # read a line from file into $_
      $in_file = $_;
      $in_file =~ s/\n//;
      $skip_flag = 0;
      for ($i=0; $i<@globe::special_files; $i++) {
         if ($in_file =~ /\/$globe::special_files[$i]/i) {
            # slash was added to name to indicate a path so that it won't
            # skip on files that only have part of the name.
            $skip_flag = 1;
         }
      } # for list of special files
      for ($i=0; $i<@globe::exclude_prefix; $i++) {
         if ((($in_file =~ /html$/i) || ($in_file =~ /htm$/i)) && ($in_file =~ /$globe::exclude_prefix[$i]/i)) {
            $skip_flag = 1;
         }
      } # for list of special files
      if ($skip_flag == 1) {
         # skip this file or do something else
         printf ("Skipping... %s\n", $in_file);
      } else {
         printf ("Working on... %s\n", $in_file);
         push (@globe::sequence, $in_file);
         if ($f_type =~ /htm/){
            &change_nav($in_file);
         }
      } # if special file
   } # while getting a special file
   close (IN_LIST);
   return (1);
} # generate_and_step_through_list

##############################################################################
##! @fn int do_dox_prev_next  ( )
##! @brief Does a cheap previous/next for Doxygen files.
##! @param None
##! @return None
##! 
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub do_dox_prev_next {
   local ($t_prev) = "";
   local ($t_next) = "";
   
   for ($j=0; $j < $#globe::sequence; $j++) {
      if (0){
         print "curr = $globe::sequence[$j]\n";
      }
      if ($j == 0) {
         $temp_n = $#globe::sequence;
         @f_name = split (/\//, $globe::sequence[$temp_n]);
         $temp_n = $#f_name;
         $t_prev = $f_name[$temp_n];
         $temp_n = $j+1;
         @f_name = split (/\//, $globe::sequence[$temp_n]);
         $temp_n = $#f_name;
         $t_next = $f_name[$temp_n];
      } elsif ($j == $#globe::sequence-1){
         $temp_n = $j-1;
         @f_name = split (/\//, $globe::sequence[$temp_n]);
         $temp_n = $#f_name;
         $t_prev = $f_name[$temp_n];
         $temp_n = 0;
         @f_name = split (/\//, $globe::sequence[$temp_n]);
         $temp_n = $#f_name;
         $t_next = $f_name[$temp_n];
      } else {
         $temp_n = $j-1;
         @f_name = split (/\//, $globe::sequence[$temp_n]);
         $temp_n = $#f_name;
         $t_prev = $f_name[$temp_n];
         $temp_n = $j+1;
         @f_name = split (/\//, $globe::sequence[$temp_n]);
         $temp_n = $#f_name;
         $t_next = $f_name[$temp_n];
      }
      if (0) {
         print "curr = $globe::sequence[$j]\nprev = $t_prev\tnext = $t_next\n";
      }
      if (open ( IN_LIST, $globe::sequence[$j])) {
         # may end up opening and closing too much.
         undef ($globe::entire_file);
         while (<IN_LIST>){
            $globe::entire_file .= $_;
         }
         close (IN_LIST);
         $globe::entire_file =~ s/$globe::voy_variable_prev_dox/$t_prev/g;
         $globe::entire_file =~ s/$globe::voy_variable_next_dox/$t_next/g;
         if (open ( IN_LIST, ">$globe::sequence[$j]")) {
            print (IN_LIST "$globe::entire_file");
         }
         close (IN_LIST);
      } # if open
   } # for @sequence
   return (1);
} # do_dox_prev_next


##############################################################################
##! @fn int change_nav
##! @brief Opens up an HTML file and replaces information between known tags.
##!
##! @param $in_file is the HTML file to edit. 
##!
##! @return Effectively writes back out to the same HTML file.
##!
##! This function performs all tasks associated with updating the individual
##! HTML file. It reads the entire HTML file into memory. 
##! 
##! In general, this routine looks for paired HTML comments containing a
##! tag name and "begin" or "end". Whatever is between these paired comments
##! is replaced with the updated information extracted from the master.
##! The sequence of replacements is: <ul>
##! 
##! <li>If the HTML file has $globe::m_define{variable}[0], it means that it
##! came from a FrameMaker book. As such, these variables contain information
##! regarding the previous topic, next topic, whether or not it is a first topic
##! or last topic for the chapter, etc. This information is extracted from the
##! HTML file and placed into variables to be used by this topic.</li>
##! 
##! <li>Info in the file header is swapped out. If the tags aren't found,
##! an appropriate spot for the new tags is created before the terminating
##! head tag.</li>
##! 
##! <li>Info in the file footer is swapped out. If the tags aren't found,
##! an appropriate spot for the new tags is created before the terminating
##! body tag. </li>
##! 
##! <li>If the HTML file came from a FM book, replace the navigation bar
##! associated with the book (the second navigation bar.)
##! Although this navigation bar appears below the common navigation bar, it is
##! swapped out first, because if the location tags aren't found, this routine finds
##! an appropriate spot for the new tags immediately following the body tag.
##! This is easy to spot.</li>
##! 
##! <li>If the HTML file came from Doxygen, replace the navigation bar 
##! associated with Doxygen topics (the second navigation bar.)
##! Although this navigation bar appears below the common navigation bar, it is
##! swapped out first, because if the location tags aren't found, this routine finds
##! an appropriate spot for the new tags immediately following the body tag.
##! This is easy to spot.</li>
##! 
##! <li>Replace the navigation bar common to all topics in the system.
##! If the location tags aren't found, this routine finds an appropriate spot
##! for the new tags immediately following the body tag. This is why this needs
##! to happen after inserting either the FM navigation or the Doxygen navigation.</li>
##! 
##! <li>Sometimes HTML is generated by Doxygen or Mif2Go that hard-codes fonts
##! or styles that cannot be controlled by a CSS. 
##! One of the lists that can be defined in the master file is a "zap"
##! list. This is read into @globe::voy_html_zap hash. Each item in the zap list has
##! three elements:<ol>
##! <li>start of the offending tag but incomplete.</li>
##! <li>end of first offending tag.</li>
##! <li>end of tag pairs.</li></ol>
##! The HTML tags defined in the zap list get taken out of the HTML document.</li>
##! 
##! <li>Items in the header, footer, or navigation have items coming from the master
##! that were not fully defined. Part of the voy_order provides the complete
##! description. The HTML document gets the complete definition of these items.
##! This includes: </li><ul>
##! <li>PDF file name.</li>
##! <li>The name of the manual or group.</li>
##! <li>The document number.</li>
##! <li>Document title is changed to match the first heading.</li></ul>
##! </ul>
##!
##! <li>The title is extracted into the index token list.</li>
##! 
##! <li>The headings are added to the index token list as well as the topic order
##! list.</li>
##! 
##! <li>Object tags are how FM index tokens are passed to MS-HTML using Mif2Go.
##! These are converted into special HTML comments. They are also extracted into
##! the index token list.</li>
##! 
##! <li>If the file came from Doxygen, this parses
##! out interesting items to add to the index token list.</li>
##!
##! While processing the file, hash tables are created: <ul>
##! <li> @globe::book_topics: key is html file name, <Title> is stored</li>
##! <li> @globe::book_topic_order: key is html file name, next html file is stored</li>
##! <li> @globe::book_topic_level: key is html file name, <Headn> is stored</li></ul>
##!
##! This is important in order to handle subsequent creation of the mini-TOC file
##! (tree.html) in another routine.
##!  
##! @bug This is a big routine that accomplishes many different tasks while
##! it has the HTML file in memory. Some of the items have to be handled in the
##! order given for them to be swapped and positioned properly.
##!
##! @ingroup tp_nav
##############################################################################
sub change_nav {
   local ($in_file) = $_[0];  # name of input html file
   local ($rel_to_file) = "";
   local ($out_file) = "_temp";
   local ($out_line) = "";
   
   local ($globe::entire_file) = "";
   local ($globe::have_header) = 0;
   local ($globe::have_footer) = 0;
   local ($globe::have_common_top) = 0;
   local ($globe::have_dox) = 0;
   local ($target_cnt) = 0;
   local ($target_prefix) = "_ggg_";
   
   if (0) {
      printf ("Here is the infile again = %s\n", $in_file);
   }
   
   #### 
   # Change path used in index file
   ####
   $rel_to_file = $in_file;
   @path_chunk = split ( /\//, $rel_to_file, 2);
   if (@path_chunk =~ /2/) {
      $rel_to_file = "$path_chunk[1]";
   }
   @path_chunk = split ( /\//, $in_file);
   if ($#path_chunk > 0) {
      # Although $globe::m_var{value}{curr} is set by mif2go, it is not set by Doxygen.
      $globe::m_var{value}{curr} = "$path_chunk[$#path_chunk]";
   }
      if (0){
         print "glenn $in_file\n";
         print "glenn $rel_to_file\n";
         print "glenn $globe::m_var{value}{curr}\n";
         exit(1);
      }
   ####   

   $out_file = "$globe::path$out_file";
   
   unless (open(IN_HTML, $in_file)) {
      print "Cannot open file \"$in_file\"\n";
      return(1);
   }
   #####
   # Read in entire input file.
   #####
   while (<IN_HTML>) {
      $globe::entire_file .= $_;
   }
   close (IN_HTML);
   
   unless (open(OUT_HTML, ">$out_file")) {
      print "Cannot open file \"$out_file\"\n";
      return(1);
   }

   #####
   # Make sure our input HTML file is well formed
   #####
   $globe::entire_file =~ s/\<\/Head\>/\<\/head\>/g;
   $globe::entire_file =~ s/\<\/HEAD\>/\<\/head\>/g;
   $globe::entire_file =~ s/\<Head\>/\<head\>/g;
   $globe::entire_file =~ s/\<HEAD\>/\<head\>/g;
   $globe::entire_file =~ s/\<\/Body\>/\<\/body\>/g;
   $globe::entire_file =~ s/\<\/BODY\>/\<\/body\>/g;
   $globe::entire_file =~ s/\<Body\>/\<body\>/g;
   $globe::entire_file =~ s/\<BODY\>/\<body\>/g;
   $globe::entire_file =~ s/\<Link rel\=/\<link rel\=/g;
   $globe::entire_file =~ s/\<Link Rel\=/\<link rel\=/g;
   $globe::entire_file =~ s/\<LINK REL\=/\<link rel\=/g;
   $globe::entire_file =~ s/\<Link/\<link/g;
   $globe::entire_file =~ s/\<LINK/\<link/g;

   #####
   # change in structure   
   #####
   $globe::entire_file =~ s/\<\!\-\- END/\<\!\-\- end/g;
   $globe::entire_file =~ s/\<\!\-\- BEGIN/\<\!\-\- begin/g;
   
   # Remove any style sheets that might be hanging out.
   undef(@pre_chunks);
   if ($globe::entire_file =~ /\<link/){
      @pre_chunks = split ( /\<link/, $globe::entire_file, 2);
      @post_chunks = split ( /\>/, $pre_chunks[1], 2);
      $globe::entire_file = sprintf( "%s%s", $pre_chunks[0], $post_chunks[1]);
   }

   # Test the Input file for various items and swap out as needed.
   undef(@pre_chunks);
   undef(@post_chunks);
   
 
   ##################
   # Get the previous and next variables
   # and the current variables
   # Applies only if it is a fm book file.
   ##################
   $not_critical = 0;
   $def_type = "variable";
   if (0) {
      print "$def_type $globe::m_define{$def_type}[0]\n";
   }
   if ( $globe::entire_file =~ /$globe::m_define{$def_type}[0]/ )  {
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                $globe::m_define{$def_type}[0], 
                $globe::m_define{$def_type}[1], 
                $not_critical);
      if ($piece) {
         $globe::m_info{$def_type} = sprintf( "%s%s%s", 
                $globe::m_define{$def_type}[0], 
                $piece, 
                $globe::m_define{$def_type}[1]); 

         VAR_VAL: foreach $var (keys %{$globe::m_var{define}}) {
            if ($var =~ /end/){
               next VAR_VAL;
            }
            ($before, $globe::m_var{value}{$var}, $after) = &globe::get_tag_chunk( $globe::m_info{$def_type},
                   $globe::m_var{define}{$var}, 
                   $globe::m_var{define}{end}, 
                   $not_critical);
            if (0) {
               # print "$def_type $var $globe::m_var{value}{$var} $globe::m_info{$def_type}\n";
               print "Stuff:: $def_type $var $globe::m_var{value}{$var} \n";
            }
         } # foreach $var

      } else {
         push (@file_errors, "$def_type definition messed up.");
      }
      # First and Last flags
      if ($globe::m_var{value}{firstlast} =~ /1\#0/){
         push (@globe::book_first_topic, $in_file);
         $globe::m_var{value}{prev} = $globe::voy_variable_prev_chapter;
      } elsif ($globe::m_var{value}{firstlast} =~ /0\#1/){
         push (@globe::book_last_topic, $in_file);
         # The last topic doesn't really have a next, 
         # except the next chapter which we don't know at this point.
         $globe::m_var{value}{next} = $globe::voy_variable_next_chapter;
      } elsif ($globe::m_var{value}{firstlast} =~ /1\#1/){  # handles single topic chapters
         push (@globe::book_first_topic, $in_file);
         $globe::m_var{value}{prev} = $globe::voy_variable_prev_chapter;
         push (@globe::book_last_topic, $in_file);
         # The last topic doesn't really have a next, 
         # except the next chapter which we don't know at this point.
         $globe::m_var{value}{next} = $globe::voy_variable_next_chapter;
      }
    
      if (0) {
         foreach $var (keys %{$globe::m_var{value}}) {
            print "$var $globe::m_var{value}{$var} $in_file\n"; 
         } # foreach $var
         exit(1);
      }
   }


   if (0){ # debug
      print "$globe::entire_file";
      exit(1);
   } # debug


   ##################
   # voyant header needs to go before </head>
   ##################
   $not_critical = 1;
   $reinsert_tags = 0;
   $def_type = "header";
   if (! &file_chunk_change ($def_type, $not_critical, "</head>", 0, $reinsert_tags, $in_file) ) {
      # Could not put proper chunk in; write out what we have an abort.
      # print (OUT_HTML "$globe::entire_file");
      close (OUT_HTML);
      if (0){ # debug
          print "glenn3\n$globe::entire_file";
          exit(1);
      } # debug
      return(0);
   } # voyant_header

   if (0){ # debug
      print "glenn4\n$globe::entire_file";
      print "oh $def_type $globe::m_info{$def_type}\noh again\n";
      exit(1);
   } # debug
 
   # get rid of any spurious navigation.
   while ($globe::entire_file =~ $globe::m_define{nav_book}[0]){
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              $globe::m_define{nav_book}[0], 
                              $globe::m_define{nav_book}[1],
                              0);
      if ((1) || ($piece)) {
         # Turned this on, because the tags could be empty.
         $globe::entire_file = join ("", $before, $after);
      }
   }
   while ($globe::entire_file =~ /$globe::m_define{nav_doxygen}[0]/){
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              $globe::m_define{nav_doxygen}[0], 
                              $globe::m_define{nav_doxygen}[1],
                              0);
      if ((1) || ($piece)) {
         # Turned this on, because the tags could be empty.
         $globe::entire_file = join ("", $before, $after);
      }
   }
   if (0){ # debug
      print "glenn6\n";
      exit(1);
   } # debug
   while ($globe::entire_file =~ $globe::m_define{nav_common}[0]){
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              $globe::m_define{nav_common}[0], 
                              $globe::m_define{nav_common}[1],
                              0);
      if ((1) || ($piece)) {
         # Turned this on, because the tags could be empty.
         $globe::entire_file = join ("", $before, $after);
      }
   }
   while ($globe::entire_file =~ $globe::m_define{footer}[0]){
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              $globe::m_define{footer}[0], 
                              $globe::m_define{footer}[1],
                              0);
      if ((1) || ($piece)) {
         # Turned this on, because the tags could be empty.
         $globe::entire_file = join ("", $before, $after);
      }
   }
   if (0) {
      print "Finishing up here... premature.\n";
      # print $globe::entire_file;
      print (OUT_HTML "$globe::entire_file");
      close (OUT_HTML);
      system ("cp $out_file $in_file");
      return (0);
   }
   

   ##################
   # voyant footer needs to go before </body>
   ##################
   $not_critical = 1;
   $reinsert_tags = 0;
   $def_type = "footer";
   if ($globe::entire_file =~ /\<\/body\>/) { 
      &file_chunk_change ($def_type, $not_critical, "<\/body>", 0, $reinsert_tags, $in_file);
      # Could not put proper chunk in; write out what we have an abort.
   } else {
      # Does not have proper body
      push (@file_errors, "ERROR: Does not have </body> in $in_file.");
      return (0);
   } # voyant footer

   ##################
   # make footer the unique for FM and doxygen
   ##################
   if ($globe::entire_file =~ /$globe::m_define{variable}[0]/) {
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              $globe::m_define{footer}[0], 
                              $globe::m_define{footer}[1],
                              0);
      if ($piece) {
         $nemp = $globe::voy_footer_fm;
         $nemp =~ s/$globe::m_var{define}{prev}/$globe::m_var{value}{prev}/g;
         $nemp =~ s/$globe::m_var{define}{next}/$globe::m_var{value}{next}/g;
         
         $globe::entire_file = join ("", $before, $nemp, $after);
      }
   } else {
      ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              $globe::m_define{footer}[0], 
                              $globe::m_define{footer}[1],
                              0);
      if ($piece) {
         $globe::entire_file = join ("", $before, $globe::voy_footer_dox, $after);
      }
   } 

   if (0){ # debug
      print "glenn4\n$globe::entire_file";
      print "oh $def_type $globe::m_info{$def_type}\noh again\n";
      exit(1);
   } # debug

   ##################
   # mif2go navigation from the FM book
   # DO BEFORE "common top navigation bar" which will slip between <body> and this.
   # Tweak navigation for this page who knows previous and next
   ##################
   $not_critical = 0;
   $reinsert_tags = 0;


   if ($globe::entire_file =~ /$globe::m_define{variable}[0]/) {
      ##################
      # Swap out the FM navigation...
      ##################
      $def_type = "nav_book";
      # Place holder so that we don't loose it.
      $temp_place_holder = $globe::m_info{$def_type};
      # whereever we see definitions, put in the actual values.
      $globe::m_info{$def_type} =~ s/$globe::m_var{define}{prev}/$globe::m_var{value}{prev}/g;
      $globe::m_info{$def_type} =~ s/$globe::m_var{define}{next}/$globe::m_var{value}{next}/g;
   } else {
      ##################
      # Swap out the Doxygen navigation...
      ##################
      $def_type = "nav_doxygen";
      # Place holder so that we don't lose it.
      $temp_place_holder = $globe::m_info{$def_type};
   } # FM Navigation
   &file_chunk_change ($def_type, $not_critical, "<body>", 1, $reinsert_tags, $in_file);
   
   #####
   # Get back to original state for next file.
   #####
   $globe::m_info{$def_type} = $temp_place_holder;

   if (0){ # debug
      print "glenn4\n$globe::entire_file";
      print "oh $def_type $globe::m_info{$def_type}\noh again\n";
      exit(1);
   } # debug


   if (0) {
      print "Finishing up here... premature.\n";
      # print $globe::entire_file;
      print (OUT_HTML "$globe::entire_file");
      close (OUT_HTML);
      system ("cp $out_file $in_file");
      return (0);
   }

   
   #####
   # Strip out native Doxygen navigation
   #####
   if (1) { # 03/30/2002 new definition of doxygen navigation to remove
      DOX_STRIP: while ($globe::entire_file =~ /$globe::dox_nav_defb3/) {
         $very_critical = 0;
         ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                                 "<center>", 
                                 "</center>",
                                 $very_critical);
         if ($after =~ /$globe::dox_nav_defb3/) {
            # You missed your definition;
            # Rebuild some pieces and try one more time.
            $before .= "\<center\>$piece\<\/center\>";
            ($before2, $piece, $after2) = &globe::get_tag_chunk( $after,
                                 "<center>", 
                                 "</center>",
                                 $very_critical);
            if ($after2 =~ /$globe::dox_nav_defb3/) {
               push (@file_errors, "ERROR: Could not get rid of native dox nav in $in_file.");
               break DOX_STRIP;
            } else {
               # You got it this time and can rebuild.
               $globe::entire_file = join ("", $before, $before2, $after2);
            }
         } else {
            # Rebuild the file
            $globe::entire_file = join ("", $before, $after);
         }
      } # while it has the definition
   } # if new definition
   
    
   ##################
   # voyant common top navigation bar should go after <body>
   # Do AFTER FM or Doxygen navigation
   ##################
   $not_critical = 0;
   $reinsert_tags = 0;
   $def_type = "nav_common";

   if (! &file_chunk_change ($def_type, $not_critical, "<body>", 1, $reinsert_tags, $in_file) ) {
      # Could not put proper chunk in; 
      # Not necessarily a problem.
   } # Common Navigation

   if (0){ # debug
      print "glenn4\n$globe::entire_file";
      print "oh $def_type $globe::m_info{$def_type}\noh again\n";
      exit(1);
   } # debug

   
   
   ##################
   # Handle the PDF file names
   ##################
   $temp = "$globe::rel_path_to_start_point$globe::pdf_dir$globe::master_order_pdf[$globe::master_order_key]";
   $globe::entire_file =~ s/$globe::this_pdf_flag/$temp/g;
   $globe::entire_file =~ s/$globe::rel_path_to_start_point_def/$globe::rel_path_to_start_point/g;
   
   ##################
   # Handle the manual or group names
   ##################
   $temp = $globe::master_order_title[$globe::master_order_key];
   $temp =~ s/[\s]+/\&nbsp\;/g;
   if (0) {
      print "$globe::master_order_key: we think title is: \"$temp\".\n";
      if (0) {
         for ($i=0; $i <= $#{$globe::m_info{_array_order}}; $i++){
               print "$i title $globe::m_info{_array_order}[$i]{title}\n";
         } # each $i
      }
      if (0) {
         for ($i=0; $i <= $#{$globe::m_info{_array_order}}; $i++){
            foreach $k (keys %{$globe::m_info{_array_order}[$i]} ) {
               print "$i $k $globe::m_info{_array_order}[$i]{$k}\n";
            } # each $k key
         } # each $i
      }
   }
   
   if ($globe::entire_file =~ /$globe::xmanual/){
      $globe::entire_file =~ s/$globe::xmanual/$temp/g;
   } 
   if ($globe::entire_file =~ /$globe::xgroup/){
      $globe::entire_file =~ s/$globe::xgroup/$temp/g;
   }
   $globe::entire_file =~ s/$globe::fix_path_flag/$globe::fix_path_to/g;
   
   #####
   # Take care of current file name which might be in header,
   # as specified in template and needed in tree applet.
   #####
   if ($globe::entire_file =~ /$globe::m_var{define}{curr2}/){
      $globe::entire_file =~ s/$globe::m_var{define}{curr2}/$rel_to_file/g;
   } 
   
   
   

   ##################
   # Handle the document number
   ##################
   if ($globe::entire_file =~ /$globe::voy_variable_doc_num/){
      $globe::entire_file =~ s/$globe::voy_variable_doc_num/$globe::m_info{_array_order}[$globe::master_order_key]{num}/;
   } 
   
   
   #################################
   # Remove certain specified HTML tags
   #################################
   if ((@globe::voy_html_zap < 1) || ($globe::voy_html_zap[0] =~ /none/)){
      # Do nothing
      # print "$in_file has no tags to zap.\n";
   } else {
      # Remove the tags in question
      for ($i=0; $i<@globe::voy_html_zap; $i++){
         # Determine start and Stop tags
         @zap_tag = split ( /\,/, $globe::voy_html_zap[$i]);
         # clean up tag
         for ($j=0; $j <@zap_tag; $j++){
            $zap_tag[$j] =~ s/^[\s]*\"//;
            $zap_tag[$j] =~ s/\"[\s]*$//;
            # print "Zap this =$zap_tag[$j]\n";
         }
         $t_cnt=0;
         while ($globe::entire_file =~ /$zap_tag[0]/i){  # case insensitive test
             $t_cnt++;
             print "$t_cnt: Removing everything between \"$zap_tag[0]\" and \"$zap_tag[1]\"\n";
             # First tag part A
             @pre_chunks = split ( /$zap_tag[0]/, $globe::entire_file, 2);
             # First tag part B
             @post_chunks = split ( /$zap_tag[1]/, $pre_chunks[1], 2);
             if (@zap_tag > 2){
                # Second tag
                # Hit the first occurrence of the ending tag
                $post_chunks[1] =~ s/$zap_tag[2]//i; # do it only once, case insensitive
             }
             # rebuild
             $globe::entire_file = join ("", $pre_chunks[0], $post_chunks[1]);
         } # while it still is infested with the tag
      } # for all zap tags
      # print "$in_file was tested for tags to zap.\n";
   } # if zap tags to remove
   undef(@pre_chunks);
   undef(@post_chunks);
   #################################

   #################################
   # Change the title if appropriate.
   # Add the title to the index token list.
   # Add headings to the list to be included in TOC.
   #################################
   $tit_flag = 0;
   @title_chunks = split ( /$globe::m_define{title}[0]|$globe::m_define{title}[1]/, $globe::entire_file, 3);
   # Changed to always have it do the title.
   # if ($title_chunks[1] =~ /$globe::m_default_title/i) {
   &title_in_topic_jumps;

   if ((0) && (@globe::index_info >0)) { #debug
      foreach $thing (@globe::index_info){
          print "---- $thing\n";
      }
      # exit(1);
   } # debug
   if (0) { #debug
      print "---- Entire file 5 ----\n$globe::entire_file\n-----5 \n";
      exit(1);
   } # debug
   undef(@chunks);
   undef(@title_chunks);
   undef(@h1_chunks);
   undef(@h1_end_chunks);
   #################################

   #################################
   # Handle <object> tags and change if appropriate.
   # This will turn them into voyant index tokens
   # and adds them to the token list.
   # If the file gets processed again, it handles
   # the index tokens by putting them in a list.
   #################################
   if ($globe::entire_file =~ /$globe::ms_object_b/){
      # 07/03/2002 We are no longer exporting MSHTML that uses objects.
      # Hence if we come across this, we should issue an error.
      push (@file_errors, "HTML file has ms object index tokens; re-export with new mif2htm.ini file tokens.");
      
      if (0) { # 07/03/2002
         # convert object keywords into voyant index tokens; put them in list
         # need to keep pre_object_chunk[0] and post_obj_chunk[1]
         while ($globe::entire_file =~ /$globe::ms_object_b/){
            @pre_obj_chunk = split ( /$globe::ms_object_b/, $globe::entire_file, 2);
            @post_obj_chunk = split ( /$globe::ms_object_e/, $pre_obj_chunk[1], 2);
            $temp_voy_token = "";
            while ($post_obj_chunk[0] =~ /$globe::ms_value_b/){
               @pre_value = split ( /$globe::ms_value_b/, $post_obj_chunk[0], 2);
               @post_value = split ( /$globe::ms_value_e/, $pre_value[1], 2);
               # $post_value[0] contains keyword
               $temp_voy_token .= "$globe::m_define{index_token}[0]$post_value[0]$globe::m_define{index_token}[1]"; 
               $post_value[0] =~ s/\,\s/$globe::word_c_boundary/;
               # update the index token list appropriately
               push (@globe::index_info, "$post_value[0]$globe::word_url_boundary<a href=\"$rel_to_file\" target=\"basefrm\">");
               # removes value line from object tag 
               $post_obj_chunk[0] = join ("", $pre_value[0], $post_value[1]);  
            } # while $globe::ms_value_b
            # rebuilt file and remove object; replace with voyant token.
            $globe::entire_file = join ("", $pre_obj_chunk[0], $temp_voy_token, $post_obj_chunk[1]);
         } # while $globe::ms_object_b still present
      } # 07/03/2002
   } elsif ($globe::entire_file =~ /$globe::m_define{index_token}[0]/) {
      # 07/03/2002 We are no longer using the old token markers.
      # All index tokens come either from doxytag or mif2go.
      # Hence, we'll remove the old markers
      push (@file_errors, "HTML file has old index tokens; re-export with new mif2htm.ini file tokens.");

      if (1) { # 07/03/02 This will eventually go away
         # check for voyant index tokens; put them in list
         @pre_token = split ( /$globe::m_define{index_token}[0]/, $globe::entire_file);
         # purposely skip 0, because that's before the token
         for ($i=1; $i<@pre_token; $i++){
            @post_token = split ( /$globe::m_define{index_token}[1]/, $pre_token[$i], 2);
            # post_token[0] has the token
            # print "$post_token[0] in $in_file\n";
            $post_value[0] =~ s/\,\s/$globe::word_c_boundary/;
            push (@globe::index_info, "$post_token[0]$globe::word_url_boundary<a href=\"$rel_to_file\" target=\"basefrm\">");
         } # for $i the tokens
         # print "$globe::entire_file\n";
      } # 07/03/02 This will eventually go away
   }
   undef (@pre_obj_chunk);
   undef (@post_obj_chunk);
   undef ($temp_voy_token);
   undef (@pre_token);
   undef (@post_token);
   #################################
   
   #################################
   # 07/03/2002 Handle the new index tokens
   #################################
   if ($globe::entire_file =~ /$globe::m_define{index_token2}[2]/) { # if new index tokens 
      $t_shrinking_all = $globe::entire_file;
      $not_critical = 0;
      while ($t_shrinking_all =~ /$globe::m_define{index_token2}[0]/) {
         ($before, $piece, $after) = &globe::get_tag_chunk( $t_shrinking_all,
                              $globe::m_define{index_token2}[0], 
                              $globe::m_define{index_token2}[1], 
                              $not_critical);
      if (0) {
         print "piece=>$piece\n";
      }
         if ($piece =~ /$globe::m_define{index_token2}[2]/) {
            # this is an index token
            @url = split ("\"", $piece, 2);
            # $url[0] contains the target name
            
            # $globe::m_define{index_token2}[4] => value=\"
            # $globe::m_define{index_token2}[5] => \"
            # The value is the text we desire
            ($before2, $piece2, $after2) = &globe::get_tag_chunk( $piece,
                              $globe::m_define{index_token2}[4], 
                              $globe::m_define{index_token2}[5], 
                              $not_critical);
            
            # $piece2 contains the index text that we desire
            if ($piece2 =~ /\:[\s]+/) {
               # This handles multi-level index tokens from FM
               $piece2 =~ s/\:[\s]+/$globe::word_c_boundary/;
            }
            ####
            # Build an index token
            ####
            push (@globe::index_info, "$piece2$globe::word_url_boundary<a href=\"$rel_to_file\#$url[0]\" target=\"basefrm\">");            
            if (0) {
               print "New Index=>$piece2$globe::word_url_boundary<a href=\"$rel_to_file\#$url[0]\"...\n";
            }
         }
         # Rebuild without targets so while loop stops and go on.
         $t_shrinking_all = join ("", $before, $after);
      } # shrinking while
      if (0) {
         print "we're stopping here.\n";
         print "0 $globe::m_define{index_token2}[0]\n";
         print "1 $globe::m_define{index_token2}[1]\n";
         print "2 $globe::m_define{index_token2}[2]\n";
         print "3 $globe::m_define{index_token2}[3]\n";
         print "4 $globe::m_define{index_token2}[4]\n";
         print "5 $globe::m_define{index_token2}[5]\n";
         exit(1);
      }
   } # if new index tokens 

   #################################
   # Old version:
   # Items from group files to add to index.
   # If the file name starts with "group", look for elements to add to index.
   # 
   # <a name="xx" doxytag="group::code_item"></a> is the general syntax.
   # Hence to find them, we'll use:
   #local ($globe::gi_filename_clue) = "group";  # clue that files beginning with "group" need additional parsing.
   #local ($globe::gi_name) = "<a name\=";  # gi: first clue to look for for index
   #local ($globe::gi_dox) = "doxytag\="; # gi: second clue to look for for index
   #local ($globe::gi_group_separator) = "\:\:"; # gi: separator from group and code item; code item is indexed
   #
   #################################
   # if (($rel_to_file =~ /$globe::gi_filename_clue/i) && ($globe::entire_file =~ /$globe::gi_dox/i)){
   
   #################################
   # New version:
   # Items to add to index, doesn't matter what the file is named.
   # If the file has $globe::gi_dox = ("doxytag\="),
   # that target anchor will be used to create an index entry.
   # 
   # <a name="xx" doxytag="group::code_item"></a> is the general syntax.
   # Hence to find them, we'll use:
   #local ($globe::gi_name) = "<a name\=";  # gi: first clue to look for for index
   #local ($globe::gi_dox) = "doxytag\="; # gi: second clue to look for for index
   #local ($globe::gi_group_separator) = "\:\:"; # gi: separator from group and code item; code item is indexed
   #
   #################################
   if ($globe::entire_file =~ /$globe::gi_dox/i){
      #####
      # if the file is a group file
      # AND if it contains at least one doxytag
      # THEN it is acceptable to add some more tokens
      #####
      @grobe_target_chunks = split (/$globe::gi_name/i, $globe::entire_file);
      foreach $gt_chunk (@grobe_target_chunks){
         # Get to the end of the name tag
         if ($gt_chunk =~ /$globe::gi_dox/i){
            @fine_target_chunks = split (/\>/, $gt_chunk, 2);
            # print "interested in: $fine_target_chunks[0]\n";
            @target_n_item = split (/$globe::gi_dox/i, $fine_target_chunks[0]);
            $ttt_target = $target_n_item[0];
            $ttt_target =~ s/\"//g;
            $ttt_target =~ s/\s//g;
            # print "2nd part interested in: $target_n_item[1]\n";
            @item_chunk = split (/$globe::gi_group_separator/, $target_n_item[1], 2);
            $ttt_item = $item_chunk[1]; 
            $ttt_item =~ s/\"//g;
            $ttt_item =~ s/\s//g;
            $ttt_item =~ s/\(//g;
            $ttt_item =~ s/\)//g;
            if (0){ #debug
               print "$ttt_item$globe::word_url_boundary<a href=\"$rel_to_file\#$ttt_target\" target=\"basefrm\">\n";
            } # debug
            if ($ttt_item =~ /\w+/){
               # if not empty; don't write any empty index tokens
               $ttt_save_url = "<a href=\"$rel_to_file\#$ttt_target\" target=\"basefrm\">";
               push (@globe::index_info, "$ttt_item$globe::word_url_boundary$ttt_save_url");
            }
            #####
            # Finshes the normal index tokens
            #####
            
            #####
            # Handle any "Used by" references while we're here;
            # These are specifically designed for the BAPI Acknowledgement Packets.
            #
            # $ttt_item from above contains important title material that will  
            # be part of the second level of an index token.
            #
            # $ttt_save_url is also important as the target for this new index token
            #
            # At this point, the $gt_chunk may or may not contain $ack_text_flag.
            # If it does, let's get that and make an index token.
            #####
            if (1){  # glenn your test
            $_ig_flag = 0;
            if ($gt_chunk =~ /$globe::ack_text_flag/){
               # print "got the text flag; =$globe::ack_text_flag=\n$gt_chunk\n";
               @ack_split1 = split (/$globe::ack_text_flag/, $gt_chunk, 2);
               @ack_split2 = split (/$globe::ack_end/, $ack_split1[1], 2);
               foreach $_ig_thing (@globe::ack_ignore){
                  # print "A thing to ignore:$_ig_thing\n";
                  if ($ack_split2[0] =~ /$_ig_thing/) {
                      # then we're done
                      $_ig_flag = 1; 
                  }
               }
               if ($_ig_flag == 0) { 
                  # Continue to create a token
                  # strip out HTML codes first
                  while ($ack_split2[0] =~ /\</) {
                     @_ack_html_chunk1 = split (/\</, $ack_split2[0], 2);
                     @_ack_html_chunk2 = split (/\>/, $_ack_html_chunk1[1], 2);
                     $ack_split2[0] = "$_ack_html_chunk1[0]$_ack_html_chunk2[1]";
                  }
                  $ack_split2[0] =~ s/\n//g;
                  $ack_split2[0] =~ s/^\s*//;
                  $ack_split2[0] =~ s/\,//;
                  # print "in>>>>> $ack_split2[0]\n";
                  @ack_split3 = split (/\s+/, $ack_split2[0]);
                  NEXT_ACK_ITEM: foreach $_ack_item (@ack_split3){
                     $_ack_item =~ s/\s*$//g;
                     $_ack_item =~ s/\,$//g;
                     if (0){
                        print "out === $_ack_item\n";
                     }
                      
                     if ("or" =~ /$_ack_item/){
                        next NEXT_ACK_ITEM;
                     }
                     if (0){
                        print "$_ack_item$globe::word_c_boundary$globe::ack_used_by$ttt_item\n";
                     }
                     if (0){
                        print "$_ack_item$globe::word_c_boundary$globe::ack_used_by$ttt_item$globe::word_url_boundary$ttt_save_url\n";
                     }
                     push (@globe::index_info, "$_ack_item$globe::word_c_boundary$globe::ack_used_by$ttt_item$globe::word_url_boundary$ttt_save_url");
                  } 
               } # ignore flag
            }
            } # glenn test
            
         } # if for doxytag
      } # foreach grobe target chunk 
      if (0){ #debug
         print "Here's one: $rel_to_file\n";
         exit(1);
      } #debug
   } # if the file is a group file and we want index access
   #################################

      
   print (OUT_HTML "$globe::entire_file");
   close (OUT_HTML);
   
   # gcm Where you can stop it from overwriting files.  
   system ("cp $out_file $in_file");
   
   undef ($in_file);
   undef ($out_file);
   undef ($globe::entire_file);
   undef ($globe::have_header);
   undef ($globe::have_footer);
   undef ($globe::have_common_top);
   undef ($globe::have_dox);
   return (1);
} # change_nav


##############################################################################
# 06/21/2002 Tore out of change_nav in order to tweak output.
##############################################################################
sub title_in_topic_jumps {
   if (0) {
      print "Entering title_in_topic_jumps\n";
   }
   if (1) {

      # 06/24/2002 Remove all targets that we created from previous run
      while ($globe::entire_file =~ /\<a[\s]+name[\s]*\=\"$target_prefix/i) { 
         ($before, $piece, $after) = &globe::get_tag_chunk( $globe::entire_file,
                              "<a name=\"$target_prefix", 
                              "><\/a>", 
                              $very_critical);
         if (0) {
            print "Removed target:::$piece\n";
         }
         $globe::entire_file = join ("", $before, $after);
      }

      #####
      # find an appropriate h1 to use as the title in the document
      # Start by dividing into beg/end heading chunks
      #####
      @h1_start_chunks = split ( /$globe::voy_h1_b/, $globe::entire_file);
      for ($f=1; $f<@h1_start_chunks; $f++){
         #####
         # purposely ignore the first chunk, because that's uninteresting
         #####
         @h1_rear_chunks = split ( /$globe::voy_h1_e/, $h1_start_chunks[$f], 2);
         #####
         # h1_start have all where ever an H1 started.
         # h1_rear subdivide that even further: 
         # h1_rear[0] will have the title in there somewhere with other HTML formatting.
         # h1_rear[1] will have the rest of the topic; uninteresting
         # Get rid of left over h1 tag info.
         #####
         @h1_title_chunks = split ( /\>/, $h1_rear_chunks[0], 2);

         #####
         # Handle the book topic level
         # 02/11/2002 Changed to start at $i=1; equality increased to $i >= 5;
         # equivalent level is set to [1].
         #####
         FND_LVL2: for ($i=1; $i<@globe::fm_format_levels; $i++){
#            if ($h1_title_chunks[0] =~ /class\=\"$globe::fm_format_levels[$i]\"/){
            # Removed closing (") so that it will match on "Head#Kwp", too.
            if ($h1_title_chunks[0] =~ /class\=\"$globe::fm_format_levels[$i]/){
               # Manual fake out: BE CAREFUL; Index references set at beginning
               if ($i >= 5) {
                  # TitleChapter, TitleAppendix, TitleFront should be given equal weighting,
                  # Hence assign TitleChapter [0] to what would have been TitleAppendix [4]
                  # or TitleFront [5].
                  $this_h1_level = $globe::fm_format_levels[1]; # key is html file name, <Headn> is stored
               } else {
                  $this_h1_level = $globe::fm_format_levels[$i]; # key is html file name, <Headn> is stored
               }
               next FND_LVL2;
            }
         }
         # print "$this_h1_level and Hi\n";

         
         #####
         # Remove any other HTML code that might be within the title.
         # This includes targets...
         #####
         $have_target = 0;
         $h1_target = 0;
         while ($h1_title_chunks[1] =~ /\</) {
            @chunks = split ( /\<|\>/ , $h1_title_chunks[1], 3);
            $h1_title_chunks[1] = join ("", $chunks[0], $chunks[2]);
            if (($chunks[1] =~ /a[\s]+name[\s]*\=/i) && ($have_target < 1)){
               # Get the named target
               @tar_b_chunks = split (/[\s]+name[\s]*\=[\s]*\"/i, $chunks[1], 2);
               @tar_e_chunks = split (/\"/, $tar_b_chunks[1], 2);
               # 06/21/2002 Throw out targets generated by me in a previous run
               if ($tar_e_chunks[0] !~ /$target_prefix/) {
                  # This is a target we can live with
                  $h1_target = $tar_e_chunks[0];
                  $have_target++;
               }
            }
            undef (@tar_b_chunks);
            undef (@tar_e_chunks);
            undef (@chunks);
         }
         
         # 06/21/2002 Create unique targets
         if (!$h1_target) {
            $h1_target = sprintf ("%s%u", $target_prefix , $target_cnt);
            $target_cnt++;
            
            # insert the name target into the file at that location
            # build up the pieces so it can be done later
            @anchor_chunks = split ( /\>/, $h1_rear_chunks[0], 2);
            $anchor_chunks[1] = "<a name=\"$h1_target\"\>\<\/a\>" .$anchor_chunks[1];
            $h1_rear_chunks[0] = join ("\>", @anchor_chunks);
            $h1_start_chunks[$f] = join ("$globe::voy_h1_e", @h1_rear_chunks);
            if (0) {
               print "glenn $h1_target\n";
               # print "glenn $h1_start_chunks[$f]\n";
            }
         }
         
         # remove bad characters from name
         $h1_title_chunks[1] =~ s/\(//g;
         $h1_title_chunks[1] =~ s/\)//g;
         $h1_title_chunks[1] =~ s/[\s]+/ /g;
         # $h1_title_chunks[1] =~ s/  / /g;
         if (0) { # debug
            print "Found and parsed h1 title to use = $h1_title_chunks[1]\n";
            print "Internal target = $h1_target\n";
            print "$this_h1_level\n";
            # exit(1);
         }

         #####
         # Take care of new Title and index
         #####
         if ($tit_flag < 1) {
            $tit_flag++;
            # Rebuild Title and file
            $title_chunks[1] = "$globe::m_define{title}[0]$h1_title_chunks[1]$globe::m_define{title}[1]";
            $globe::entire_file = join ("", @title_chunks);
            #####
            # Put information into index
            #####
            push (@globe::index_info, "$h1_title_chunks[1]$globe::word_url_boundary<a href=\"$rel_to_file\" target=\"basefrm\">");
         } else {
            #####
            # title taken care of, but need to index h1.
            # Put information into index
            #####
            push (@globe::index_info, "$h1_title_chunks[1]$globe::word_url_boundary<a href=\"$rel_to_file\#$h1_target\" target=\"basefrm\">");
         }


         #####
         # Create book topic lists
         # Handle the book topic level
         #####
         $book_topics[$f] = $h1_title_chunks[1];
         $book_targets[$f] = $h1_target; 
         $book_topic_level[$f] = $this_h1_level; # key is html file name, <Headn> is stored

         if (0) { #debug
            print "---- Entire file ----\n$globe::entire_file\n-----\n";
            exit(1);
         } # debug
         
      } # for $f the number of <h1> tags
      
      # 06/21/2002 Rebuild the file with targets inside
      if ($#h1_start_chunks > 1) {
         $globe::entire_file = join ("$globe::voy_h1_b", @h1_start_chunks);
      }
      

      if (0){
         for ($f=1; $f<= $#h1_start_chunks; $f++){
             print "$f = $book_topic_level[$f]\n";
             print "$f = $book_topics[$f]\n";
             # print "$f = $book_targets[$f]\n";
         }
      }
      
      #####
      # Loop through again to take care of topic list
      #####
      for ($f=1; $f<= $#h1_start_chunks; $f++){
         #####
         # Create book topic lists
         #####
         if (0) {
            print "$f = $book_targets[$f]\n";
         }
         if (($f == 1) && ($f == $#h1_start_chunks)) {
            # if first element and only one element
            $globe::book_topics{$in_file} = $book_topics[$f];    # key is html file name, <Title> is stored
            $globe::book_topic_order{$in_file} = $globe::m_var{value}{next};    # key is html file name, next html file is stored
            $globe::book_topic_level{$in_file} = $book_topic_level[$f]; # key is html file name, <Headn> is stored
         } elsif (($f == 1) && ($f < $#h1_start_chunks)) {
            # if first element and more than one element
            $globe::book_topics{$in_file} = $book_topics[$f];    # key is html file name, <Title> is stored
            $globe::book_topic_order{$in_file} = "$globe::m_var{value}{curr}\#$book_targets[$f+1]";    # key is html file name, next html file is stored
            $globe::book_topic_level{$in_file} = $book_topic_level[$f]; # key is html file name, <Headn> is stored
         } elsif (($f > 1) && ($f < $#h1_start_chunks)) {
            # if not first element and not last element
            $thing = "$in_file\#$book_targets[$f]";
            $globe::book_topics{$thing} = $book_topics[$f];    # key is html file name, <Title> is stored
            $globe::book_topic_order{$thing} = "$globe::m_var{value}{curr}\#$book_targets[$f+1]";    # key is html file name, next html file is stored
            $globe::book_topic_level{$thing} = $book_topic_level[$f]; # key is html file name, <Headn> is stored
         } elsif (($f > 1) && ($f >= $#h1_start_chunks)) {
            # if not first element and the last element
            $thing = "$in_file\#$book_targets[$f]";
            $globe::book_topics{$thing} = $book_topics[$f];    # key is html file name, <Title> is stored
            $globe::book_topic_order{$thing} = $globe::m_var{value}{next};    # key is html file name, next html file is stored
            $globe::book_topic_level{$thing} = $book_topic_level[$f]; # key is html file name, <Headn> is stored
         }
         
      } # for $f the number of <h1> tags
      
   } # if there are titles
   
   if (0) { #debug
      $thing_cnt = 0;
      print "======= this is what we know ======\n";
      if (0){ # inner debug
         foreach $thing (keys %globe::book_topics){
            print "$thing = topic = $globe::book_topics{$thing}\n";
            $thing_cnt++;
         }
         print "Thing cnt $thing_cnt\n";
         $thing_cnt = 0;
         foreach $thing (keys %globe::book_topic_level){
            print "$thing = level = $globe::book_topic_level{$thing}\n";
            $thing_cnt++;
         }
         print "Thing cnt $thing_cnt\n";
         $thing_cnt = 0;
         foreach $thing (sort keys %globe::book_topic_order){
            print "$thing = order = $globe::book_topic_order{$thing}\n"; 
            $thing_cnt++;
         }
         print "Thing cnt $thing_cnt\n";
         $thing_cnt = 0;
         foreach $thing (sort keys %globe::book_topics){
            print "$globe::book_topic_level{$thing} $globe::book_topics{$thing}\n"; 
            $thing_cnt++;
         }
         print "Thing cnt $thing_cnt\n";
         $thing_cnt = 0;
         print "----- all together now -----\n";
         foreach $thing (keys %globe::book_topics){
            print "$globe::book_topic_level{$thing} = $thing = $globe::book_topics{$thing} = $globe::book_topic_order{$thing}\n"; 
            $thing_cnt++;
         }
         print "Thing cnt $thing_cnt\n";
         $thing_cnt = 0;
      } # inner debug
         print "----- all together now -----\n";
         foreach $thing (keys %globe::book_topics){
            print "dd $globe::book_topic_level{$thing} = $thing = $globe::book_topic_order{$thing}\n"; 
            $thing_cnt++;
         }
         print "Thing cnt $thing_cnt\n";
         $thing_cnt = 0;
      # exit(1);
   }

   if ((0) && (@globe::index_info >0)) { #debug
      foreach $thing (@globe::index_info){
          print "---- $thing\n";
      }
      # exit(1);
   } # debug
   if (0) { #debug
      print "---- Entire file 5 ----\n$globe::entire_file\n-----5 \n";
      exit(1);
   } # debug
   undef(@chunks);
   undef(@title_chunks);
   undef(@h1_chunks);
   undef(@h1_end_chunks);
   #################################

} # title_in_topic_jumps




##############################################################################
##! @fn int file_chunk_change
##! @brief Replaces the information for the given chunk. (Not complete).
##! 
##! @param _type_define The key into the m_define and m_info hashes.
##! @param _criticality specifies whether or not to record error messages.
##! @param _where_flag The tag where this is supposed to go.
##! @param _after if 0, set before, otherwise after
##! @param _in_file Name of file being worked on.
##!
##! @retval success 1 if found tags, 0 if missing.
##! 
##! @bug Uses the $globe::entire_file
##! @ingroup tp_toc
##############################################################################
sub file_chunk_change {
   $_type_define = $_[0];
   $_criticality = $_[1];
   $_where_flag = $_[2];
   $_after = $_[3];
   $_insert_tags = $_[4];  # pass through
   $_in_file = $_[5];

   if (0) {
      print "file_chunk_change $_type_define $globe::m_define{$_type_define}[0]\n$_in_file\n";
   }
   if (( $globe::entire_file =~ /$globe::m_define{$_type_define}[1]/ ) && ( $globe::entire_file =~ /$globe::m_define{$_type_define}[0]/ ) ) {
      ($okay, $globe::entire_file) = &globe::replace_tag_chunk ( $globe::entire_file,
                $globe::m_define{$_type_define}[0], 
                $globe::m_define{$_type_define}[1], 
                $globe::m_info{$_type_define}, $_insert_tags, $_criticality);
      if (!$okay) {
         push (@file_errors, "$_type_define definition messed up in $in_file.");
      }
      if (0){
         print "file_chunk_change ok=$okay\n";
      }

   } else {
      # Figure out where the $_where_flag should go
      if ( $globe::entire_file =~ /$_where_flag/ ) {
         @pre_chunks = split ( /$_where_flag/, $globe::entire_file, 2);
         if ($_after) {
            $globe::entire_file = sprintf("%s%s%s%s", 
               $pre_chunks[0], 
               $_where_flag,
               $globe::m_info{$_type_define}, 
               $pre_chunks[1]);
         } else {
            $globe::entire_file = sprintf("%s%s%s%s", 
               $pre_chunks[0], 
               $globe::m_info{$_type_define}, 
               $_where_flag,
               $pre_chunks[1]);
         }
      } else {
         push (@file_errors, "WARNING: No $_where_flag defined in $_in_file.");
         # Do only if no head/body tags
         print "WARNING: No $_where_flag defined.\n";
         return (0);
      } # figuring out where it should go
   }
   return (1);

} # file_chunk_change




##############################################################################
##! @fn int organize_topic_order
##! @brief Organizes the topic orders.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub organize_topic_order {
   if (0) {
      print "organize_topic_order...\n";
   }
   if (@globe::book_first_topic > 0) {
      # First topics that I know about
      if (0) {
         print "==== First topics not yet ordered ====\n";
         for ($i=0; $i<@globe::book_first_topic; $i++){
            print "$globe::book_first_topic[$i]\n";
         }
         print "Master Chapter Order ===\n";
         for ($i=0; $i<@globe::voy_master_topic_order; $i++){
            print "$i $globe::voy_master_topic_order[$i]\n";
         }
      }

      # Sort the book_first_topic array into the same order as the master_topic_order
      # The master_topic_order has more topics that the book_first_topic array
   
      for ($i=0; $i < @globe::book_first_topic-1; $i++) {
         for ($j=$i+1; $j < @globe::book_first_topic; $j++) {
            # Look for the location in the first_topic list where the item is;
            # Swap it with the one where it should be.
            if ($globe::book_first_topic[$j] < /$globe::book_first_topic[$i]/i) { # case insentive
               # Swaps the topics with where it should be
               $temp = $globe::book_first_topic[$j];
               $globe::book_first_topic[$j] = $globe::book_first_topic[$i];
               $globe::book_first_topic[$i] = $temp;
            } else {
               # leave alone.
            } # if looking for swap
         } # for #j
      } # for $i
   
      # Sort the book_last_topic array
      for ($i=0; $i < @globe::book_last_topic-1; $i++) {
         for ($j=$i+1; $j < @globe::book_last_topic; $j++) {
            # Look for the location in the first_topic list where the item is;
            # Swap it with the one where it should be.
            if ($globe::book_last_topic[$j] < /$globe::book_last_topic[$i]/i) { # case insentive
               # Swaps the topics with where it should be
               $temp = $globe::book_last_topic[$j];
               $globe::book_last_topic[$j] = $globe::book_last_topic[$i];
               $globe::book_last_topic[$i] = $temp;
            } else {
               # leave alone.
            } # if looking for swap
         } # for #j
      } # for $i
      if (0){
         for ($i=0; $i < @globe::book_last_topic-1; $i++) {
            print "$i = $globe::book_first_topic[$i] == $globe::book_last_topic[$i]\n";
         }
      }
      # clean up last topic of a chapter
      &first_last_prev_next_at_boundary();
      # create a tree file for the fm book
      &create_fm_nav_tree();
   }
} # organize_topic_order

##############################################################################
##! @fn int first_last_prev_next_at_boundary
##! @brief Fixes the boundaries between chapters to fix previous/next
##! references. 
##!
##! @param  @globe::book_first_topic hash table with the first topics for all
##! chapters.
##! @param  @globe::book_last_topic hash table with the last topics for all
##! chapters.
##!
##! @return Actually opens all first and last HTML files for a chapter and
##! replaces the previous/next links with the appropriate references. The
##! first and last chapters wrap to each other.
##!
##! This is called after all HTML files in the directory have been processed
##! and the hash tables for first and last topics filled in.
##!
##! When called, all first and last topics are known. This fixes the last topic
##! in a chapter which has an undefined "next" topic. This defines the next
##! topic as the first topic in the next chapter. If last chapter, it wraps
##! around to the very first topic in book.
##!
##! @bug In order for this to work correctly, the file names for the chapters
##! have to be named correctly so that they sort into a natural order. This
##! is achieved by having the generation place the same prefix on all HTML
##! files for a given chapter. I chose a two-digit number followed by an
##! underscore. In this manner, all files for a given chapter are grouped
##! together by the prefix. Earlier processing of the files determined which
##! of those topics was the first and last for the chapter.
##!
##! @ingroup tp_nav
##############################################################################
sub first_last_prev_next_at_boundary {
   local ($out_file) = "_temp";
   local ($temp_content) = "";
   local ($temp_html) = "";
   
NEXT_FIRST:   for ($i=0; $i < @globe::book_first_topic ; $i++){
      # print "$i = first = $globe::book_first_topic[$i]\n$i = last= $globe::book_last_topic[$i]\n";
      unless (open(IN_FIRST, "$globe::book_first_topic[$i]")) {
         print "Cannot open file \"$globe::book_first_topic[$i]\"\n";
         next NEXT_FIRST;
      }
      $temp_content = "";
      while (<IN_FIRST>){
         $temp_content .= $_;
      }
      close (IN_FIRST);
      # parse what the next topic should be
      if ($i !~ /0/) {
         @h_chunk = split ( /\//, $globe::book_last_topic[$i-1]);
      } else {
         # wrap around to last topic in chapter
         @h_chunk = split ( /\//, $globe::book_last_topic[@globe::book_last_topic-1]);
      }
      $temp_html = $h_chunk[@h_chunk-1];
      # print "$i = prev file= $temp_html\n";
      # Replace the next topic holder with what it should be
      if ($temp_content =~ /$globe::voy_variable_prev_chapter/i) {
         $temp_content =~ s/$globe::voy_variable_prev_chapter/$temp_html/g;
         unless (open(OUT_FIRST, ">$globe::path$out_file")) {
            print "Cannot open file \"$globe::path$out_file\"\n";
            next NEXT_LAST;
         }
         print (OUT_FIRST $temp_content);
         close (OUT_FIRST);
         system ("cp $globe::path$out_file $globe::book_first_topic[$i]");
      } # if there is something to replace

   } # for all first topics in array

# Modifies the last file to wrap and point to the first.

NEXT_LAST:   for ($i=0; $i < @globe::book_last_topic ; $i++){
      # print "$i = first = $globe::book_first_topic[$i]\n$i = last= $globe::book_last_topic[$i]\n";
      unless (open(IN_LAST, "$globe::book_last_topic[$i]")) {
         print "Cannot open file \"$globe::book_last_topic[$i]\"\n";
         next NEXT_LAST;
      }
      $temp_content = "";
      while (<IN_LAST>){
         $temp_content .= $_;
      }
      close (IN_LAST);
      $test_value = @globe::book_last_topic-1;
      # parse what the next topic should be
      if ($i !~ /$test_value/) {
         @h_chunk = split ( /\//, $globe::book_first_topic[$i+1]);
      } else {
         # wrap around to first topic in chapter
         @h_chunk = split ( /\//, $globe::book_first_topic[0]);
      }
      $temp_html = $h_chunk[@h_chunk-1]; # file name
      # print "$i = next file= $temp_html\n";
      # Replace the next topic holder with what it should be
      if ($temp_content =~ /$globe::voy_variable_next_chapter/i) {
         $temp_content =~ s/$globe::voy_variable_next_chapter/$temp_html/g;
         unless (open(OUT_LAST, ">$globe::path$out_file")) {
            print "Cannot open file \"$globe::path$out_file\"\n";
            next NEXT_LAST;
         }
         print (OUT_LAST $temp_content);
         close (OUT_LAST);
         system ("cp $globe::path$out_file $globe::book_last_topic[$i]");
      } # if there is something to replace

   } # for all last topics in array
} # first_last_prev_next_at_boundary



##############################################################################
##! @fn int create_fm_nav_tree
##! @brief Creates a tree file which is the TOC for the FrameMaker book.
##!
##! @param @globe::book_first_topic hash table.
##! @param @globe::fm_format_levels hash table with the FrameMaker format
##! level for that topic. This is typically its heading number.
##!
##! @return Outputs a tree.html file that is the TOC for this book.
##!
##! In order to accomplish this, the hash tables with the topic order must
##! be generated first. This routine starts at the very first first_topic
##! and steps through its next links. Then it goes to the next first_topic
##! (e.g., the first topic of the next chapter) and steps through its next
##! links.
##!
##! When stepping through the topics, it notes the heading level. A 1-bit clear
##! GIF is used whose width is varied depending upon heading level to get
##! placement in the tree.html that is appropriate for the TOC.
##!
##! @bug In order for this to work correctly, the file names for the chapters
##! have to be named correctly so that they sort into a natural order. This
##! is achieved by having the generation place the same prefix on all HTML
##! files for a given chapter. I chose a two-digit number followed by an
##! underscore. In this manner, all files for a given chapter are grouped
##! together by the prefix. Earlier processing of the files determined which
##! of those topics was the first and last for the chapter.
##!
##! This used information that came specifically from our implementation
##! of styles in FrameMaker.
##!
##! @ingroup tp_nav
##############################################################################
sub create_fm_nav_tree {
   if (0) {
      print "gcm create_fm_nav_tree...\n";
      if (0) {
         print "first last $globe::m_var{value}{firstlast}\n";
      }
      if (1) {
         foreach $thing (keys %globe::book_topics){
            print "$thing = $globe::book_topic_level{$thing} = $globe::book_topic_order{$thing}\n"; 
            $thing_cnt++;
         }
      }
      # exit(1);
      print "++++++++++++++++\n";
   }
   local ($htree_file) = "tree.html";
   local ($script_file) = "tree.script";
   
   $num_start = $#globe::book_first_topic;
   $num_end = $#globe::book_last_topic;
   
   if ($num_start > $num_end) {
      # It is okay if there are more ending topics than starting.
      # Hopefully, all starting topics when traced through "next" will
      # eventually terminate on one of the ending topics.
      print "ERROR: number of chapter starts ($num_start) does not equal num of end of chapters ($num_end).\n";
      print "\t\tNo tree file created!!!\n";
      push (@file_errors, "ERROR: $in_file\n\tnumber of chapter starts ($num_start) does not equal num of end of chapters ($num_end).");
      return(1);
   } 
   
   $temp = $globe::book_first_topic[0];
   @name_chunks = split ( /\//, $temp);
   # if the syntax is:
   # doc_publish/book_innovox_api/File.htm
   # then the first element is doc_publish and the last is the file name.
   # I want the directory.
   $temp = "";
   for ($i=0; $i< $#name_chunks; $i++) {
      $temp .= "$name_chunks[$i]/";
   }
   $htree_file = "$temp$htree_file";
   $script_file = "$temp$script_file";
   
   print "Preparing to write new file = $htree_file\n";
   print "Preparing to write new file = $script_file\n";
   
   unless (open(OUT_HTREE, ">$htree_file")) {
      print "Cannot open file \"$htree_file\"\n";
      return(1);
   }
   unless (open(OUT_SCRIPT, ">$script_file")) {
      print "Cannot open file \"$script_file\"\n";
      return(1);
   }
 
   # Re-use $temp
   # CREATE TITLE
   # Title now comes from the navigation files.
   # $temp = $name_chunks[1];
   # $temp =~ s/book_//g;
   # $temp =~ s/cref_//g;
   # $temp =~ s/cit/CIT/g;
   # $temp =~ s/dtoss/DTOSS/g;
   # $temp =~ s/api/API/g;
   # $temp =~ s/_/ /g;
   # $temp =~ s/(\w+)/\u$1/g;   # changes first letter of words to be all caps
   
   $tcntgcm = 0;
   ##################################################
   # Root level of tree.
   ##################################################
   print (OUT_HTREE "<html>\n<head>\n$globe::m_info{header}\n</head>\n<body>\n");
   print (OUT_HTREE "$globe::m_define{topic_list}[0]");  # This does not have a return on purpose.
   print (OUT_HTREE "$globe::chapter_division");  # This does not have a return on purpose.
   print (OUT_HTREE "0");  # This does not have a return on purpose.
   print (OUT_HTREE "$globe::chapter_division_e");  # This does not have a return on purpose.
   print (OUT_HTREE "<pre class =\"navnarrow\"><a class=\"navigation\" href=\"$globe::empty_html\" target=\"treefrm\"><img src=\"nav_folderopen.gif\" border=\"0\"><b>$globe::master_order_title[$globe::master_order_key]</b></a></pre>\n");
   
   #####
   # whether or not nested script files are used.
   #####
   local ($nested_script) = 1;  # 0 means not, 1 means are
   if ($nested_script == 0) {
      #####
      # Not using nested script files
      #####
      print (OUT_SCRIPT "Item level=1 ");
      print (OUT_SCRIPT "image=nav_folderclosed.gif ");
      # print (OUT_SCRIPT "url=$temp_remember_url,basefrm ");
      print (OUT_SCRIPT "selected=YES ");
      print (OUT_SCRIPT "text=$globe::master_order_title[$globe::master_order_key]\r\n");
   } else {
      #####
      # Using nested script files
      # No title is necessary.
      #####
   }

   ##################################################
   
   ##################################################
   # This is the main loop; it starts at a first topic and then traces the next
   # topics from there. When a next topic is determined to be a last topic for
   # that section, it then goes and gets the next first topic.
   ##################################################
   for ($i=0; $i <= $#globe::book_first_topic; $i++) {
      $keyabs = $globe::book_first_topic[$i];
      @name_chunks = split ( /\//, $globe::book_first_topic[$i]);
      $html_name = $name_chunks[$#name_chunks];
      pop (@name_chunks);
      $html_path = join ("\/", @name_chunks);


      # $globe::last_book_level should be 0 at this point
      $globe::last_book_level = 0;
      $temp_remember_title = "_start_";
      $temp_remember_url = "_start_";
      $current_level = 0;    
      $remember_current_level = 0;

      # Default values
      $keep_going = 1;
      $almost_done = 0;
      
      # 06/24/2002 Changed so that it won't put dual entries into tree files
      # if it has a next that is the "next chapter".
      if ($globe::book_topic_order{$keyabs} =~ /$globe::voy_variable_next_chapter/) {  
          # This topic was found to be a last topic
          if (0) {
             print "++ending $keyabs $globe::book_topic_order{$keyabs}\n";
          }
          if ($globe::m_var{value}{firstlast} =~ /1\#1/) {
             # if this is the only html page, then don't do loop again.
             $keep_going = 0;
          }
      } # 06/24/2002

      ##################################################
      # Loop to trace through from the current topic to the next topic.
      # It keeps going until it determines that the next topic is the last
      # topic for that section.
      ##################################################

      do {  # keep_going
         if ($almost_done) {
            $keep_going = 0; # sets that this is the last time through; picks up very last element
         }
         if (0) {
            print "another one = $keyabs\n";
         }
         @name_chunks = split ( /\//, $keyabs);
         $html_name = $name_chunks[$#name_chunks];
         pop (@name_chunks);
         $html_path = join ("\/", @name_chunks);
      
         for ($lev_test=0; $lev_test <= $#globe::fm_format_levels; $lev_test++){
            if ($globe::book_topic_level{$keyabs} =~ /$globe::fm_format_levels[$lev_test]/ ) {
               # This is the level we're on
               $current_level = $lev_test;
            }
         } # lev_test
         for ($k=0; $k<=$current_level+1; $k++){
            # print out the appropriate number of leading spaces
            $temp = 10 * $k;
         } # for $k printing space

         #####
         # Made it always on so that we always have level.
         #####
         print (OUT_HTREE "$globe::chapter_division");
         print (OUT_HTREE "$current_level");
         print (OUT_HTREE "$globe::chapter_division_e");
         #####
         print (OUT_HTREE "<pre class =\"navnarrow\"><img src=\"clear.gif\" width=\"$temp\" height=\"0\">");
         if (0) { #debug
             print "$temp = ";
         } # debug

         $temp_remember_title = $globe::book_topics{$keyabs};
         $temp_remember_url = $html_name;
         #####
         # incremented for the script
         #####
         if ($nested_script == 0) {
            #####
            # Old:
            #   Not using nested script files.
            #####
            $temp_c_level = $current_level+1;
         } else {
            # New:
            #   So we can use nested script files
            #####
            $temp_c_level = $current_level;
         }

         print (OUT_HTREE "<a class=\"navigation\" href=\"$temp_remember_url\" target=\"basefrm\"><img src=\"nav_doc.gif\" border=\"0\"><b>$temp_remember_title</b></a></pre>\n");
         print (OUT_SCRIPT "Item level=$temp_c_level ");
         print (OUT_SCRIPT "image=nav_doc.gif ");
         print (OUT_SCRIPT "url=$temp_remember_url,basefrm ");
         print (OUT_SCRIPT "selected=YES ");
         print (OUT_SCRIPT "text=$temp_remember_title\r\n");
         $tcntgcm++;
         if (0) { #debug
             # print "gcm $html_name\n";
             print "$temp_remember_title\n";
         } # debug


         ##################################################
         # Determine where the next topic is tracing it
         # in hash.
         ##################################################
         $next = $globe::book_topic_order{$keyabs};  # $next doesn't have full path
         $next_abs = join ("\/", $html_path, $next); # $next_abs has full path
         $found_next = 0;
         if (0) {# debug
            print "Next topic=$next; $next_abs\n";
         } # debug
         
         ##################################################
         # Find the next topic in the huge hash
         ##################################################
         SEARCH_NEXT: foreach $key (keys %globe::book_topic_order){
            if (0) {# debug
               print "1 Next topic=$key\n";
            } # debug
            if (($key =~ /$next_abs/i) && ($next_abs =~ /$key/i)) {
               # handles partial matches so that only exact matches win
               $keyabs = $key;     # This is the next topic
               $found_next = 1;
               if (0) {# debug
                  print "2 Next topic=$keyabs\n";
               } # debug
               last SEARCH_NEXT;
            }
         } # foreach $key

         ##################################################
         # Determine if topic is a last topic.
         # The first_topic and last_topic array do not match index-index
         # Check each possible last topic with the current topic to see if last
         ##################################################
         if (1) { # 06/24/2002 
           if ($globe::book_topic_order{$keyabs} =~ /$globe::voy_variable_next_chapter/) {  
               # This topic was found to be a last topic
               if (0) {
                  print "++ending $keyabs $globe::book_topic_order{$keyabs}\n";
               }
               if ((0) && ($globe::m_var{value}{firstlast} =~ /1\#1/)) {
                  # if this is the only html page, then don't do loop again.
                  $keep_going = 0;
               } else {
                  $almost_done = 1;
               }
           }
         } # 06/24/2002

         # The following test is no longer absolutely valid; needs tweaking.
         if (($found_next =~ /0/) && ($globe::book_topic_order{$keyabs} !~ /$globe::voy_variable_next_chapter/)) {
            # The desired next topic wasn't parsed, so loop won't end
            print "WARNING: ($keyabs) references non-existent next topic ($next).\n";
            push (@file_errors, "WARNING: ($keyabs) references non-existent next topic ($next).");
            # Terminate this loop.
            $keep_going = 0;
         }
         ##################################################

      } while ($keep_going);

   } # for first topics... go get next first topic (beginning of new section)

   ##################################################
   # Clean up navigation.
   ##################################################
   print (OUT_HTREE "$globe::m_define{topic_list}[1]");  
   print (OUT_HTREE "\n$globe::m_info{footer}\n</body>\n</html>\n");
   ##################################################
   
   print "Total number of topics $tcntgcm\n";
   close (OUT_HTREE);
   close (OUT_SCRIPT);

} # create_fm_nav_tree



##############################################################################
##! @fn int END
##! @brief Code to execute when first entered.
##!
##! @param None. 
##!
##! @return None.
##!
##! @bug None
##! @ingroup tp_nav
##############################################################################
sub END {
   undef ($_file_list);     # "_file_list";
   undef ($_index_file); # = "_index_list";
   undef ($in_file);     # "";
   undef ($f_type);     # "htm";

   
   ##############################################################################
   ## Memory clean-up.
   ##############################################################################
   if ($no_scope_file > 0){
      &globe::memory_clean_up();
   }

   if (@file_errors) {
      print "\n============  Summary of errors ==================================\n";
      for ($i=0; $i<@file_errors; $i++){
         print "$i = $file_errors[$i]\n";
      }
   }

   print "\n============  Finished voyant_nav.pl ==================================\n";
} # END

