///////////////////////////////////////////////////////////////
//! @file
//! @brief Defines the data structures used to convert an
//! xml file into another format.
//!
//! This is defined in C++ for easy incorporation into other
//! applications (e.g., diagMenu). However, it is used
//! by perl.
//!
//! The intent is to have the definition in one place so
//! that changes to our in-house xml files can be understood
//! quickly by all applicable programs.
//!
// $Id: $
//
// Revision Information:
//
// $Log: $
///////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////
//! The xml tags look similar to:
//! <diagmenu item="/IMOD" audience="manufacturing">
//! </diagmenu item="/IMOD" audience="manufacturing">
//! 
//! The internal modifiers (e.g., item, audience) can be in
//! any order with any amount of white space in between.
//! There can also be any amount of white space on either side
//! of equals sign (=).
//!
//! The opening and closing tags (diagmenu) and modifiers
//! (item, audience) _are_ case sensitive.
//! The contents of the (item) modifier _is_ case sensitive.
//! The contents of the (audience) modifier is _not_ case 
//! sensitive.
//!
///////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////
//! @brief Specifies the starting and ending characters for
//! all xml tags.
//! %perl% @xmlTagDelimiter = ("<", ">", "</", ">")
//! $xmlTagDelimiter{start} = 
///////////////////////////////////////////////////////////////
const char *xmlTagDelimiter[4] = {"<", ">", "</", ">"};

///////////////////////////////////////////////////////////////
//! @brief Specifies the valid xml tags.
//! %perl% @xmlTags = ("diagmenu", "xhelp");
///////////////////////////////////////////////////////////////
const char *xmlTags[] = {"diagmenu", "xhelp"};

///////////////////////////////////////////////////////////////
//! @brief Specifies the allowable modifiers for a particular
//! xml tag.
//! %perl% @{$xmlTags{diagmenu}{modifier}} = ("item", "audience");
//! $xmlTags{diagmenu}{modifier}[0] = "item"
//! $xmlTags{diagmenu}{modifier}[1] = "audience"
///////////////////////////////////////////////////////////////
const char *diagmenu[] = {"item", "audience"};
const char *xhelp[] = {"item", "audience"};

///////////////////////////////////////////////////////////////
//! @brief Specifies the allowable values for a particular
//! xml tag modifier.
//! %perl% @{$xmlTags{diagmenu}{audience}} = ("customer", "manufacturing", "engineering");
//! $xmlTags{diagmenu}{audience}[0] = "customer";
//! $xmlTags{diagmenu}{audience}[1] = "manufacturing";
//! $xmlTags{diagmenu}{audience}[2] = "engineering";
///////////////////////////////////////////////////////////////
const char *audience[] = {"customer", "manufacturing", "engineering"};

